/* 
    %W%
    %D% %T%

    PROGRAM NAME:  Manifold
   
    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.
   
    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   February 22, 1999                     */

#ifndef __MFATLAS_H__
#define __MFATLAS_H__

#include <stdlib.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C"
 {
#endif

#include <MFBase.h>

#include <MFKVector.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <MFNKMatrix.h>
#include <MFChart.h>
#include <MFImplicitMF.h>

struct MFAtlasSt;
typedef struct MFAtlasSt *MFAtlas;
#include <MFContinuationMethod.h>

MFAtlas MFCreateAtlas(MFImplicitMF);
void MFFreeAtlas(MFAtlas);

int MFAtlasAddChart(MFAtlas,MFNVector);
void MFAtlasReduceChartRadius(MFAtlas,int,double);
/*int MFAtlasAddChartWithApproxTS(MFAtlas,MFNVector,MFNKMatrix);*/
int MFAtlasAddChartWithApproxTS(MFAtlas,MFNVector,MFNKMatrix,double,double);

int MFAtlasK(MFAtlas);
int MFAtlasN(MFAtlas);

MFImplicitMF MFAtlasMF(MFAtlas);
int MFAtlasNumberOfCharts(MFAtlas);
MFChart MFAtlasChart(MFAtlas,int);
double MFAtlasChartRadius(MFAtlas,int);
MFNVector MFAtlasChartCenter(MFAtlas,int);
MFNVector MFAtlasCenterOfChart(MFAtlas,int);
MFNKMatrix MFAtlasChartTangentSpace(MFAtlas,int);
int MFAtlasIsPointInChart(MFAtlas,int,MFKVector);
int MFAtlasIsChartNearBoundary(MFAtlas,int);
void MFAtlasEvaluateChart(MFAtlas,int,MFKVector,MFNVector);

int MFAtlasNumberOfChartNeighbors(MFAtlas,int);
int MFAtlasChartNeighbor(MFAtlas,int,int);

int MFAtlasNumberOfChartsWithBoundary(MFAtlas);
int MFAtlasChartWithBoundary(MFAtlas,int);

int MFAtlasPointOnBoundaryInsideRegion(MFAtlas,MFNRegion,MFNVector,MFNKMatrix*,double*);

MFAtlas MFReadAtlas(FILE*);
void MFWriteAtlas(FILE*,MFAtlas);
void MFWriteChartCenters(FILE*,MFAtlas);
int MFReadChartCenters(FILE*,MFNVector**,MFNKMatrix**,double**);

void MFAtlasSetEpsilon(MFAtlas,double);
void MFAtlasSetRMin(MFAtlas,double);
void MFAtlasSetDotMin(MFAtlas,double);
void MFAtlasSetVerbose(MFAtlas,int);

void MFAtlasReduceChartRadius(MFAtlas,int,double);

struct MFChartSt;

MFAtlas MFComputeAtlas(MFContinuationMethod,MFImplicitMF,MFNRegion,MFNVector);
MFAtlas MFComputeAtlasWithTangent(MFContinuationMethod,MFImplicitMF,MFNRegion,MFNVector,MFNKMatrix);
MFAtlas MFComputeAtlasMultiple(MFContinuationMethod,MFImplicitMF,MFNRegion,int,MFNVector*);
MFAtlas MFComputeAtlasMultipleWithTangents(MFContinuationMethod,MFImplicitMF,MFNRegion,int,MFNVector*,MFNKMatrix*);
MFAtlas MFAnimateAtlas(MFImplicitMF,MFNRegion,MFNVector,double,int,int,char*,int,int,int);
MFAtlas MFComputeInvariantMap(MFImplicitMF,MFNRegion,MFNVector,MFNKMatrix,double,double,int,int);

double MFVolumeOfChart(struct MFChartSt*,MFNRegion);
double MFVolumeOfAtlas(MFAtlas,MFNRegion);

int MFAtlasAddChartWithAll(MFAtlas,MFNVector,MFNKMatrix,double);
int MFAtlasPointOnBoundaryWOProject(MFAtlas,MFNRegion,MFKVector);

int MFComputePointsOn3dEdge(double,double**,double,double,double,double,double,double);
int MFComputePointsOn3dPolygon(double,int,double*,int,double*,double*,double**,double**);

MFNVector MFAtlasGetPointOnBoundaryChart(MFAtlas,MFNRegion,int,double);
int MFAtlasGetSingularChartWithBoundary(MFAtlas,MFNRegion);

void MFExtendAtlas(MFAtlas,MFContinuationMethod,MFImplicitMF,MFNRegion,MFNVector);
void MFExtendAtlasMultiple(MFAtlas,MFContinuationMethod,MFImplicitMF,MFNRegion,int,MFNVector*);
void MFExtendAtlasWithTangent(MFAtlas,MFContinuationMethod,MFImplicitMF,MFNRegion,MFNVector,MFNKMatrix);
void MFExtendAtlasMultipleWithTangents(MFAtlas,MFContinuationMethod,MFImplicitMF,MFNRegion,int,MFNVector*,MFNKMatrix*);
void MFCloseAtlas(MFContinuationMethod,MFAtlas);
void MFFlushAtlas(MFContinuationMethod,MFAtlas);

int MFAtlasIsChartSingular(MFAtlas,int);
double MFAtlasChartSuggestedRadius(MFAtlas,int);

void MFAtlasAddClipF(MFAtlas,double (*)(MFNVector));
void MFAtlasClearClipF(MFAtlas);

void MFAtlasSetExpFactor(MFAtlas,double);
double MFAtlasGetExpFactor(MFAtlas);

void MFAtlasClosePlotfile(MFAtlas);
void MFAtlasCloseCenterfile(MFAtlas);

#ifdef __cplusplus
 }
#endif
#endif
