/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFPrint.h>
#include <math.h>

void findSolution(double,double,double,double,double,double,double*,double*);
int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j,n;
  int n0;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  MFNVector ll,ur;
  MFContinuationMethod H;

/* Complex Cusp Catastrophe */

/*   u*(u*u-r)-s = 0  */
/*   (x+i*y)*((x+i*y)*(x+i*y)-r)-s = 0  */
/*   (x+i*y)*((x*x-y*y-r)+i*2*x*y)-s = 0  */
/*   x*((x*x-y*y-r)+i*2*x*y)+i*y*((x*x-y*y-r)+i*2*x*y)-s = 0  */

/*   x*(x*x-y*y-r)-2*x*y*y - s = 0 */
/*   2*x*x*y+y*(x*x-y*y-r) = 0  */

  M=MFIMFCreateAlgebraicExpressionWithRadius("[x,y,r,s]","[x*(x*x-y*y-r)-2*x*y*y-s,2*x*x*y+y*(x*x-2*y*y-r)]",.05);

  n=MFIMF_N(M);
  ll=MFCreateNVector(n);
  ur=MFCreateNVector(n);
  MFNVSetC(ll,0,-5.);MFNVSetC(ur,0,5.);
  MFNVSetC(ll,1,-5.);MFNVSetC(ur,1,5.);
  MFNVSetC(ll,2,-1.);MFNVSetC(ur,2,1.);
  MFNVSetC(ll,3,-1.); MFNVSetC(ur,3,1.); 
  Omega=MFNRegionCreateHyperCubeByCorners(n,ll,ur);
  MFFreeNVector(ll);
  MFFreeNVector(ur);
  
  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetDotMin(H,.9);
  MFHendersonsMethodSetMaxCharts(H,100000);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,0);
  MFHendersonsMethodSetBranchSwitch(H,200);
  MFHendersonsMethodSetDumpToPlotFile(H,1);   /* Write polyhedra to a plotfile */
  MFHendersonsMethodSetDumpToCenterFile(H,0); /* Write points to a file */
  MFHendersonsMethodSetDumpToRestartFile(H,0);

  u0=MFCreateNVector(n);

/* Trivial */

  MFNVSetC(u0,0,-.4);
  MFNVSetC(u0,1,0.);
  MFNVSetC(u0,2,0.);
  MFNVSetC(u0,3,-.4*.4*.4);
  MFHendersonsMethodSetFilename(H,"Cusp");

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
