/* 
    @(#)ComputeExpression.c	1.7
    03/02/25 10:33:35
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFContinuationMethod H;

  M=MFIMFCreateAlgebraicExpression("[x,y,z]","[(100*x)**2+y**2+z**2 -1]");
/*M=MFIMFCreateAlgebraicExpression("[x,y,z]","[x**2+y**2-1+.1*sin(6*3.1415926*x),z-x*y]");*/
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,1.1);

  u0=MFCreateNVector(n);
  MFNVSetC(u0,0, 0.);
  MFNVSetC(u0,1, 1.);
  MFNVSetC(u0,2, 0.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.001);
  MFHendersonsMethodSetMaxCharts(H,3000);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,0);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"Expression");

  S=MFComputeAtlas(H,M,Omega,u0);

/* Use this if you don't set dump to plotfile */
/*
  fid=fopen("Expression.atlas","w");
  MFWriteAtlas(fid,S);
  fclose(fid);
  printf("Done writing Atlas\n");fflush(stdout);
*/

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
