/* 
    @(#)ComputeForcedOscillator.c	1.5
    02/04/19 14:48:25
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

#define MFPI 3.14159265358979323846264338327950288

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFNVector ug;
  MFNVector t[2];
  MFNKMatrix Tan;
  double tau;

  int nt=600;

  double p=1.;
  double q=4.;
  double e,B;
  MFContinuationMethod H;

  M=MFIMFCreateForcedOscillator(nt,q);
  n=nt+2;

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,1.);         /* Max distance from TS to M */
  MFHendersonsMethodSetMaxCharts(H,-1);       /* -1 means infinite */
  MFHendersonsMethodSetVerbose(H,1);          /* Write info to stdout */
  MFHendersonsMethodSetPage(H,1);             /* Page out non-interior polyhedra */
  MFHendersonsMethodSetDumpToPlotFile(H,1);   /* Write polyhedra to a plotfile */
  MFHendersonsMethodSetDumpToCenterFile(H,0); /* Write points to a file */

  ug=MFCreateNVector(n);

#define FOURONE
#ifdef FOURONE
  MFHendersonsMethodSetFilename(H,"ForcedOscillatorFourOneA");
  p=1.;
  q=4.;
  e=11.4;
  B=.05*20;
#endif
#ifdef THREEEONE
  MFHendersonsMethodSetFilename(H,"ForcedOscillatorThreeOne");
  p=1.;
  q=3.;
  e=7.5;
  B=.04*20;
#endif

  for(i=0;i<nt;i++)
   {
    tau=q*2*MFPI*i/nt;
    MFNVSetC(ug,i,2*sin(p*tau/q));
   }
  MFNVSetC(ug,nt+0,e);
  MFNVSetC(ug,nt+1,B);
  ForcedWrite(ug);

  t[0]=MFCreateNVector(n);
  for(i=0;i<nt;i++)
    MFNVSetC(t[0],i, 0.);
  MFNVSetC(t[0],nt+0,1.);
  MFNVSetC(t[0],nt+1,0.);

  t[1]=MFCreateNVector(n);
  for(i=0;i<nt;i++)
    MFNVSetC(t[1],i,0.);
  MFNVSetC(t[1],nt+0,0.);
  MFNVSetC(t[1],nt+1,1.);

  Tan=MFCreateNKMatrix(2,t);
  u0=MFCreateNVector(n);
  if(!MFIMFProject(M,ug,Tan,u0))
   {
    printf("Initial project failed\n");return 8;
   }
  MFFreeNVector(ug);
  MFFreeNKMatrix(Tan);

/*Omega=MFNRegionCreateForcedOscillator(n,q,0.,20.,0.,.5*20);*/
  Omega=MFNRegionCreateForcedOscillator(n,q,0.,30.,0.,.8*20);

  S=MFComputeAtlas(H,M,Omega,u0);
  MFCloseAtlas(H,S);
  printf("Done writing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);
  ForcedCloser();

  return(0);
 }
