/* 

    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0,u;
  MFNKMatrix Phi;
  FILE *fid;
  MFContinuationMethod H;

  M=MFIMFCreateAlgebraicExpression("[x,y,z]",
  "[((sqrt((x-.8)**2+y**2)-.8)**2+z**2 -.5**2)*((sqrt((x+.8)**2+y**2)-.8)**2+z**2 -.5**2)-.05]");
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,2.4);

/* This bit starts with an approximate point on M, and calls project to get a point on M */

  u=MFCreateNVector(n);
  MFNVSetC(u,0, 1.2);
  MFNVSetC(u,1, 0.);
  MFNVSetC(u,2, 0.);
  u0=MFCreateNVector(n);
  Phi=MFCreateNKMatrixWithData(3,2,(double*)NULL);

  MFNKMSetC(Phi,0,0,0.);
  MFNKMSetC(Phi,1,0,1.);
  MFNKMSetC(Phi,2,0,0.);
  MFNKMSetC(Phi,0,1,0.);
  MFNKMSetC(Phi,1,1,0.);
  MFNKMSetC(Phi,2,1,1.);

  if(!MFIMFProject(M,u,Phi,u0))return 8;

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,0);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"GenusTwo");

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
