/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>

double XGreaterThanMinusOne(MFNVector u){return -MFNV_C(u,0)-1.;}
double XLessThanOne(MFNVector u){return MFNV_C(u,0)-1.;}
double YGreaterThanMinusOne(MFNVector u){return -MFNV_C(u,1)-1.;}
double YLessThanOne(MFNVector u){return MFNV_C(u,1)-1.;}

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFContinuationMethod H;

  M=MFIMFCreatePlane();
  n=MFIMF_N(M);
  Omega=MFNRegionCreateRectangle(-1.,-1.,1.,1.);

  u0=MFCreateNVector(n);
  MFNVSetC(u0,0, 0.);
  MFNVSetC(u0,1, 0.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.1);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"PlaneClip");

  MFHendersonsMethodAddClipF(H,XGreaterThanMinusOne);
  MFHendersonsMethodAddClipF(H,XLessThanOne);
  MFHendersonsMethodAddClipF(H,YGreaterThanMinusOne);
  MFHendersonsMethodAddClipF(H,YLessThanOne);

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
