/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFDX.h>
#include <MFFortran.h>
#include <math.h>
int MFRodProjectToDraw(MFNVector,double*,void*);

double MzPositive(MFNVector);
double MzNegative(MFNVector);
double MzSmallPositive(MFNVector);
double MzSmallNegative(MFNVector);
double Mx0Positive(MFNVector);
double Mx0Negative(MFNVector);
double FPositive(MFNVector);
double DNegative(MFNVector);
double DPositive(MFNVector);
double FLessThanSixteen(MFNVector);
double DSmallNegative(MFNVector);
double DSmallPositive(MFNVector);
double DSmallerNegative(MFNVector);
double DSmallerPositive(MFNVector);

double K(double);
double Ki(double);

double MFTPBVPTestTangent(MFImplicitMF,MFNVector,MFNKMatrix);
MFNKMatrix ComputeTangent(MFImplicitMF,int,int,int,int,MFNVector);
int MFTPBVPAnalyze(MFImplicitMF,MFNVector);
void MFTPBVPNSpaceSetPeriodicParameter(MFNSpace,int,double);
int MFRodStopN(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*);
int MFRodStopD(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*);
int MFRodProjectToDraw(MFNVector,double*,void*);
MFNVector GetInitialPoint(MFImplicitMF,double,double,double,double,double);
MFNVector GetInitialPointMx0Mz(MFImplicitMF,double,double,double,double);
MFNVector GetInitialPointMx0f(MFImplicitMF,double,double,double,double);
MFNVector GetInitialPointI(MFImplicitMF,double,double,double,double);
void flipf(int,int,MFNVector);
void flipMz(int,int,MFNVector);
void flipMx0(int,int,MFNVector);
void flipMx0AndMz(int,int,MFNVector);
void moveToPrinciple(int,int,MFNVector);
void moveToPositive(int,int,MFNVector);

struct MFTPBVPData
 {
  int nx;
  int nu;
  int np;
  int nbc;
  int nic;
  int k;
  MFTPBVPFFUNCTION f;
  MFTPBVPFFUNCTION fu;
  MFTPBVPFFUNCTION fl;
  MFTPBVPAFUNCTION a;
  MFTPBVPAFUNCTION au;
  MFTPBVPAFUNCTION al;
  MFTPBVPLFUNCTION l;
  MFTPBVPLFUNCTION lu;
  MFTPBVPLFUNCTION ll;
  MFTPBVPMFUNCTION m;
  MFTPBVPMFUNCTION ml;
  MFNSpace space;
 };

#define PI 3.14159265358979323846264338327950288

/*
u'=f(u,t,l);                     nu differential equations
a(u(0),u(1),l)=0;                nbc=nu equations
int l(u(t),t,l) dt + m(l) = 0;   nc equations
*/

int GetNullVector(int,int,int,int,double*,double*,double*);

/* mz  a - p[0] */
/* mx0 c - p[1] */
/* f   t - p[2] */
/*     theta0 - p[3] */

void f(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *f)
 {
  f[0]=PI*u[1];
  f[1]=PI*(p[2]*u[0]*u[5]-p[0]*u[3]);
  f[2]=PI*u[3];
  f[3]=PI*(p[2]*u[2]*u[5]-p[1]*u[5]+p[0]*u[1]);
  f[4]=PI*u[5];
  f[5]=PI*(-p[2]*u[2]*u[3]-p[2]*u[0]*u[1]+p[1]*u[3]);

  return;
 }

void fu(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *fu)
 {
  int i;

  for(i=0;i<nu*nu;i++)fu[i]=0.;

  fu[0+nu*1]=PI;

  fu[1+nu*0]=PI*p[2]*u[5];
  fu[1+nu*3]=-PI*p[0];
  fu[1+nu*5]=PI*p[2]*u[0];

  fu[2+nu*3]=PI;

  fu[3+nu*1]=PI*p[0];
  fu[3+nu*2]=PI*p[2]*u[5];
  fu[3+nu*5]=PI*(p[2]*u[2]-p[1]);

  fu[4+nu*5]=PI;

  fu[5+nu*0]=-PI*p[2]*u[1];
  fu[5+nu*1]=-PI*p[2]*u[0];
  fu[5+nu*2]=-PI*p[2]*u[3];
  fu[5+nu*3]=PI*(-p[2]*u[2]+p[1]);

  return;
 }

void fl(double r, int nu, double *u, int np, double *p, double *u0, double *l0, double *fl)
 {
  int i;

  for(i=0;i<nu*np;i++)fl[i]=0.;

  fl[1+nu*0]=-PI*u[3];
  fl[1+nu*2]=PI*u[0]*u[5];
  fl[3+nu*0]=PI*u[1];
  fl[3+nu*1]=-PI*u[5];
  fl[3+nu*2]=PI*u[2]*u[5];
  fl[5+nu*1]=PI*u[3];
  fl[5+nu*2]=-PI*(u[2]*u[3]+u[0]*u[1]);

  return;
 }

void a(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *a)
 {
  a[0]=uL[0];
  a[1]=uL[1]-sin(p[3]);
  a[2]=uL[2];
  a[3]=uL[3];
  a[4]=uL[4];
  a[5]=uL[5]-cos(p[3]);
  a[6]=uR[3];
  a[7]=uR[0]*uR[5]-uR[4]*uR[1];

  return;
 }

void au(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *au)
 {
  int i;

  for(i=0;i<nbc*2*nu;i++)au[i]=0.;

  au[0+nbc*0]=1.;
  au[1+nbc*1]=1.;
  au[2+nbc*2]=1.;
  au[3+nbc*3]=1.;
  au[4+nbc*4]=1.;
  au[5+nbc*5]=1.;
  au[6+nbc*(6+3)]=1.;
  au[7+nbc*(6+0)]=uR[5];
  au[7+nbc*(6+1)]=-uR[4];
  au[7+nbc*(6+4)]=-uR[1];
  au[7+nbc*(6+5)]=uR[0];

  return;
 }

void al(int nbc, int nu, double *uL, double *uR, int np, double *p, double *u0L, double *u0R,double *l0, double *al)
 {
  int i;

  for(i=0;i<nbc*np;i++)al[i]=0.;

  al[1+nbc*3]=-cos(p[3]);
  al[5+nbc*3]= sin(p[3]);

  return;
 }

void l(int nic, double r, int nu, double *u, int np, double *p,double *u0,double *l0, double *l)
 {
  int i;

  for(i=0;i<nic;i++)l[i]=0.;

  if(nic>0)
    l[0]=u[1]*u[1]+u[3]*u[3]+u[5]*u[5];
  if(nic>1)
    l[1]=u[0]*u[0]+u[2]*(u[2]-p[1])+u[5];

  return;
 }

void lu(int nic, double r, int nu, double *u, int np, double *p,double *u0,double *l0, double *lu)
 {
  int i;

  for(i=0;i<nu*nic;i++)lu[i]=0.;

  if(nic>0)
   {
    lu[0+nic*1]=2*u[1];
    lu[0+nic*3]=2*u[3];
    lu[0+nic*5]=2*u[5];
   }
  if(nic>1)
   {
    lu[1+nic*0]=2*u[0];
    lu[1+nic*2]=2*u[2]-p[1];
    lu[1+nic*5]=1.;
   }

  return;
 }

void ll(int nic, double r, int nu, double *u, int np, double *p,double *u0,double *l0, double *ll)
 {
  int i;

  for(i=0;i<np*nic;i++)ll[i]=0.;
  if(nic>0)ll[1+nic*8]=-u[2];

  return;
 }

void m(int nic, int np, double *p, double *p0, double *m)
 {
  int i;

  for(i=0;i<nic;i++)m[i]=0.;
  if(nic>0)m[0]=-pow(p[2],4);
  if(nic>1)m[1]=-p[3];

  return;
 }

void ml(int nic, int np, double *p, double *p0, double *ml)
 {
  int i;

  for(i=0;i<np*nic;i++)ml[i]=0.;

  if(nic>0)ml[0+nic*2]=-4.*pow(p[2],3);
  if(nic>1)ml[1+nic*3]=-1.;

  return;
 }

#define U0I {u0[n0]=GetInitialPointI(M,Mx0,Mz,F,theta0);if(u0[n0]!=(MFNVector)NULL){printf("u0[%d] is (%lf,%lf,%lf,%lf)\n",n0,MFNV_C(u0[n0],nx*nu+0),MFNV_C(u0[n0],nx*nu+1),MFNV_C(u0[n0],nx*nu+2),MFNV_C(u0[n0],nx*nu+3));n0++;}}

#define U0Mx0Mz {u0[n0]=GetInitialPointMx0Mz(M,Mx0,Mz,F,theta0);if(u0[n0]!=(MFNVector)NULL){printf("u0[%d] is (%lf,%lf,%lf,%lf)\n",n0,MFNV_C(u0[n0],nx*nu+0),MFNV_C(u0[n0],nx*nu+1),MFNV_C(u0[n0],nx*nu+2),MFNV_C(u0[n0],nx*nu+3));n0++;}}

#define FLIPMz {u0[n0]=MFCloneNVector(u0[n0-1]);flipMz(nx,nu,u0[n0]);n0++;}
#define FLIPMx0 {u0[n0]=MFCloneNVector(u0[n0-1]);flipMx0(nx,nu,u0[n0]);n0++;}

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j,n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0[500];
  int n0;
  FILE *fid;
  MFNVector ug;
  MFNKMatrix Tan;

  double C,A,t;
  double Mx0,Mz,F,theta0;
  double r,dr,x1,y1,x2,y2,x3,y3;
  double xy0[6],p0[4],p1[4];
  double error,e0,e1,corr;
  double dA,dC,dy1,dy3,a00,a01,a10,a11,de0,de1;
  int itimes;
  MFNVector tug;
  double eps;

  int nx=150;
  int np=4;
  int nu=6;
  int nbc=8;
  int nic=0;
  int k;
  MFContinuationMethod H;
  char name[80];

/* ( a,-c, t, th) <-> ( a, c, t,2PI-th) */
/* (-a, c, t, th) <-> ( a, c, t,2PI-th) */
/* ( a, c,-t, th) <-> ( a, c, t,th-PI) (x,y,z,dx,dy,dz)-> (-x,-y,-z,-dx,-dy,-dz) */

  k=nu+np-nbc-nic;
  
  M=MFIMFCreateTPBVP(k,nx,nu,np,f,fu,fl,nbc,a,au,al,nic,l,lu,ll,m,ml);
  MFTPBVPNSpaceSetPeriodicParameter(MFIMFNSpace(M),3,2*PI);
  MFIMFSetStop(M,MFRodStopN);
  MFIMFSetR(M,.3);
  MFIMFSetProjectForDraw(M,MFRodProjectToDraw);
  n=MFIMF_N(M);
  printf("TPBVP calculated k to be %d\n",MFIMF_K(M));

  p0[0]= -0.05;   /* Mz .1  */
  p0[1]= -10.5;   /* Mx0.1  */
  p0[2]= -.1;   /* f  */
  p0[3]= -1*PI;  /* th */

  p1[0]= 10.5;  /* a  Mz */
  p1[1]=  0.05;   /* c  Mx0 */
  p1[2]= 16.;   /* 16 f  */
  p1[3]= 1*PI;  /* th */
  Omega=MFNRegionCreateTPBVP(nx,nu,np,p0,p1,-200.,200.);

/* 2- Mz<0, Mx0<0 */
  n0=0;
  Mx0=-1.5483;Mz= 1.9552;F=4.1227;theta0=-1.4082;U0Mx0Mz
  Mx0=-0.2279;Mz= 1.2260;F=4.5294;theta0=-0.9955;U0Mx0Mz
  printf("Number of converged initial points = %d\n",n0);fflush(stdout);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.1*60); /* This looks strange, but it's because of the 6 eqs, on intervals of length PI */
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetPageEvery(H,100);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetBranchSwitch(H,0);
  MFHendersonsMethodSetFilename(H,"Rod");

  MFHendersonsMethodAddClipF(H,MzPositive);
  MFHendersonsMethodAddClipF(H,Mx0Negative);
  MFHendersonsMethodAddClipF(H,FPositive);
  MFHendersonsMethodAddClipF(H,DNegative);
  MFHendersonsMethodAddClipF(H,FLessThanSixteen);

  S=MFComputeAtlasMultiple(H,M,Omega,n0,u0);
  MFFlushAtlas(H,S);
  MFFreeNRegion(Omega);
  for(i=0;i<n0;i++)MFFreeNVector(u0[i]);
  p0[0]= -10.5;   /* Mz .1  */
  p0[1]=  0.05;   /* Mx0.1  */
  p0[2]= -.1;   /* f  */
  p0[2]=  .05;   /* f  */
  p0[3]= -4*PI;  /* th */

  p1[0]= -.05;  /* a  Mz */
  p1[1]= 10.5;   /* c  Mx0 */
  p1[2]= 16.;   /* 16 f  */
  p1[3]= 4*PI;  /* th */
  Omega=MFNRegionCreateTPBVP(nx,nu,np,p0,p1,-200.,200.);

/* 2- Mz<0, Mx0>0 */
  n0=0;
  Mx0=1.4595;Mz=-0.2071;F=3.5751;theta0=-1.0814;U0Mx0Mz
  Mx0=1.3491;Mz=-0.0881;F=3.3506;theta0=-1.1738;U0Mx0Mz
/* 4- Mz<0, Mx0>0 */
  Mx0=1.4294;Mz=-0.2155;F=3.9047;theta0=-0.9269;U0Mx0Mz
  printf("Number of converged initial points = %d\n",n0);fflush(stdout);

  MFHendersonsMethodClearClipF(H);
  MFHendersonsMethodAddClipF(H,MzNegative);
  MFHendersonsMethodAddClipF(H,Mx0Positive);
  MFHendersonsMethodAddClipF(H,DNegative);
  MFHendersonsMethodAddClipF(H,FPositive);
  MFHendersonsMethodAddClipF(H,FLessThanSixteen);

  MFExtendAtlasMultiple(S,H,M,Omega,n0,u0);
  MFFlushAtlas(H,S);
  MFFreeNRegion(Omega);
  for(i=0;i<n0;i++)MFFreeNVector(u0[i]);

  p0[0]= -.15;   /* Mz .1  */
  p0[1]= -0.01;   /* Mx0.1  */
  p0[2]= -.15;   /* f  */
  p0[2]=  .05;   /* f  */
  p0[3]= -4*PI;  /* th */

  p1[0]=  .15;  /* a  Mz */
  p1[1]= 10.5;   /* c  Mx0 */
  p1[2]= 16.;   /* 16 f  */
  p1[3]= 4*PI;  /* th */
  Omega=MFNRegionCreateTPBVP(nx,nu,np,p0,p1,-200.,200.);
/* 2- Mz=0, Mx0>0 */
  n0=0;
  Mx0=1.6306;Mz=-0.0109;F=1.8838;theta0=-0.4894;U0Mx0Mz

  printf("Number of converged initial points = %d\n",n0);fflush(stdout);

  MFHendersonsMethodClearClipF(H);
  MFHendersonsMethodAddClipF(H,MzSmallPositive);
  MFHendersonsMethodAddClipF(H,MzSmallNegative);
  MFHendersonsMethodAddClipF(H,Mx0Positive);
  MFHendersonsMethodAddClipF(H,FPositive);
  MFHendersonsMethodAddClipF(H,DNegative);
  MFHendersonsMethodAddClipF(H,FLessThanSixteen);

  MFExtendAtlasMultiple(S,H,M,Omega,n0,u0);
  MFFlushAtlas(H,S);
  MFFreeNRegion(Omega);
  for(i=0;i<n0;i++)MFFreeNVector(u0[i]);

  p0[0]= -10.5;   /* Mz .1  */
  p0[1]= -0.05;   /* Mx0.1  */
  p0[2]= -.1;   /* f  */
  p0[2]=  .05;   /* f  */
  p0[3]= -4*PI;  /* th */

  p1[0]= -.05;  /* a  Mz */
  p1[1]= 10.5;   /* c  Mx0 */
  p1[2]= 16.;   /* 16 f  */
  p1[3]= 4*PI;  /* th */
  Omega=MFNRegionCreateTPBVP(nx,nu,np,p0,p1,-200.,200.);

/* 2- Mz<0, Mx0>0 */
  n0=0;
  Mx0=-1.5436;Mz= 2.1364;F= 0.5304;theta0=-.3170;U0Mx0Mz
  flipMx0AndMz(nx,nu,u0[n0-1]);

  printf("Number of converged initial points = %d\n",n0);fflush(stdout);

  MFHendersonsMethodClearClipF(H);
  MFHendersonsMethodAddClipF(H,MzNegative);
  MFHendersonsMethodAddClipF(H,Mx0Positive);
  MFHendersonsMethodAddClipF(H,DNegative);
  MFHendersonsMethodAddClipF(H,FPositive);
  MFHendersonsMethodAddClipF(H,FLessThanSixteen);

  MFExtendAtlasMultiple(S,H,M,Omega,n0,u0);
  MFFlushAtlas(H,S);
  MFFreeNRegion(Omega);
  for(i=0;i<n0;i++)MFFreeNVector(u0[i]);
  p0[0]=  0.05;   /* Mz .1  */
  p0[1]= -0.1;   /* Mx0.1  */
  p0[2]= -.1;   /* f  */
  p0[2]=  .05;   /* f  */
  p0[3]= -2*PI;  /* th */

  p1[0]=  10.5;   /* a  Mz */
  p1[1]=  10.5;   /* c  Mx0 */
  p1[2]= 16.;   /* 16 f  */
  p1[3]= 2*PI;  /* th */
  Omega=MFNRegionCreateTPBVP(nx,nu,np,p0,p1,-200.,200.);

/* 2- Buckled Ring */
  n0=0;

/* 4-  */
  Mx0= 0.5276;Mz= 0.8750;F= 3.1051;theta0=-1.3618;U0Mx0Mz
/* 4-  */
  Mx0= 0.2372;Mz= 0.9332;F= 4.0547;theta0=-1.0424;U0Mx0Mz
/* 4-  (new this trip)*/
  Mx0= 0.7051;Mz= 0.5268;F= 3.7947;theta0=-1.0477;U0Mx0Mz 
/* 2-  */
  Mx0= 0.8811;Mz= 0.3322;F= 2.7780;theta0=-1.6469;U0I

  printf("Number of converged initial points %d\n",n0);fflush(stdout);

  MFHendersonsMethodClearClipF(H);
  MFHendersonsMethodAddClipF(H,MzPositive);
  MFHendersonsMethodAddClipF(H,Mx0Positive);
  MFHendersonsMethodAddClipF(H,FPositive);
  MFHendersonsMethodAddClipF(H,DNegative);
  MFHendersonsMethodAddClipF(H,FLessThanSixteen);

  MFExtendAtlasMultiple(S,H,M,Omega,n0,u0);

  printf("Done Computing, now trying to write it\n");fflush(stdout);
  MFCloseAtlas(H,S);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeHendersonsMethod(H);
  MFFreeNRegion(Omega);
  for(i=0;i<n0;i++)MFFreeNVector(u0[i]);

  printf("Normal termination of ComputeRod\n");fflush(stdout);
  return(0);
 }

void MFTPBVPEvaluateIntegralConstraints(MFImplicitMF,MFNVector,MFNVector,double*);
void MFTPBVPEvaluateBoundaryConditions(MFImplicitMF,MFNVector,MFNVector,double*);

MFNKMatrix ComputeTangent(MFImplicitMF M, int nx, int nu, int np, int nic, MFNVector ug)
 {
  MFNVector dug[4];
  MFNVector tug;
  MFNKMatrix Tan0;
  static double *u0=(double*)NULL;
  static double *p0=(double*)NULL;
  static double *TMat=(double*)NULL;
  static double *bc=(double*)NULL;
  static double *dbc=(double*)NULL;
  static double *ic=(double*)NULL;
  static double *dic=(double*)NULL;
  int i,j,n;
  double eps;

  u0=(double*)realloc((void*)u0,nu*sizeof(double));
  p0=(double*)realloc((void*)p0,np*sizeof(double));
  TMat=(double*)realloc((void*)TMat,(nu+np)*(nu+np)*sizeof(double));
  bc=(double*)realloc((void*)bc,nu*sizeof(double));
  dbc=(double*)realloc((void*)dbc,nu*sizeof(double));
  ic=(double*)realloc((void*)ic,nic*sizeof(double));
  dic=(double*)realloc((void*)dic,nic*sizeof(double));

  printf("In ComputeTangent\n");

  for(j=0;j<nu;j++)u0[j]=0.;
  for(j=0;j<np;j++)p0[j]=0.;
  eps=1.e-3;

  tug=MFTPBVPIntegrateForTangent(M,ug,u0,p0);
  MFTPBVPEvaluateBoundaryConditions(M,tug,tug,bc);
  MFTPBVPEvaluateIntegralConstraints(M,tug,tug,ic);
  MFFreeNVector(tug);

  for(i=0;i<nu+np;i++)
   {
    for(j=0;j<nu;j++)u0[j]=0.;
    for(j=0;j<np;j++)p0[j]=0.;

    if(i<nu)u0[i]=eps;
      else p0[i-nu]=eps;

    tug=MFTPBVPIntegrateForTangent(M,ug,u0,p0);
    MFTPBVPEvaluateBoundaryConditions(M,tug,tug,dbc);
    MFTPBVPEvaluateIntegralConstraints(M,tug,tug,dic);
    MFFreeNVector(tug);

    for(j=0;j<nu;j++)
      TMat[j+(nu+np)*i]=(dbc[j]-bc[j])/eps;
    for(j=0;j<nic;j++)
      TMat[nu+j+(nu+np)*i]=(dic[j]-ic[j])/eps;
    for(j=0;j<np-nic;j++)
      TMat[nu+nic+j+(nu+np)*i]=0.;

   }

  for(i=0;i<nu+np;i++)
   {
    if(i==0)printf("T=[%10.3le",TMat[i+(nu+np)*0]);
     else printf("  [%10.3le",TMat[i+(nu+np)*0]);
    for(j=1;j<nu+np;j++)
      printf(" %10.3le",TMat[i+(nu+np)*j]);
    printf("]\n");
   }

/* for each null vector of a */
  printf("Find tangent\n");
  n=1;
  for(j=0;j<n;j++)
   {
    n=GetNullVector(j,nu,np,j,TMat,u0,p0);
    if(j==0){printf("There were %d zero eigenvalues\n",n);fflush(stdout);}
    printf("%d[",j);
    for(i=0;i<nu;i++)
     printf(" %10.3le",u0[i]);
    for(i=0;i<np;i++)
     printf(" %10.3le",p0[i]);
    printf("]\n");
 
    dug[j]=MFTPBVPIntegrateForTangent(M,ug,u0,p0);
   }
  Tan0=MFCreateNKMatrix(MFIMF_K(M),dug);
/*MFGramSchmidt(Tan0);*/
/*printf("error in Phi %le\n",MFTPBVPTestTangent(M,ug,Tan0));fflush(stdout);*/

  return Tan0;
 }

int GetNullVector(int time, int nu, int np, int ev,double *A,double *x,double *p)
 {
  static int i,j,l;
  static double mag;
  static int info;
  static char jobvl;
  static char jobvr;
  static int ldA;
  static double *wr=(double*)NULL;
  static double *wi=(double*)NULL;
  static int ldvl;
  static double *vl=(double*)NULL;
  static int ldvr;
  static double *vr=(double*)NULL;
  static double *work=(double*)NULL;
  static int lwork;

  ldA=nu+np;
  wr=(double*)realloc((void*)wr,ldA*sizeof(double));
  wi=(double*)realloc((void*)wi,ldA*sizeof(double));
  ldvl=ldA;
  vl=(double*)realloc((void*)vl,ldvl*ldvl*sizeof(double));
  ldvr=ldA;
  vr=(double*)realloc((void*)vr,ldvr*ldvr*sizeof(double));
  lwork=4*ldA;
  work=(double*)realloc((void*)work,lwork*sizeof(double));
  info=0;
  jobvl='N';
  jobvr='V';
  if(time==0)
    F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&ldA,A,&ldA,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);

  j=0;
  for(i=0;i<nu+np;i++)
   {
    mag=sqrt(wr[i]*wr[i]+wi[i]*wi[i]);
    if(mag<1.e-10)
     {
      if(j==ev)
       {
        for(l=0;l<nu;l++)x[l]=vr[l+(nu+np)*i];
        for(l=0;l<np;l++)p[l]=vr[nu+l+(nu+np)*i];
       }
      j++;
     }
   }

  return j;
 }

MFNVector GetInitialPoint(MFImplicitMF M,double C,double A, double t, double theta0,double dA)
 {
  MFNVector u0,ug,ut[2];
  MFNKMatrix Tan;
  double *r0,dr;
  int i,n,nx,nu,np;
  int rc;

  printf("GetInitialPoint\n");fflush(stdout);

  n=MFIMF_N(M);
  nx=MFTPBVPGetNX(M);
  nu=MFTPBVPGetNU(M);
  np=MFTPBVPGetNP(M);
  r0=(double*)malloc(nx*sizeof(double));

  dr=1./(nx-2);
  for(i=0;i<nx;i++)r0[i]=(i-.5)*dr;

/* Assumes theta0= 0 or PI, C=0 */

  ug=MFCreateNVector(n);
  ut[0]=MFCreateNVector(n);
  ut[1]=MFCreateNVector(n);
  for(i=0;i<nx;i++)
   {
    MFNVSetC(ug,nu*i+0,0.);
    MFNVSetC(ug,nu*i+1,0.);
    MFNVSetC(ug,nu*i+2,0.);
    MFNVSetC(ug,nu*i+3,0.);
    MFNVSetC(ug,nu*i+4,cos(theta0)*r0[i]*PI);
    MFNVSetC(ug,nu*i+5,cos(theta0));

    MFNVSetC(ug,nu*nx+np+i,r0[i]);

    MFNVSetC(ut[0],nu*i+0,0.);
    MFNVSetC(ut[0],nu*i+1,0.);
    MFNVSetC(ut[0],nu*i+2,0.);
    MFNVSetC(ut[0],nu*i+3,0.);
    MFNVSetC(ut[0],nu*i+4,0.);
    MFNVSetC(ut[0],nu*i+5,0.);

    MFNVSetC(ut[0],nu*nx+np+i,r0[i]);

    MFNVSetC(ut[1],nu*i+0,0.);
    MFNVSetC(ut[1],nu*i+1,0.);
    MFNVSetC(ut[1],nu*i+2,0.);
    MFNVSetC(ut[1],nu*i+3,0.);
    MFNVSetC(ut[1],nu*i+4,0.);
    MFNVSetC(ut[1],nu*i+5,0.);
    MFNVSetC(ut[1],nu*nx+np+i,r0[i]);
   }
  MFNVSetC(ug,nu*nx+0,A);
  MFNVSetC(ug,nu*nx+1,C);
  MFNVSetC(ug,nu*nx+2,t);
  MFNVSetC(ug,nu*nx+3,theta0);

  MFNVSetC(ut[0],nu*nx+0,0.);
  MFNVSetC(ut[0],nu*nx+1,dA);
  MFNVSetC(ut[0],nu*nx+2,1.);
  MFNVSetC(ut[0],nu*nx+3,0.);

  MFNVSetC(ut[1],nu*nx+0,0.);
  MFNVSetC(ut[1],nu*nx+1,1.);
  MFNVSetC(ut[1],nu*nx+2,0.);
  MFNVSetC(ut[1],nu*nx+3,0.);
/* Try */
  MFNVSetC(ug,nu*nx+0,A+.1);
  MFNVSetC(ut[0],nu*nx+0,0.);
  MFNVSetC(ut[0],nu*nx+1,0.);
  MFNVSetC(ut[0],nu*nx+2,1.);
  MFNVSetC(ut[0],nu*nx+3,0.);
  MFNVSetC(ut[1],nu*nx+0,1.);
  MFNVSetC(ut[1],nu*nx+1,0.);
  MFNVSetC(ut[1],nu*nx+2,0.);
  MFNVSetC(ut[1],nu*nx+3,0.);
/* end Try */
  Tan=MFCreateNKMatrix(2,ut);
  Tan=MFIMFTangentSpace(M,ug);
  printf("ug=(%le,%le,%le,%le)\n",MFNV_C(ug,nu*nx+0),MFNV_C(ug,nu*nx+1),MFNV_C(ug,nu*nx+2),MFNV_C(ug,nu*nx+3));fflush(stdout);
  MFFreeNVector(ut[0]);
  MFFreeNVector(ut[1]);
  MFFreeNKMatrix(Tan);
  free(r0);
  return ug;

  u0=MFCreateNVector(n);
  rc=MFIMFProject(M,ug,Tan,u0);
  printf("u0=(%le,%le,%le,%le)\n",MFNV_C(u0,nu*nx+0),MFNV_C(u0,nu*nx+1),MFNV_C(u0,nu*nx+2),MFNV_C(u0,nu*nx+3));fflush(stdout);

  MFFreeNVector(ug);

  MFFreeNVector(ut[0]);
  MFFreeNVector(ut[1]);
  MFFreeNKMatrix(Tan);
  free(r0);
  if(rc){printf("Project Successful\n");fflush(stdout);}
  if(rc)return u0;
  printf("Project Failed\n");fflush(stdout);
  MFFreeNVector(u0);
  return (MFNVector)NULL;
 }

MFNVector GetInitialPointMx0Mz(MFImplicitMF M,double Mx0,double Mz, double f, double theta0)
 {
  double r,dr;
  double xy0[6],p0[4],p1[4];
  double *r0;
  double error,e0,e1,corr;
  double dMz,dMx0,dy1,dy3,a00,a01,a10,a11,de0,de1;
  int itimes;
  MFNVector ug,tug;
  MFNVector u0;
  MFNKMatrix Tan;
  int i,n,nx,nu,np;
  double xnorm;
  int verbose=0;

  if(verbose){printf("GetInitialPointMx0Mz\n");fflush(stdout);}

  n=MFIMF_N(M);
  nx=MFTPBVPGetNX(M);
  nu=MFTPBVPGetNU(M);
  np=MFTPBVPGetNP(M);

  r0=(double*)malloc(nx*sizeof(double));
  dr=1./(nx-2);
  for(i=0;i<nx;i++)r0[i]=(i-.5)*dr;

  error=1.;
  itimes=0;

  xy0[0]=0.;
  xy0[1]=sin(theta0);
  xy0[2]=0.;
  xy0[3]=0.;
  xy0[4]=0.;
  xy0[5]=cos(theta0);
  p0[0]=Mz;
  p0[1]=Mx0;
  p0[2]=f;
  p0[3]=theta0;
  if(verbose){printf("Iter (%lf,%lf,%lf,%lf)\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);}

  ug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

  e0=(MFNV_C(ug,nu*(nx-1)+3)+MFNV_C(ug,nu*(nx-2)+3))/2;
  e1=(MFNV_C(ug,nu*(nx-1)  )+MFNV_C(ug,nu*(nx-2)  ))/2*
         (MFNV_C(ug,nu*(nx-1)+5)+MFNV_C(ug,nu*(nx-2)+5))/2
   -(MFNV_C(ug,nu*(nx-1)+4)+MFNV_C(ug,nu*(nx-2)+4))/2*
         (MFNV_C(ug,nu*(nx-1)+1)+MFNV_C(ug,nu*(nx-2)+1))/2;

  error=sqrt(e0*e0+e1*e1);

/*printf(" Iteration to find u0:\n");*/
  while(error>1.e-10 && itimes<200)
   {
    if(verbose){printf("%d error=%le Mx0(C)=%lf, Mz(A)=%lf, f(t)=%lf, th0=%lf\n",itimes,error,p0[0],p0[1],p0[2],p0[3]);}

    dMz=1.e-5;
    dMx0=1.e-5;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz+dMz;
    p0[1]=Mx0;
    p0[2]=f;
    p0[3]=theta0;

    tug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

    de0=(MFNV_C(tug,nu*(nx-1)+3)+MFNV_C(tug,nu*(nx-2)+3))/2;
    de1=(MFNV_C(tug,nu*(nx-1)  )+MFNV_C(tug,nu*(nx-2)  ))/2*
         (MFNV_C(tug,nu*(nx-1)+5)+MFNV_C(tug,nu*(nx-2)+5))/2
   -(MFNV_C(tug,nu*(nx-1)+4)+MFNV_C(tug,nu*(nx-2)+4))/2*
         (MFNV_C(tug,nu*(nx-1)+1)+MFNV_C(tug,nu*(nx-2)+1))/2;
    MFFreeNVector(tug);
    a00=(de0-e0)/dMz;
    a10=(de1-e1)/dMz;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz;
    p0[1]=Mx0+dMx0;
    p0[2]=f;
    p0[3]=theta0;

    tug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);
    de0=(MFNV_C(tug,nu*(nx-1)+3)+MFNV_C(tug,nu*(nx-2)+3))/2;
    de1=(MFNV_C(tug,nu*(nx-1)  )+MFNV_C(tug,nu*(nx-2)  ))/2*
         (MFNV_C(tug,nu*(nx-1)+5)+MFNV_C(tug,nu*(nx-2)+5))/2
   -(MFNV_C(tug,nu*(nx-1)+4)+MFNV_C(tug,nu*(nx-2)+4))/2*
         (MFNV_C(tug,nu*(nx-1)+1)+MFNV_C(tug,nu*(nx-2)+1))/2;
    MFFreeNVector(tug);
    a01=(de0-e0)/dMx0;
    a11=(de1-e1)/dMx0;

    dMz =-( a11*e0-a01*e1)/(a00*a11-a01*a10);
    dMx0=-(-a10*e0+a00*e1)/(a00*a11-a01*a10);
    corr=sqrt(dMz*dMz+dMx0*dMx0);
    xnorm=sqrt(Mz*Mz+Mx0*Mx0);
    if(verbose){printf("      Delta=(%lf,%lf) ||=%lf X=(%lf,%lf),||=%lf\n",dMx0,dMz,corr,Mx0,Mz,sqrt(Mz*Mz+Mx0*Mx0));fflush(stdout);}
/*  if(corr>.1){dMz=.1*dMz/corr;dMx0=.1*dMx0/corr;}*/
    if(0&&corr>1e-2*xnorm){dMz=1.e-2*xnorm*dMz/corr;dMx0=1.e-2*xnorm*dMx0/corr;}
    Mz=Mz+dMz;
    Mx0=Mx0+dMx0;
    itimes++;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz;
    p0[1]=Mx0;
    p0[2]=f;
    p0[3]=theta0;

    MFFreeNVector(ug);
    ug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

    e0=(MFNV_C(ug,nu*(nx-1)+3)+MFNV_C(ug,nu*(nx-2)+3))/2;
    e1=(MFNV_C(ug,nu*(nx-1)  )+MFNV_C(ug,nu*(nx-2)  ))/2*
           (MFNV_C(ug,nu*(nx-1)+5)+MFNV_C(ug,nu*(nx-2)+5))/2
     -(MFNV_C(ug,nu*(nx-1)+4)+MFNV_C(ug,nu*(nx-2)+4))/2*
           (MFNV_C(ug,nu*(nx-1)+1)+MFNV_C(ug,nu*(nx-2)+1))/2;

    error=sqrt(e0*e0+e1*e1);
   }
  if(error<1.e-8)
   {
    if(verbose){printf("  got(%lf,%lf,%lf,%lf)\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);}
    Tan=MFIMFTangentSpace(M,ug);
    u0=MFCreateNVector(n);
    MFIMFProject(M,ug,Tan,u0);
    MFFreeNKMatrix(Tan);
   }else{
    u0=(MFNVector)NULL;
    if(verbose){printf("  failed\n");fflush(stdout);}
   }
  MFFreeNVector(ug);
  free(r0);
  return u0;
 }

void flipf(int nx,int nu,MFNVector u)
 {
  int i;
  double th0;

  for(i=0;i<nx;i++)
   {
    MFNVSetC(u,nu*i+0,-MFNV_C(u,nu*i+0));
    MFNVSetC(u,nu*i+1,-MFNV_C(u,nu*i+1));
    MFNVSetC(u,nu*i+2,-MFNV_C(u,nu*i+2));
    MFNVSetC(u,nu*i+3,-MFNV_C(u,nu*i+3));
    MFNVSetC(u,nu*i+4,-MFNV_C(u,nu*i+4));
    MFNVSetC(u,nu*i+5,-MFNV_C(u,nu*i+5));
   }
  MFNVSetC(u,nx*nu+2,-MFNV_C(u,nx*nu+2));
  th0=MFNV_C(u,nx*nu+3)+3.1415926;
  if(th0>2*3.1415926)th0=th0-2*3.1415926;
  MFNVSetC(u,nx*nu+3,th0);

  return;
 }

void flipMx0AndMz(int nx,int nu,MFNVector u)
 {
  int i;
  double th0;

  for(i=0;i<nx;i++)
   {
    MFNVSetC(u,nu*i+2,-MFNV_C(u,nu*i+2));
    MFNVSetC(u,nu*i+3,-MFNV_C(u,nu*i+3));
   }
  MFNVSetC(u,nx*nu+0,-MFNV_C(u,nx*nu+0));
  MFNVSetC(u,nx*nu+1,-MFNV_C(u,nx*nu+1));

  return;
 }

void flipMz(int nx,int nu,MFNVector u)
 {
  int i;
  double th0;

  for(i=0;i<nx;i++)
   {
    MFNVSetC(u,nu*i+0,-MFNV_C(u,nu*i+0));
    MFNVSetC(u,nu*i+1,-MFNV_C(u,nu*i+1));
   }
  MFNVSetC(u,nx*nu+0,-MFNV_C(u,nx*nu+0));
  th0=-MFNV_C(u,nx*nu+3);
  while(th0>2*3.1415926)th0=th0-2*3.1415926;
  while(th0<0.)th0=th0+2*3.1415926;
  MFNVSetC(u,nx*nu+3,th0);

  return;
 }

void flipMx0(int nx,int nu,MFNVector u)
 {
  flipMx0AndMz(nx,nu,u);
  flipMz(nx,nu,u);
  return;
 }

void moveToPrinciple(int nx,int nu,MFNVector u)
 {
  printf("moveToPrinciple, was (%lf,%lf,%lf,%lf)\n",MFNV_C(u,nx*nu+1),MFNV_C(u,nx*nu+0),MFNV_C(u,nx*nu+2),MFNV_C(u,nx*nu+3));
  while(MFNV_C(u,nx*nu+3)<-PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)+2*PI);}
  while(MFNV_C(u,nx*nu+3)> PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)-2*PI);}

  if(MFNV_C(u,nx*nu+3)>PI/2)flipf(nx,nu,u);
  while(MFNV_C(u,nx*nu+3)<-PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)+2*PI);}
  while(MFNV_C(u,nx*nu+3)> PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)-2*PI);}
  if(MFNV_C(u,nx*nu+3)<-PI/2)flipf(nx,nu,u);
  while(MFNV_C(u,nx*nu+3)<-PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)+2*PI);}
  while(MFNV_C(u,nx*nu+3)> PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)-2*PI);}
  while(MFNV_C(u,nx*nu+3)< 0.)flipMz(nx,nu,u);
  printf("       result is     (%lf,%lf,%lf,%lf)\n",MFNV_C(u,nx*nu+1),MFNV_C(u,nx*nu+0),MFNV_C(u,nx*nu+2),MFNV_C(u,nx*nu+3));

  return;
 }

MFNVector GetInitialPointI(MFImplicitMF M,double Mx0,double Mz, double f, double theta0)
 {
  double r,dr;
  double xy0[6],p0[4],p1[4];
  double *r0;
  MFNVector u;
  MFNVector ug;
  MFNVector du[2];
  MFNKMatrix Tan;
  int i,n,nx,nu,np;

  printf("GetInitialPointI\n");fflush(stdout);

  n=MFIMF_N(M);
  nx=MFTPBVPGetNX(M);
  nu=MFTPBVPGetNU(M);
  np=MFTPBVPGetNP(M);

  r0=(double*)malloc(nx*sizeof(double));
  dr=1./(nx-2);
  for(i=0;i<nx;i++)r0[i]=(i-.5)*dr;

  xy0[0]=0.;
  xy0[1]=sin(theta0);
  xy0[2]=0.;
  xy0[3]=0.;
  xy0[4]=0.;
  xy0[5]=cos(theta0);
  p0[0]=Mz;
  p0[1]=Mx0;
  p0[2]=f;
  p0[3]=theta0;
  printf("     (%lf,%lf,%lf,%lf)\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);

  ug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);
  Tan=MFIMFTangentSpace(M,ug);
  u=MFCreateNVector(n);
/*
  du[0]=MFCreateNVector(n);
  du[1]=MFCreateNVector(n);
  for(i=0;i<nx*nu;i++)
   {
    MFNVSetC(du[0],i,0.);
    MFNVSetC(du[1],i,0.);
   }
  MFNVSetC(du[0],nx*nu+0,0.);MFNVSetC(du[1],nx*nu+0,0.);
  MFNVSetC(du[0],nx*nu+1,0.);MFNVSetC(du[1],nx*nu+1,0.);
  MFNVSetC(du[0],nx*nu+2,1.);MFNVSetC(du[1],nx*nu+2,0.);
  MFNVSetC(du[0],nx*nu+3,0.);MFNVSetC(du[1],nx*nu+3,1.);
  Tan=MFCreateNKMatrix(2,du);
*/

  if(!MFIMFProject(M,ug,Tan,u))
   {
    printf("     (%lf,%lf,%lf,%lf) failed\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);
    MFFreeNVector(u);
    u=(MFNVector)NULL;
   }else{
    printf("  got(%lf,%lf,%lf,%lf)\n",MFNV_C(u,nx*nu+1),MFNV_C(u,nx*nu+0),MFNV_C(u,nx*nu+2),MFNV_C(u,nx*nu+3));fflush(stdout);
   }

  free(r0);
  MFFreeNVector(ug);
/*
  MFFreeNVector(du[0]);
  MFFreeNVector(du[1]);*/

  MFFreeNKMatrix(Tan);

  return u;
 }

void moveToPositive(int nx,int nu,MFNVector u)
 {
  printf("moveToPositive, was (%lf,%lf,%lf,%lf)\n",MFNV_C(u,nx*nu+1),MFNV_C(u,nx*nu+0),MFNV_C(u,nx*nu+2),MFNV_C(u,nx*nu+3));

  if(MFNV_C(u,nx*nu+0)<0)flipMz(nx,nu,u);
  if(MFNV_C(u,nx*nu+1)<0)flipMx0(nx,nu,u);
  if(MFNV_C(u,nx*nu+2)<0)flipf(nx,nu,u);

  while(MFNV_C(u,nx*nu+3)<0){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)+2*PI);}
  while(MFNV_C(u,nx*nu+3)>2*PI){MFNVSetC(u,nx*nu+3,MFNV_C(u,nx*nu+3)-2*PI);}

  printf("       result is     (%lf,%lf,%lf,%lf)\n",MFNV_C(u,nx*nu+1),MFNV_C(u,nx*nu+0),MFNV_C(u,nx*nu+2),MFNV_C(u,nx*nu+3));

  return;
 }

MFNVector GetInitialPointMx0f(MFImplicitMF M,double Mz,double Mx0, double f, double theta0)
 {
  double r,dr;
  double xy0[6],p0[4],p1[4];
  double *r0;
  double error,e0,e1,corr;
  double df,dMx0,dy1,dy3,a00,a01,a10,a11,de0,de1;
  int itimes;
  MFNVector ug,tug;
  MFNVector u0;
  MFNKMatrix Tan;
  int i,n,nx,nu,np;
  double xnorm;
  printf("GetInitialPointMx0f\n");fflush(stdout);

  n=MFIMF_N(M);
  nx=MFTPBVPGetNX(M);
  nu=MFTPBVPGetNU(M);
  np=MFTPBVPGetNP(M);

  r0=(double*)malloc(nx*sizeof(double));
  dr=1./(nx-2);
  for(i=0;i<nx;i++)r0[i]=(i-.5)*dr;

  error=1.;
  itimes=0;

  xy0[0]=0.;
  xy0[1]=sin(theta0);
  xy0[2]=0.;
  xy0[3]=0.;
  xy0[4]=0.;
  xy0[5]=cos(theta0);
  p0[0]=Mz;
  p0[1]=Mx0;
  p0[2]=f;
  p0[3]=theta0;
  printf("Iter (%lf,%lf,%lf,%lf)\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);

  ug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

  e0=(MFNV_C(ug,nu*(nx-1)+3)+MFNV_C(ug,nu*(nx-2)+3))/2;
  e1=(MFNV_C(ug,nu*(nx-1)  )+MFNV_C(ug,nu*(nx-2)  ))/2*
         (MFNV_C(ug,nu*(nx-1)+5)+MFNV_C(ug,nu*(nx-2)+5))/2
   -(MFNV_C(ug,nu*(nx-1)+4)+MFNV_C(ug,nu*(nx-2)+4))/2*
         (MFNV_C(ug,nu*(nx-1)+1)+MFNV_C(ug,nu*(nx-2)+1))/2;

  error=sqrt(e0*e0+e1*e1);

/*printf(" Iteration to find u0:\n");*/
  while(error>1.e-10 && itimes<200)
   {
    printf("%d error=%le Mx0(C)=%lf, Mz(A)=%lf, f(t)=%lf, th0=%lf\n",itimes,error,p0[0],p0[1],p0[2],p0[3]);

    df=1.e-5;
    dMx0=1.e-5;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz;
    p0[1]=Mx0;
    p0[2]=f+df;
    p0[3]=theta0;

    tug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

    de0=(MFNV_C(tug,nu*(nx-1)+3)+MFNV_C(tug,nu*(nx-2)+3))/2;
    de1=(MFNV_C(tug,nu*(nx-1)  )+MFNV_C(tug,nu*(nx-2)  ))/2*
         (MFNV_C(tug,nu*(nx-1)+5)+MFNV_C(tug,nu*(nx-2)+5))/2
   -(MFNV_C(tug,nu*(nx-1)+4)+MFNV_C(tug,nu*(nx-2)+4))/2*
         (MFNV_C(tug,nu*(nx-1)+1)+MFNV_C(tug,nu*(nx-2)+1))/2;
    MFFreeNVector(tug);
    a00=(de0-e0)/df;
    a10=(de1-e1)/df;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz;
    p0[1]=Mx0+dMx0;
    p0[2]=f;
    p0[3]=theta0;

    tug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);
    de0=(MFNV_C(tug,nu*(nx-1)+3)+MFNV_C(tug,nu*(nx-2)+3))/2;
    de1=(MFNV_C(tug,nu*(nx-1)  )+MFNV_C(tug,nu*(nx-2)  ))/2*
         (MFNV_C(tug,nu*(nx-1)+5)+MFNV_C(tug,nu*(nx-2)+5))/2
   -(MFNV_C(tug,nu*(nx-1)+4)+MFNV_C(tug,nu*(nx-2)+4))/2*
         (MFNV_C(tug,nu*(nx-1)+1)+MFNV_C(tug,nu*(nx-2)+1))/2;
    MFFreeNVector(tug);
    a01=(de0-e0)/dMx0;
    a11=(de1-e1)/dMx0;

    df  =-( a11*e0-a01*e1)/(a00*a11-a01*a10);
    dMx0=-(-a10*e0+a00*e1)/(a00*a11-a01*a10);
    corr=sqrt(df*df+dMx0*dMx0);
    xnorm=sqrt(f*f+Mx0*Mx0);
    printf("      Delta=(%lf,%lf) ||=%lf X=(%lf,%lf),||=%lf\n",dMx0,df,corr,Mx0,f,sqrt(Mx0*Mx0+f*f));fflush(stdout);
/*  if(corr>.1){df =.1*df /corr;dMx0=.1*dMx0/corr;}*/
/*  if(corr>1e-2*xnorm){df =1.e-2*xnorm*df /corr;dMx0=1.e-2*xnorm*dMx0/corr;}*/
    f=f+df;
    Mx0=Mx0+dMx0;
    itimes++;

    xy0[0]=0.;
    xy0[1]=sin(theta0);
    xy0[2]=0.;
    xy0[3]=0.;
    xy0[4]=0.;
    xy0[5]=cos(theta0);
    p0[0]=Mz;
    p0[1]=Mx0;
    p0[2]=f;
    p0[3]=theta0;

    MFFreeNVector(ug);
    ug=MFTPBVPIntegrateForInitialSolution(M,xy0,p0,r0);

    e0=(MFNV_C(ug,nu*(nx-1)+3)+MFNV_C(ug,nu*(nx-2)+3))/2;
    e1=(MFNV_C(ug,nu*(nx-1)  )+MFNV_C(ug,nu*(nx-2)  ))/2*
           (MFNV_C(ug,nu*(nx-1)+5)+MFNV_C(ug,nu*(nx-2)+5))/2
     -(MFNV_C(ug,nu*(nx-1)+4)+MFNV_C(ug,nu*(nx-2)+4))/2*
           (MFNV_C(ug,nu*(nx-1)+1)+MFNV_C(ug,nu*(nx-2)+1))/2;

    error=sqrt(e0*e0+e1*e1);
   }
  if(error<1.e-8)
   { 
    printf("  got(%lf,%lf,%lf,%lf)\n",p0[1],p0[0],p0[2],p0[3]);fflush(stdout);
    Tan=MFIMFTangentSpace(M,ug);
    u0=MFCreateNVector(n);
    MFIMFProject(M,ug,Tan,u0);
    MFFreeNKMatrix(Tan);
   }else{
    printf("  failed\n");fflush(stdout);
    u0=(MFNVector)NULL;
   }
  MFFreeNVector(ug);
  free(r0);

  return u0;
 }

int MFRodStopN(MFImplicitMF M,MFNVector U0,MFNKMatrix Phi0,MFNVector U1,MFNKMatrix Phi1,void *data)
 {
  double u0,um,up;
  double theta0,Mx0,Mz,f;
  double delta,Omega,T,gnu,m;
  double x1,y1,x2,y2,x3,y3,del0,del1;
  int s00,s01,s02;
  int s10,s11,s12;
  int N0,N1;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(U0,nx*nu+0);
  Mx0   =MFNV_C(U0,nx*nu+1);
  f     =MFNV_C(U0,nx*nu+2);
  theta0=MFNV_C(U0,nx*nu+3);

  s00=1;
  if(theta0<PI)s00=-1;
  s02=1;
  if(theta0<-PI)s02=-1;
  s01=1;
  if(Mz<0)s01=-1;

  u0=cos(theta0);
  delta=(Mz*Mz+Mx0*Mx0)*(Mz*Mz+Mx0*Mx0)+8*f*(2*f+u0*(Mx0*Mx0-Mz*Mz))-16*Mz*Mx0*f*sin(theta0);
  up=(Mz*Mz+Mx0*Mx0+sqrt(delta))/4/f;
  um=(Mz*Mz+Mx0*Mx0-sqrt(delta))/4/f;

  gnu=up-u0;
  if(um<u0)gnu=up-um;
  m=fabs(u0-um)/gnu;
  if(m>0)T=2*sqrt(2.)*K(sqrt(m))/sqrt(fabs(f*gnu));
   else  T=2*sqrt(2.)*Ki(sqrt(-m))/sqrt(fabs(f*gnu));
  Omega=2*PI/T;

  N0=2*floor((Omega-1.)/2.)+1;
  if(um<u0)N0=N0+1;

  x1=(MFNV_C(U0,nu*(nx-2)+0)+MFNV_C(U0,nu*(nx-1)+0))/2;
  y1=(MFNV_C(U0,nu*(nx-2)+1)+MFNV_C(U0,nu*(nx-1)+1))/2;
  x2=(MFNV_C(U0,nu*(nx-2)+2)+MFNV_C(U0,nu*(nx-1)+2))/2;
  y2=(MFNV_C(U0,nu*(nx-2)+3)+MFNV_C(U0,nu*(nx-1)+3))/2;
  x3=(MFNV_C(U0,nu*(nx-2)+4)+MFNV_C(U0,nu*(nx-1)+4))/2;
  y3=(MFNV_C(U0,nu*(nx-2)+5)+MFNV_C(U0,nu*(nx-1)+5))/2;
  del0=1.-(x1*y1+x2*y2+x3*y3)/3.1415926;
  if(del0<1.)N0=-N0;

  Mz    =MFNV_C(U1,nx*nu+0);
  Mx0   =MFNV_C(U1,nx*nu+1);
  f     =MFNV_C(U1,nx*nu+2);
  theta0=MFNV_C(U1,nx*nu+3);

  s10=1;
  if(theta0<PI)s10=-1;
  s12=1;
  if(theta0<-PI)s12=-1;
  s11=1;
  if(Mz<0)s11=-1;

  u0=cos(theta0);
  delta=(Mz*Mz+Mx0*Mx0)*(Mz*Mz+Mx0*Mx0)+8*f*(2*f+u0*(Mx0*Mx0-Mz*Mz))-16*Mz*Mx0*f*sin(theta0);
  up=(Mz*Mz+Mx0*Mx0+sqrt(delta))/4/f;
  um=(Mz*Mz+Mx0*Mx0-sqrt(delta))/4/f;

  gnu=up-u0;
  if(um<u0)gnu=up-um;
  m=fabs(u0-um)/gnu;
  if(m>0)T=2*sqrt(2.)*K(sqrt(m))/sqrt(fabs(f*gnu));
   else  T=2*sqrt(2.)*Ki(sqrt(-m))/sqrt(fabs(f*gnu));
  Omega=2*PI/T;

  N1=2*floor((Omega-1.)/2.)+1;
  if(um<u0)N1=N1+1;

  x1=(MFNV_C(U1,nu*(nx-2)+0)+MFNV_C(U1,nu*(nx-1)+0))/2;
  y1=(MFNV_C(U1,nu*(nx-2)+1)+MFNV_C(U1,nu*(nx-1)+1))/2;
  x2=(MFNV_C(U1,nu*(nx-2)+2)+MFNV_C(U1,nu*(nx-1)+2))/2;
  y2=(MFNV_C(U1,nu*(nx-2)+3)+MFNV_C(U1,nu*(nx-1)+3))/2;
  x3=(MFNV_C(U1,nu*(nx-2)+4)+MFNV_C(U1,nu*(nx-1)+4))/2;
  y3=(MFNV_C(U1,nu*(nx-2)+5)+MFNV_C(U1,nu*(nx-1)+5))/2;
  del1=1.-(x1*y1+x2*y2+x3*y3)/3.1415926;
  if(del1<1.)N1=-N1;

/*printf("StopN. N0=%d, del0=%lf, s00=%d, s01=%d\n",N0,del0,s00,s01);
  printf("       N1=%d, del1=%lf, s10=%d, s11=%d\n",N1,del1,s10,s11);
  printf("  result=%d\n",N0!=N1 || del0>1 || del1>1 || s00*s10<0&&s01*s11<0);*/

  return N0!=N1 || del0>1 || del1>1 || s00*s10<0&&s01*s11<0
                                    || s00*s10<0&&s02*s12<0;
 }


int MFRodStopD(MFImplicitMF M, MFNVector U0,MFNKMatrix Phi0,MFNVector U1,MFNKMatrix Phi1,void *data)
 {
  double x1,y1,x2,y2,x3,y3;
  double d0,d1;
  int nx=150;
  int np=4;
  int nu=6;

  x1=(MFNV_C(U0,nu*(nx-2)+0)+MFNV_C(U0,nu*(nx-1)+0))/2;
  y1=(MFNV_C(U0,nu*(nx-2)+1)+MFNV_C(U0,nu*(nx-1)+1))/2;
  x2=(MFNV_C(U0,nu*(nx-2)+2)+MFNV_C(U0,nu*(nx-1)+2))/2;
  y2=(MFNV_C(U0,nu*(nx-2)+3)+MFNV_C(U0,nu*(nx-1)+3))/2;
  x3=(MFNV_C(U0,nu*(nx-2)+4)+MFNV_C(U0,nu*(nx-1)+4))/2;
  y3=(MFNV_C(U0,nu*(nx-2)+5)+MFNV_C(U0,nu*(nx-1)+5))/2;
  d0=x1*y1+x2*y2+x3*y3;

  x1=(MFNV_C(U1,nu*(nx-2)+0)+MFNV_C(U1,nu*(nx-1)+0))/2;
  y1=(MFNV_C(U1,nu*(nx-2)+1)+MFNV_C(U1,nu*(nx-1)+1))/2;
  x2=(MFNV_C(U1,nu*(nx-2)+2)+MFNV_C(U1,nu*(nx-1)+2))/2;
  y2=(MFNV_C(U1,nu*(nx-2)+3)+MFNV_C(U1,nu*(nx-1)+3))/2;
  x3=(MFNV_C(U1,nu*(nx-2)+4)+MFNV_C(U1,nu*(nx-1)+4))/2;
  y3=(MFNV_C(U1,nu*(nx-2)+5)+MFNV_C(U1,nu*(nx-1)+5))/2;
  d1=x1*y1+x2*y2+x3*y3;

  return fabs(d0)>.05 || fabs(d1)>.05;
 }

int MFRodProjectToDraw(MFNVector u, double *x, void *d)
 {
  static char RoutineName[]={"MFRodProjectToDraw"};
  struct MFTPBVPData *data;
  int i,j;
  int n,nu,nx,np;
  double min,max;
  double a,c,t;
  double x1,y1,x2,y2,x3,y3,kp0,kp;
  double mx0,mz,f;
  double theta0,m3;
  double u0,up,um,delta;
  double T,Omega,R,Wr;
  double dist,dth0;
  double d1x,d1y,d1z;
  double gnu;
  double m;
  int verbose=0;
  double Delta,del;
  int N;
  char name[1024];
  FILE *fid;

#ifdef MFDEBUG
  _heap_check();
#endif

  data=(struct MFTPBVPData*)d;

  if(x==(double*)NULL)return 9;

  n=MFNV_NC(u);

  nu=data->nu;
  nx=data->nx;
  np=data->np;

  if(verbose){printf("%s, n=%d, nu=%d, nx=%d, np=%d, comp=%d\n",RoutineName,n,nu,nx,np,nx*nu+np+nx+1);fflush(stdout);}
  min=0;max=0;
  a=MFNV_C(u,nx*nu+0);
  c=MFNV_C(u,nx*nu+1);
  t=MFNV_C(u,nx*nu+2);

  y1=(MFNV_C(u,1)+MFNV_C(u,nu+1))/2;
  y3=(MFNV_C(u,5)+MFNV_C(u,nu+5))/2;

  mz=a;
  mx0=c;
  f=t;
  theta0=MFNV_C(u,nx*nu+3);
  m3=mx0*y1+mz*y3;

  u0=cos(theta0);
  delta=(mz*mz+mx0*mx0)*(mz*mz+mx0*mx0)+8*f*(2*f+u0*(mx0*mx0-mz*mz))-16*mz*mx0*f*sin(theta0);
  if(fabs(f)>1.e-5)
   {
    up=(mz*mz+mx0*mx0+sqrt(delta))/4/f;
    um=(mz*mz+mx0*mx0-sqrt(delta))/4/f;
   }else{
    up=1000.;
    um=0.;
   }

  gnu=up-u0;
  if(um<u0)gnu=up-um;
  m=fabs(u0-um)/gnu;
  if(verbose)
   {
    if(m>0){printf("   before K(%lf)\n",sqrt(m));fflush(stdout);}
     else {printf("   before Ki(%lf)\n",sqrt(-m));fflush(stdout);}
   }
  if(fabs(m)>1.e-5)
   {
    if(m>0)T=2*sqrt(2.)*K(sqrt(m))/sqrt(fabs(f*gnu));
     else  T=2*sqrt(2.)*Ki(sqrt(-m))/sqrt(fabs(f*gnu));
   }else T=PI/2.;
  if(verbose){printf("   after  K()\n");fflush(stdout);}
  Omega=2*PI/T;

  N=2*floor((Omega-1.)/2.)+1;
  if(um<u0)N=N+1;

  if(N>1.e5)N=0;
  if(N<-1.e5)N=0;
  if(N!=N)N=0;

  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  del=1.-(x1*y1+x2*y2+x3*y3)/3.1415926;
  if(del<1.)N=-N;

  R=0.;
  Wr=0.;

  d1x=0.;
  d1y=1.;
  d1z=0.;
  x1=(MFNV_C(u,0)+MFNV_C(u,nu+0))/2;
  y1=(MFNV_C(u,1)+MFNV_C(u,nu+1))/2;
  x2=(MFNV_C(u,2)+MFNV_C(u,nu+2))/2;
  y2=(MFNV_C(u,3)+MFNV_C(u,nu+3))/2;
  x3=(MFNV_C(u,4)+MFNV_C(u,nu+4))/2;
  y3=(MFNV_C(u,5)+MFNV_C(u,nu+5))/2;
  kp0=t*x1*y1+(t*x2-c)*y2;
  for(i=1;i<nx-1;i++)
   {
    x1=(MFNV_C(u,nu*i+0)+MFNV_C(u,nu*(i+1)+0))/2;
    y1=(MFNV_C(u,nu*i+1)+MFNV_C(u,nu*(i+1)+1))/2;
    x2=(MFNV_C(u,nu*i+2)+MFNV_C(u,nu*(i+1)+2))/2;
    y2=(MFNV_C(u,nu*i+3)+MFNV_C(u,nu*(i+1)+3))/2;
    x3=(MFNV_C(u,nu*i+4)+MFNV_C(u,nu*(i+1)+4))/2;
    y3=(MFNV_C(u,nu*i+5)+MFNV_C(u,nu*(i+1)+5))/2;

    kp=t*x1*y1+(t*x2-c)*y2;
    if(kp*kp0<=0.)
     {
      if(i==1)
       {
        if(kp>0.)min+=1;
        if(kp<0.)max+=1;
       }else{
        if(kp>0.)min+=2;
        if(kp<0.)max+=2;
       }
     }
    kp0=kp;
   }

  x[0]=mx0;
  x[1]=mz;
  x[2]=f;
  x[3]=theta0;
  x[4]=x1*y1+x2*y2+x3*y3;
  x[5]=up;
  x[6]=um;
  x[7]=N;
  x[8]=Omega;

  for(i=0;i<9;i++)if(x[i]!=x[i])x[i]=0.;

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return 0;
 }

double MzPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return -Mz;
 }

double Mx0Positive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return -Mx0;
 }

double Mx0Negative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return Mx0;
 }

double FPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return -f;
 }

double DNegative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return d-1;
 }

double DPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return 1-d;
 }

double FLessThanSixteen(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return f-16.;
 }

double MzNegative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return Mz;
 }

double MzSmallPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return Mz-.05;
 }

double MzSmallNegative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=x1*y1+x2*y2+x3*y3-1;

  return -Mz-.05;
 }

double DSmallPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return d-1-.1;
 }

double DSmallNegative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return 1-d-.1;
 }

double DSmallerPositive(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return d-1-.05;
 }

double DSmallerNegative(MFNVector u)
 {
  double x1,y1,x2,y2,x3,y3;
  double d,Mx0,Mz,f;
  int nx=150;
  int np=4;
  int nu=6;

  Mz    =MFNV_C(u,nx*nu+0);
  Mx0   =MFNV_C(u,nx*nu+1);
  f     =MFNV_C(u,nx*nu+2);
  x1=(MFNV_C(u,nu*(nx-2)+0)+MFNV_C(u,nu*(nx-1)+0))/2;
  y1=(MFNV_C(u,nu*(nx-2)+1)+MFNV_C(u,nu*(nx-1)+1))/2;
  x2=(MFNV_C(u,nu*(nx-2)+2)+MFNV_C(u,nu*(nx-1)+2))/2;
  y2=(MFNV_C(u,nu*(nx-2)+3)+MFNV_C(u,nu*(nx-1)+3))/2;
  x3=(MFNV_C(u,nu*(nx-2)+4)+MFNV_C(u,nu*(nx-1)+4))/2;
  y3=(MFNV_C(u,nu*(nx-2)+5)+MFNV_C(u,nu*(nx-1)+5))/2;
  d=1-(x1*y1+x2*y2+x3*y3)/3.1415926;

  return 1-d-.05;
 }

double K(double p)
 {
/* This from Spanier and Oldham "An Atlas of Functions" page 61:8 */
  double g,a,t,e,s,k;

  if(p<.067)return .5*PI*(16.-5*p*p)/(16-9*p*p);
  if(1.-p<1.e-15)return 1.e20;

  g=sqrt(1-p*p);
  a=1.;
  t=1.;
  e=1.+g*g;
  s=g*a;
  while(t*(a*a-s)>6e-12)
   {
    s=g*a;
    a=(a+g)/2;
    g=sqrt(s);
    t=2*t;
    e-=t*(a*a-s);
   }
  k=.5*PI/g;

  return k;
 }

double Ki(double p)
 {
  double x;
  double result;

  x=p/sqrt(1+p*p);
  result=K(x)/sqrt(1.+x*x);

  return result;
 }
