/*
    %W%
    %D% %T%

    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  MFNRegion Omega;
  MFAtlas A;
  MFNVector u0;
  MFContinuationMethod H;

  M=MFIMFCreateAlgebraicExpression("[x,y,z]",
                                   "[x*x+y*y+z*z-1.]");
  Omega=MFNRegionCreateHyperCube(3,1.1);

  u0=MFCreateNVector(3);
  MFNVSetC(u0,0, 0.);
  MFNVSetC(u0,1, 1.);
  MFNVSetC(u0,2, 0.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.1);         /* Max distance from TS to M */
  MFHendersonsMethodSetMaxCharts(H,-1);       /* -1 means infinite */
  MFHendersonsMethodSetVerbose(H,1);          /* Write info to stdout */
  MFHendersonsMethodSetPage(H,1);             /* Page out non-interior polyhedra */
  MFHendersonsMethodSetDumpToPlotFile(H,1);   /* Write polyhedra to a plotfile */
  MFHendersonsMethodSetDumpToCenterFile(H,0); /* Write points to a file */
  MFHendersonsMethodSetFilename(H,"Sphere");

  A=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,A);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(A);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);
  MFFreeHendersonsMethod(H);

  return(0);
 }
