/* 
    @(#)ComputeSphereSub.c	1.4
    02/04/19 14:47:54
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>
#include <MFFortran.h>

int DrawSphere(MFNVector,double*,void*);

static void F(int *n,double *z,int *m,double *f)
 {
  f[0]= z[0]*z[0]+z[1]*z[1]+z[2]*z[2]-1.;

  return;
 }

static void dF(int *n,double *z,int *m,double *df)
 {
  int i,N;

  N=*m;
  df[0+N* 0]= 2*z[0];
  df[0+N* 1]= 2*z[1];
  df[0+N* 2]= 2*z[2];

  return;
 }

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j;
  int n,k;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0[2];
  MFContinuationMethod H;

  n=3;
  k=2;

  M=MFIMFCreateAlgebraicSubroutine(n,k,F,dF);
  MFIMFSetProjectForDraw(M,DrawSphere);

  Omega=MFNRegionCreateHyperCube(n,1.1);

  u0[0]=MFCreateNVector(n);
  for(j=0;j<n;j++)MFNVSetC(u0[0],j,0.);
  MFNVSetC(u0[0],1,1.);

  u0[1]=MFCreateNVector(n);
  for(j=0;j<n;j++)MFNVSetC(u0[1],j,0.);
  MFNVSetC(u0[1],2,1.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetMaxCharts(H,100);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"SphereSub");

  S=MFComputeAtlasMultiple(H,M,Omega,2,u0);

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0[0]);
  MFFreeNVector(u0[1]);

  return(0);
 }

int DrawSphere(MFNVector v,double *u,void *d)
 {
  if(u==(double*)NULL||v==(MFNVector)NULL) return 3;

  u[0]=MFNV_C(v,0);
  u[1]=MFNV_C(v,1);
  u[2]=MFNV_C(v,2);

  return 0;
 }
