/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFPrint.h>
#include <math.h>

int MFTaylor24ProjectToDraw(MFNVector,double*,void*);

void findSolution(double,double,double,double,double,double,double*,double*);
int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j,n;
  int n0;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  MFNVector ll,ur;
  MFContinuationMethod H;

/* Model for Taylor (2,4) mode interaction */

  printf("Expression is %s\n","[x*(2*x*x+y*y-r+s)+x*y,y*(x*x+2*y*y-r-s)-x*x]");fflush(stdout);
  M=MFIMFCreateAlgebraicExpressionWithRadius("[x,y,r,s]","[x*(2*x*x+y*y-r+s)+x*y,y*(x*x+2*y*y-r-s)-x*x]",.05);
  MFIMFSetProjectForDraw(M,MFTaylor24ProjectToDraw);

  n=MFIMF_N(M);
  ll=MFCreateNVector(n);
  ur=MFCreateNVector(n);
  MFNVSetC(ll,0,-5.);MFNVSetC(ur,0,5.);
  MFNVSetC(ll,1,-5.);MFNVSetC(ur,1,5.);
  MFNVSetC(ll,2,-1.);MFNVSetC(ur,2,1.);
  MFNVSetC(ll,3,-1.); MFNVSetC(ur,3,1.); 
  Omega=MFNRegionCreateHyperCubeByCorners(n,ll,ur);
  MFFreeNVector(ll);
  MFFreeNVector(ur);
  
  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetDotMin(H,.9);
  MFHendersonsMethodSetMaxCharts(H,100000);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetBranchSwitch(H,200);
  MFHendersonsMethodSetDumpToPlotFile(H,1);   /* Write polyhedra to a plotfile */
  MFHendersonsMethodSetDumpToCenterFile(H,0); /* Write points to a file */
  MFHendersonsMethodSetDumpToRestartFile(H,0);

  u0=MFCreateNVector(n);

/* Trivial */

  MFNVSetC(u0,0,0.);
  MFNVSetC(u0,1,0.);
  MFNVSetC(u0,2,-.4);
  MFNVSetC(u0,3,0.);
  MFHendersonsMethodSetFilename(H,"Taylor24");

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }

int MFTaylor24ProjectToDraw(MFNVector u, double *x, void *d)
 {
  static char RoutineName[]={"MFTaylor24ProjectToDraw"};

  if(x==(double*)NULL)return 3;

  x[0]=MFNV_C(u,2);
  x[1]=MFNV_C(u,3);
  x[2]=.5*(MFNV_C(u,0)+MFNV_C(u,1));

  return 0;
 }
