/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  double Rxy=.8;
  double Rz=.5;
  MFContinuationMethod H;

  printf("Torus, xy R=%lf, z R=%lf (surface area=%lf)\n",Rxy,Rz,4*3.1415926*3.1415926*Rxy*Rz);fflush(stdout);
  M=MFIMFCreateTorus(0.,0.,0.,Rxy,Rz);
  n=MFIMF_N(M);
  Omega=MFNRegionCreateCube(-2.,-2.,-2.,2.,2.,2.);

  u0=MFCreateNVector(n);
  MFNVSetC(u0,0, 0.);
  MFNVSetC(u0,1, 1.3);
  MFNVSetC(u0,2, 0.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"Torus");

  S=MFComputeAtlas(H,M,Omega,u0);

  MFCloseAtlas(H,S);
  printf("Done computing Atlas\n");fflush(stdout);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
