/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0[2];
  FILE *fid;
  MFContinuationMethod H;

/*M=MFIMFCreateAlgebraicExpression("[x,y,z]","[(z-x)*(x+y+z)]");*/
  M=MFIMFCreateAlgebraicExpression("[x,y,z]","[(z-x-.5*(y-1)*(y-1))*(x+y+z)]");
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,1.1);

  u0[0]=MFCreateNVector(n);
  MFNVSetC(u0[0],0, 0.);
  MFNVSetC(u0[0],1, 1.);
  MFNVSetC(u0[0],2,-1.);

  u0[1]=MFCreateNVector(n);
  MFNVSetC(u0[1],0, 1.);
  MFNVSetC(u0[1],1, 1.);
  MFNVSetC(u0[1],2, 1.);

  H=MFCreateHendersonsMethod();
  MFHendersonsMethodSetEpsilon(H,.01);
  MFHendersonsMethodSetMaxCharts(H,-1);
  MFHendersonsMethodSetVerbose(H,1);
  MFHendersonsMethodSetPage(H,1);
  MFHendersonsMethodSetPageEvery(H,100000);
  MFHendersonsMethodSetDumpToPlotFile(H,1);
  MFHendersonsMethodSetDumpToCenterFile(H,0);
  MFHendersonsMethodSetFilename(H,"Transcritical");
  MFHendersonsMethodSetDotMin(H,.9);
  MFHendersonsMethodSetBranchSwitch(H,20);

  S=MFComputeAtlasMultiple(H,M,Omega,1,u0);

  MFCloseAtlas(H,S);

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0[0]);
  MFFreeNVector(u0[1]);

  return(0);
 }
