/* 
    @(#)DrawChemicalReactor.c	1.2
    02/04/19 16:20:11
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <sh.h>
#include <math.h>

void tricon(float *s,int *ldx,int *ldy,int *ldz,int *nx,int *ny,int *nz,float *xl,float *xr,float *yl,float *yr,float *zl,float *zr,float *v,int *nl,float *h,int *ired,int *igrn,int *iblu,int *iered,int *iegrn,int *ieblu,float *vol,float *area,int *contour,int *volume);

double f(double u1,double u2, double x2)
 {
  double f;

  double q_cr = 1;             /* Reference coolant flow rate */
  double V_r = 1;              /* Reference Volume */
  double qu_o  = 1;            /* Normalized feed rate */
  double gama  = 20;           /* Normalized activation energy */
  double beta  = 7.0/20;       /* Normalized heat of reaction */
  double x10   = 1.0;          /* Normalized Feed concentration */
  double x20   = 0;            /* Normalized Feed Temperature */
  double x30   = -1/20;        /* Normalized Coolant Feed Temperature */
  double xi   = 1.0;           /* Ratio of heat capcities */
  double qu_c  = 1.0;          /* Ratio of Reference flows between feed and coolant*/ 
  double phi   = 0.11;         /* Normalized rate of reaction */
  double deltaR = 0.5;         /* Normalized heat transfer coefficient    */

  f = qu_o*(x20-x2) + (beta*phi*exp(gama*x2/(1+x2))*u2*qu_o*x10)/(qu_o+phi*exp(gama*x2/(1+x2))*u2)-(deltaR*u2*qu_c*u1)*(x2-x30)/(qu_c*u1+xi*deltaR*u2);   

  return f;
 }


int main(int argc, char *argv[])
 {
  float *s;
  int nx,ny,nz;
  int ldx,ldy,ldz;
  int i,j,k;
  float xl,xr;
  float yl,yr;
  float zl,zr;
  float x,y,z,dx,dy,dz;
  float v;
  int nl;
  float h;
  int red,green,blue;
  float vol=0.;
  float area=0.;
  int contour,volume;
  int zero=0;
  int half=100;
  int full=255;
  float d,a,b;

  xl=0.;
  xr=2.;

  yl=0.;
  yr=1.5;

  zl=0.;
  zr=.5;

  nx=100;
  ny=30;
  nz=30;

  ldx=nx+1;
  ldy=ny+1;
  ldz=nz+1;

  dx=(xr-xl)/nx;
  dy=(yr-yl)/ny;
  dz=(zr-zl)/nz;

  contour=TRUE;
  volume=FALSE;
  red=zero;
  green=zero;
  blue=full;
  nl=1;
  v=0.;
  h=0.;

  s=(float*)malloc((ldx+3)*(ldy+3)*(ldz+3)*sizeof(float));
  shinit(&half,&half,&half);
  shSetOutputFilename("CSTR");
  d=100.;
  a=10.;
  b=20.;
  shview(&d,&a,&b,&xl,&xr,&yl,&yr,&zl,&zr);
  shlinc(&full,&full,&full);
  shcube(&xl,&xr,&yl,&yr,&zl,&zr);

  for(i=0;i<ldx+3;i++)
    for(j=0;j<ldy+3;j++)
      for(k=0;k<ldz+3;k++)
       {
        x=xl+(i-2)*dx;
        y=yl+(j-2)*dy;
        z=zl+(k-2)*dz;
        s[i+(ldx+3)*(j+(ldy+3)*k)]=f(x,y,z);
       }

  tricon(s,&ldx,&ldy,&ldz,&nx,&ny,&nz,&xl,&xr,&yl,&yr,&zl,&zr,&v,&nl,&h,&red,&green,&blue,&red,&green,&blue,&vol,&area,&contour,&volume);

  shpause();
  shend();

  return(0);
 }
