/* 
    @(#)IMFExpansionSpace.c	1.7
    03/06/13 17:12:07
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */

#include <MFNSpace.h>
#include <MFNVector.h>
#include <MFError.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

void MFSetError(int,char*,char*,int,char*);
static char IMFExpansionSpaceErrorMsg[256]="";

struct IMFExpansionSpaceData
 {
  int n;
 };

void MFFreeExpansionSpaceData(void*);
double IMFExpansionSpaceDistance(MFNSpace,MFNVector,MFNVector,void*);
void IMFExpansionSpaceDirection(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void IMFExpansionSpaceAdd(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void IMFExpansionSpaceScale(MFNSpace,double,MFNVector,MFNVector,void*);
double IMFExpansionSpaceInner(MFNSpace,MFNVector,MFNVector,void*);

MFNSpace IMFCreateExpansionSpace(int n)
 {
  static char RoutineName[]={"IMFCreateExpansionSpace"};
  MFNSpace this;
  struct IMFExpansionSpaceData *data;


  this=MFCreateNSpaceBaseClass("ExpansionSpace");
  MFNSpaceSetDistance(this,IMFExpansionSpaceDistance);
  MFNSpaceSetInnerProduct(this,IMFExpansionSpaceInner);
  MFNSpaceSetDirection(this,IMFExpansionSpaceDirection);
  MFNSpaceSetAdd(this,IMFExpansionSpaceAdd);
  MFNSpaceSetScale(this,IMFExpansionSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeExpansionSpaceData);
  data=(struct IMFExpansionSpaceData*)malloc(sizeof(struct IMFExpansionSpaceData));
  data->n=n;

  MFNSpaceSetData(this,(void*)data);

  return this;
 }

void MFFreeExpansionSpaceData(void *data)
 {
  static char RoutineName[]={"MFFreeExpansionSpaceData"};
  struct IMFExpansionSpaceData *d;

  d=(struct IMFExpansionSpaceData*)data;
  if(d!=(struct IMFExpansionSpaceData*)NULL)free(d);

  return;
 }

double IMFExpansionSpaceDistance(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"IMFExpansionSpaceDistance"};
  double result;
  int i;
  struct IMFExpansionSpaceData *data;

  data=(struct IMFExpansionSpaceData*)d;

  result=0.;
  for(i=0;i<data->n;i++)result+=pow(MFNV_C(v0,i)-MFNV_C(v1,i),2);
  result=sqrt(result);

  return result;
 }

void IMFExpansionSpaceDirection(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector diff,void *d)
 {
  static char RoutineName[]={"IMFExpansionSpaceDirection"};
  int i;
  struct IMFExpansionSpaceData *data;

  data=(struct IMFExpansionSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(diff,i,MFNV_C(v1,i)-MFNV_C(v0,i));

  return;
 }

double IMFExpansionSpaceInner(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"IMFExpansionSpaceInner"};
  double result;
  int i;
  int verbose;
  struct IMFExpansionSpaceData *data;

  data=(struct IMFExpansionSpaceData*)d;

  verbose=0;
  if(verbose){printf("%s 0x%8.8x\n",RoutineName,this);fflush(stdout);}
  if(verbose){printf(" v0 ");MFPrintNVector(stdout,v0);printf("\n");fflush(stdout);}
  if(verbose){printf(" v1 ");MFPrintNVector(stdout,v1);printf("\n");fflush(stdout);}
  if(verbose){printf(" n=%d\n",data->n);fflush(stdout);}

  result=0.;
  for(i=0;i<data->n;i++)
    result+=MFNV_C(v0,i)*MFNV_C(v1,i);

  return result;
 }

void IMFExpansionSpaceAdd(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector sum,void *d)
 {
  static char RoutineName[]={"IMFExpansionSpaceAdd"};
  int i;
  struct IMFExpansionSpaceData *data;

  data=(struct IMFExpansionSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(sum,i,MFNV_C(v1,i)+MFNV_C(v0,i));

  return;
 }

void IMFExpansionSpaceScale(MFNSpace this,double s, MFNVector v,MFNVector prod,void *d)
 {
  static char RoutineName[]={"IMFExpansionSpaceScale"};
  int i;
  struct IMFExpansionSpaceData *data;

  data=(struct IMFExpansionSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(prod,i,s*MFNV_C(v,i));

  return;
 }

