#include <MFAtlas.h>
#include <MFBinaryTree.h>
#include <MFChart.h>
#include <IMFFlow.h>
#include <IMFExpansion.h>
#include <IMFExpansionSpace.h>
#include <IMFIntegrateFat.h>
#include <math.h>
#include <stdio.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

/* #define DOTRAJ*/

MFBinaryTree MFAtlasGetBB(MFAtlas);
MFChart MFAtlasChart(MFAtlas,int);
int MFAtlasAddChartToList(MFAtlas,MFNVector,MFNKMatrix,double);

FILE *IMFTraj=(FILE*)NULL;
int   IMFNTraj=0.;

int MEHKellerInt(IMFFlow,double*,double,double,int);
int IMFIsNear(MFAtlas,MFChart,MFChart);
void MFChartSetPaged(MFChart);

void IMFExtendAtlasAlongFatTraj(MFAtlas I, IMFFlow F, MFNVector u0, double dt, double ti, double tf,double epsilon,MFNRegion Omega, int maxSteps, int maxCharts,double Rmax,int nSkip)
 {
  static char RoutineName[]={"IMFExtendAtlasAlongFatTraj"};

  double t,tout;
  MFNKMatrix TS;
  double R;
  int i,j;
  int p,ni;
  double d;
  int nsteps;
  int neq,n;
  IMFExpansion E;
  MFNVector u;
  MFNVector f,g;
  double *fp;
  MFKVector s;
  double *y0;
  double *y;
  MFListOfCharts L;
  MFBinaryTree BTree;
  MFNSpace space;
  int k,prev;
  IMFFlow Fp;
  int nearChart;
  double dMin;
  double Rf;
  MFChart c;
  int bail;

#ifdef DOTRAJ
  if(IMFTraj==(FILE*)NULL)
   {
    IMFTraj=fopen("IMFTraj.dx","w");
    IMFNTraj=0;
   }
#endif

  E=IMFExpansionNVGetE(u0);
  printf("Initial expansion\n");fflush(stdout);
  IMFPrintExpansion(stdout,E);fflush(stdout);
  neq=IMFExpansionDataLn(E);
  space=MFIMFNSpace(MFAtlasMF(I));
  BTree=MFAtlasGetBB(I);

  t=ti;

  n=IMFFlowN(F);
  k=IMFExpansionK(E);

  Fp=IMFCreateFatFlow(F,k);

  fp=(double*)malloc(n*sizeof(double));if(fp==(double*)NULL)abort();
  f=MFCreateWrappedNVector(n,fp);
  g=MFCreateNVector(n);
  s=MFCreateKVector(IMFExpansionK(E));

  nsteps=0;
  u=u0;
  MFRefNVector(u);
  while(1)
   {

    bail=0;
    if(t>tf){printf("Ending integration, final time reached (%lf>%lf)\n",t,tf);fflush(stdout);bail=1;}

    if(nsteps>=maxSteps){printf("Ending integration, too many steps\n");fflush(stdout);goto FreeAndReturn;}

    if(MFAtlasNumberOfCharts(I)>=maxCharts){printf("Ending integration, too many charts\n");fflush(stdout);goto FreeAndReturn;}

    if(!MFNRegionInterior(Omega,u)){printf("Ending integration, point is outside Omega. ");MFPrintNVector(stdout,u);printf("\n");fflush(stdout);bail=1;}

    IMFEvaluateFlow(F,u,fp);
    d=sqrt(MFNSpaceInner(space,f,f));
    if(d<epsilon){printf("Ending integration, |F| (%le<%le) too small (near fixed point)\n",d,epsilon);fflush(stdout);bail=1;}

    E=IMFExpansionNVGetE(u);
    TS=IMFExpansionTS(E);

    MFNSpaceScale(space,1./d,f,f);
    MFMVMulT(space,TS,f,s);
    MFMVMul(space,TS,s,g);
    d=MFNSpaceDistance(space,f,g);
    if(0&&d>epsilon){printf("Ending integration, F too far from tangent space, d=%l2\n",d);fflush(stdout);bail=1;}

    R=IMFExpansionR(E,epsilon);
    Rf=IMFFlowR(F,epsilon,u,TS);
    if(Rf<R)R=Rf;
    if(R>Rmax||R!=R)R=Rmax;
    c=MFCreateChart(MFAtlasMF(I),u,TS,R);
    if(0&&R<epsilon){printf("Ending integration, R too small\n");fflush(stdout);bail=1;}

/* find closest point to keep trajectories from getting too close */

    y0=IMFExpansionData(E);
    L=MFCreateListOfNearbyCharts(BTree,y0,R);
    ni=MFNumberOfIntersectingCharts(L);
    nearChart=-1;
    dMin=2*R;
    for(p=0;p<ni;p++)
     {
      j=MFIntersectingChart(L,p);
      if(MFChartPaged(MFAtlasChart(I,j)))continue;
      if(!IMFIsNear(I,MFAtlasChart(I,j),c))continue;
      d=MFNSpaceDistance(space,MFAtlasCenterOfChart(I,j),u);
      if(d<.5*MFAtlasChartRadius(I,j)&&(nearChart==-1||d<dMin))
       {
        dMin=d;
        nearChart=j;
       }
     }
    MFFreeListOfIntersectingCharts(L);
    MFFreeChart(c);

/* Passed checks, add chart */

/* NV Types: 
           0     normal
           1     fixed point or on c
           2     interpolation point
           3     skipped point, not on manifold - don't plot
*/

    prev=-1;
    if(  ((nearChart==-1||IMFExpansionNVGetType(u)==1)||IMFExpansionNVGetType(u)==2) && nsteps>=nSkip )
     {
      printf("%d) Point %d on fat traj.",MFAtlasNumberOfCharts(I),nsteps);MFPrintNVector(stdout,u);printf(", R=%lf, t=%lf\n",R,t);fflush(stdout);
      prev=MFAtlasAddChartWithAll(I,u,TS,R);
      if(prev>0 && (IMFExpansionNVGetType(u)==1||IMFExpansionNVGetType(u)==2))MFChartSetSingular(MFAtlasChart(I,prev));
       else if(prev>1)MFChartSetNonSingular(MFAtlasChart(I,prev));

     }else{
      printf("*) Point %d on fat traj. ",MFAtlasNumberOfCharts(I),nsteps);MFPrintNVector(stdout,u);printf(", R=%lf, t=%lf\n",R,t);fflush(stdout);
      if(nsteps<nSkip){printf("    is being skipped, %d<%d\n",nsteps,nSkip);fflush(stdout);}
        else {printf("    is too near chart %d (distance = %lf)\n",nearChart,dMin);fflush(stdout);}
      prev=MFAtlasAddChartToList(I,u,TS,R);
      IMFExpansionNVSetType(u,3);
      MFChartSetPaged(MFAtlasChart(I,prev));
     }

    if(bail){if(prev>-1)MFChartSetSingular(MFAtlasChart(I,prev));goto FreeAndReturn;}

    nsteps++;

/* Now find the next */

    y0=IMFExpansionData(E);
    y=(double*)malloc(IMFExpansionDataLn(E)*sizeof(double));if(y==(double*)NULL)abort();
    for(i=0;i<IMFExpansionDataLn(E);i++)y[i]=y0[i];
#ifdef DOTRAJ
    if(nsteps>nSkip)
     {
      for(i=0;i<IMFExpansionN(E);i++)fprintf(IMFTraj," %lf",y[i]);fprintf(IMFTraj,"\n");fflush(IMFTraj);IMFNTraj++;
     }
#endif

    d=0;
    tout=t+R;
    while(t+1.e-7<=tout && t<tf && d<1.03*R)
     {
      if(!MEHKellerInt(Fp,y,t,tout,2)){
#ifdef DOTRAJ
        for(i=0;i<IMFExpansionN(E);i++)fprintf(IMFTraj," %lf",y[i]);fprintf(IMFTraj,"\n");fflush(IMFTraj);IMFNTraj++;
#endif
        printf("Ending integration, Integrator failed to converge\n");fflush(stdout);goto FreeAndReturn;}
/* was ,5); */
    
      for(i=0;i<n;i++)
       {
        if(y[i]!=y[i]){printf("Ending integration, Integrator returned a NaN\n");fflush(stdout);goto FreeAndReturn;}
       }
      d=0.;for(i=0;i<n;i++)d+=pow(y[i]-y0[i],2);d=sqrt(d);
      t=tout;
      tout=t+.1*R;
     }
#ifdef DOTRAJ
    for(i=0;i<IMFExpansionN(E);i++)fprintf(IMFTraj," %lf",y[i]);fprintf(IMFTraj,"\n");fflush(IMFTraj);IMFNTraj++;
#endif
    E=IMFCreateExpansion(n,k);
    IMFExpansionSetDerivatives(E,y,y+n,y+n*k,(double*)NULL);
    free(y);

    u0=u;
    u=IMFCreateExpansionNVector(E,t,IMFExpansionNVGetSigma(u0),prev,0);
    IMFExpansionNVSetS0(u,IMFExpansionNVGetS0(u0));
    IMFExpansionNVSetChart0(u,IMFExpansionNVGetChart0(u0));
    MFFreeNVector(u0);
    IMFFreeExpansion(E);
   }

FreeAndReturn:
  MFFreeNVector(u);
  MFFreeNVector(f);
  free(fp);
  IMFFreeFlow(Fp);
  MFFreeNVector(g);
  MFFreeKVector(s);

  printf("done IMFExtendAtlasAlongFatTraj\n");fflush(stdout);
  return;
 }

int MEHKellerInt(IMFFlow F,double *y,double ti,double tf,int nsteps)
 {
  static char RoutineName[]={"MEHKellerInt"};
  double t,dt;
  static int meq=-1;
  static double *y0=(double*)NULL;
  static double *ym=(double*)NULL;
  static double *y1=(double*)NULL;
  static double *f=(double*)NULL;
  static double *A=(double*)NULL;
  static double *b=(double*)NULL;
  double fNorm;
  int i,j,itimes;
  double error;
  int verbose=0;
  int neq;
  MFNVector vy;
  MFNVector vym;
  MFNVector vy0;

  neq=IMFFlowN(F);

  if(verbose){printf("%s, neq=%d integrate from t=%lf to %lf\n",RoutineName,neq,ti,tf);fflush(stdout);}

  if(meq<neq)
   {
    y0=(double*)realloc((void*)y0,neq*sizeof(double));if(y0==(double*)NULL)abort();
    ym=(double*)realloc((void*)ym,neq*sizeof(double));if(ym==(double*)NULL)abort();
    y1=(double*)realloc((void*)y1,neq*sizeof(double));if(y1==(double*)NULL)abort();
    f =(double*)realloc((void*)f ,neq*sizeof(double));if(f ==(double*)NULL)abort();
    A =(double*)realloc((void*)A ,neq*neq*sizeof(double));if(A ==(double*)NULL)abort();
    b =(double*)realloc((void*)b ,neq*sizeof(double));if(b ==(double*)NULL)abort();
    meq=neq;
   }

  vy=MFCreateWrappedNVector(neq,y);
  vym=MFCreateWrappedNVector(neq,ym);
  vy0=MFCreateWrappedNVector(neq,y0);

  IMFEvaluateFlow(F,vy,f);
  fNorm=0.;for(i=0;i<neq;i++)fNorm+=f[i]*f[i]; fNorm=sqrt(fNorm);
  if(fNorm>10.)fNorm=10.;
  nsteps = 10*(tf-ti)*fNorm+2;

  t=ti;
  dt=(tf-ti)/nsteps;
  if(verbose){printf("   nsteps=%d, fNorm=%lf\n",nsteps,fNorm);fflush(stdout);}

  for(j=0;j<neq;j++)y0[j]=y[j];
  for(i=0;i<nsteps;i++)
   {
    if(verbose){printf("Keller Box for IVP, step %d/%d, t=%lf\n",i,nsteps,t);fflush(stdout);  }
    IMFEvaluateFlow(F,vy0,f);

    for(j=0;j<neq;j++)y1[j]=y0[j]+dt*f[j];

    error=1.;
    itimes=0;
    while(error>1.e-5)
     {
      for(j=0;j<neq;j++)ym[j]=.5*(y0[j]+y1[j]);
      IMFEvaluateFlow(F,vym,f);
      IMFEvaluateDerivativeOfFlow(F,vym,A);
      for(j=0;j<neq*neq;j++)A[j]=-.5*dt*A[j];
      error=0.;
      for(j=0;j<neq;j++)
       {
        b[j]=-(y1[j]-y0[j]-dt*f[j]);
        error+=b[j]*b[j];
        A[j+neq*j]=A[j+neq*j]+1.;
       }
      if(verbose){printf("Iteration %d, error=%le\n",itimes,error);fflush(stdout);  }

      MFSolveFull(neq,A,b);
      for(j=0;j<neq;j++)y1[j]+=b[j];
      itimes++;
      if(itimes>100)return 0;
     }
    t+=dt;
    for(j=0;j<neq;j++)y0[j]=y1[j];
   }
  for(j=0;j<neq;j++)y[j]=y1[j];

  MFFreeNVector(vy);
  MFFreeNVector(vym);
  MFFreeNVector(vy0);
  return 1;
 }

MFChart IMFStepAlongFatTraj(MFAtlas I, IMFFlow F, MFNVector u0, double R0, double dt, double epsilon,MFNRegion Omega, double Rmax)
 {
  static char RoutineName[]={"IMFStepAlongFatTraj"};

  double t,tout;
  MFNKMatrix TS;
  double R;
  int i,j;
  int p,ni;
  double d;
  int nsteps;
  int neq,n;
  IMFExpansion E;
  MFNVector u;
  MFNVector f,g;
  double *fp;
  MFKVector s;
  double *y0;
  double *y;
  MFListOfCharts L;
  MFBinaryTree BTree;
  MFNSpace space;
  int k,prev;
  IMFFlow Fp;
  int nearChart;
  double dMin;
  double Rf;
  MFChart c;
  int bail;

  E=IMFExpansionNVGetE(u0);
  printf("Initial expansion\n");fflush(stdout);
  IMFPrintExpansion(stdout,E);fflush(stdout);
  neq=IMFExpansionDataLn(E);
  space=MFIMFNSpace(MFAtlasMF(I));
  BTree=MFAtlasGetBB(I);

  t=0.;
  R=R0;

  n=IMFFlowN(F);
  k=IMFExpansionK(E);

  Fp=IMFCreateFatFlow(F,k);

  fp=(double*)malloc(n*sizeof(double));if(fp==(double*)NULL)abort();
  f=MFCreateWrappedNVector(n,fp);
  g=MFCreateNVector(n);
  s=MFCreateKVector(IMFExpansionK(E));

  nsteps=0;
  u=u0;
  MFRefNVector(u);
  y=(double*)malloc(IMFExpansionDataLn(E)*sizeof(double));if(y==(double*)NULL)abort();
  while(1)
   {
    E=IMFExpansionNVGetE(u);

/* find the next */

    y0=IMFExpansionData(E);
    for(i=0;i<IMFExpansionDataLn(E);i++)y[i]=y0[i];

    d=0;
    tout=t+R;
    while(t+1.e-7<=tout && d<1.03*R)
     {
      if(!MEHKellerInt(Fp,y,t,tout,2)){
       printf("Ending integration, Integrator failed to converge\n");fflush(stdout);goto FreeAndReturn;}
/* was ,5); */
    
      for(i=0;i<n;i++)
       {
        if(y[i]!=y[i]){printf("Ending integration, Integrator returned a NaN\n");fflush(stdout);goto FreeAndReturn;}
       }
      d=0.;for(i=0;i<n;i++)d+=pow(y[i]-y0[i],2);d=sqrt(d);
      t=tout;
      tout=t+.1*R;
     }
    E=IMFCreateExpansion(n,k);
    IMFExpansionSetDerivatives(E,y,y+n,y+n*k,(double*)NULL);

    u0=u;
    u=IMFCreateExpansionNVector(E,t,IMFExpansionNVGetSigma(u0),prev,0);
    MFFreeNVector(u0);
    IMFFreeExpansion(E);
    bail=0;

    if(!MFNRegionInterior(Omega,u)){printf("Ending integration, point is outside Omega. ");MFPrintNVector(stdout,u);printf("\n");fflush(stdout);bail=1;}

    IMFEvaluateFlow(F,u,fp);
    d=sqrt(MFNSpaceInner(space,f,f));
    if(d<epsilon){printf("Ending integration, |F| (%le<%le) too small (near fixed point)\n",d,epsilon);fflush(stdout);bail=1;}

    E=IMFExpansionNVGetE(u);
    TS=IMFExpansionTS(E);

    MFNSpaceScale(space,1./d,f,f);
    MFMVMulT(space,TS,f,s);
    MFMVMul(space,TS,s,g);
    d=MFNSpaceDistance(space,f,g);
    if(0&&d>epsilon){printf("Ending integration, F too far from tangent space, d=%l2\n",d);fflush(stdout);bail=1;}

    R=IMFExpansionR(E,epsilon);
    Rf=IMFFlowR(F,epsilon,u,TS);
    if(Rf<R)R=Rf;
    if(R>Rmax||R!=R)R=Rmax;
    c=MFCreateChart(MFAtlasMF(I),u,TS,R);
    if(0&&R<epsilon){printf("Ending integration, R too small\n");fflush(stdout);bail=1;}

/* find closest point to keep trajectories from getting too close */

    y0=IMFExpansionData(E);
    L=MFCreateListOfNearbyCharts(BTree,y0,R);
    ni=MFNumberOfIntersectingCharts(L);
    nearChart=-1;
    dMin=2*R;
    for(p=0;p<ni;p++)
     {
      j=MFIntersectingChart(L,p);
      if(MFChartPaged(MFAtlasChart(I,j)))continue;
      if(!IMFIsNear(I,MFAtlasChart(I,j),c))continue;
      d=MFNSpaceDistance(space,MFAtlasCenterOfChart(I,j),u);
      if(MFAtlasChartRadius(I,j)&&(nearChart==-1||d<dMin))
       {
        dMin=d;
        nearChart=j;
       }
     }
    MFFreeListOfIntersectingCharts(L);

/* Passed checks, add chart */

    prev=-1;
    if((nearChart==-1||IMFExpansionNVGetType(u)==1)||IMFExpansionNVGetType(u)==2)
     {
      goto FreeAndReturn;
     }else{
      printf("*) Point %d on fat traj. ",MFAtlasNumberOfCharts(I),nsteps);MFPrintNVector(stdout,u);printf(", R=%lf, t=%lf\n",R,t);fflush(stdout);
      printf("    is too near chart %d (distance = %lf)\n",nearChart,dMin);fflush(stdout);
     }

    if(bail){if(prev>-1)MFChartSetSingular(MFAtlasChart(I,prev));MFFreeChart(c);c=(MFChart)NULL;goto FreeAndReturn;}

    MFFreeChart(c);
    nsteps++;
   }

FreeAndReturn:
  MFFreeNVector(u);
  MFFreeNVector(f);
  free(fp);
  IMFFreeFlow(Fp);
  MFFreeNVector(g);
  MFFreeKVector(s);
    free(y);

  printf("done %s\n",RoutineName);fflush(stdout);
  return c;
 }
