/*
    @(#)MFInvMFNSpace.c	1.2
    02/04/29 16:57:33

    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   December 9, 2002   modified InvMF          */

#include <MFNSpace.h>
#include <MFNVector.h>
#include <MFError.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

void MFSetError(int,char*,char*,int,char*);
static char MFInvMFNSpaceErrorMsg[256]="";

double MFInvMFNSpaceDistance(MFNSpace,MFNVector,MFNVector,void*);
void MFInvMFNSpaceDirection(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void MFInvMFNSpaceAdd(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void MFInvMFNSpaceScale(MFNSpace,double,MFNVector,MFNVector,void*);
double MFInvMFNSpaceInner(MFNSpace,MFNVector,MFNVector,void*);
void MFInvMFNSpaceFreeData(void*);

void MFWriteInvMFNSpace(FILE*,MFNSpace,void*);
MFNSpace MFReadInvMFNSpace(FILE*);

double *MFNV_CStar(MFNVector this);

extern double scaled_dp(double*,double*,void*);

struct InvMFNSpaceData
 {
  int N;
  int n;
 };

MFNSpace InvMFCreateNSpace(int N, int n)
 {
  static char RoutineName[]={"MFCreateInvMFNSpace"};
  MFNSpace this;
  struct InvMFNSpaceData *data;


  this=MFCreateNSpaceBaseClass("InvMFNSpace");

  data=(struct InvMFNSpaceData*)malloc(sizeof(struct InvMFNSpaceData));if(data==NULL)abort();
  data->N=N;
  data->n=n;
  MFNSpaceSetData(this,data);
  MFNSpaceSetFreeData(this,MFInvMFNSpaceFreeData);

  MFNSpaceSetDistance(this,MFInvMFNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFInvMFNSpaceInner);
  MFNSpaceSetDirection(this,MFInvMFNSpaceDirection);
  MFNSpaceSetAdd(this,MFInvMFNSpaceAdd);
  MFNSpaceSetScale(this,MFInvMFNSpaceScale);

  return this;
 }

double MFInvMFNSpaceDistance(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"MFNSpaceDistance"};
  double result;
  MFNVector dv;

  dv=MFCloneNVector(v0);
  MFInvMFNSpaceDirection(this,v0,v1,dv,d);
  result=MFInvMFNSpaceInner(this,dv,dv,d);
  MFFreeNVector(dv);

  return sqrt(result);
 }

void MFInvMFNSpaceDirection(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector diff,void *d)
 {
  static char RoutineName[]={"MFNSpaceDirection"};
  int i;


  MFNVDiff(v1,v0,diff);

  return;
 }

double MFInvMFNSpaceInner(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"MFNSpaceInner"};
  double result;
  double *x0;
  double *x1;
  int i,n;
  struct InvMFNSpaceData *data;

  data=(struct InvMFNSpaceData *)d;

  x0=MFNV_CStar(v0);
  x1=MFNV_CStar(v1);
  n=data->n;
  result=0.;for(i=0;i<n;i++)result+=x0[i]*x1[i];

  return result;
 }

void MFWriteInvMFNSpace(FILE *fid,MFNSpace this,void *d)
 {
  static char RoutineName[]={"MFWriteInvMF"};
  int i;

  return;
 }

void MFInvMFNSpaceAdd(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector sum,void *d)
 {
  static char RoutineName[]={"MFNSpaceAdd"};
  int i;

  MFNVAdd(v0,v1,sum);

  return;
 }

void MFInvMFNSpaceScale(MFNSpace this,double s, MFNVector v,MFNVector w,void *d)
 {
  static char RoutineName[]={"MFNSpaceScale"};
  double *pv;
  double *pw;
  int i,n;

  pv=MFNV_CStar(v);
  pw=MFNV_CStar(w);
  n=MFNV_NC(v);

  for(i=0;i<n;i++)pw[i]=s*pv[i];

  return;
 }

void MFInvMFNSpaceFreeData(void *d)
 {
  struct InvMFNSpaceData *data;

  data=(struct InvMFNSpaceData *)d;
  if(data!=(struct InvMFNSpaceData*)NULL)free(data);
 }
