#include <MFAtlas.h>
#include <IMFExpansion.h>
#include <IMFFlow.h>
#include <IMFNSpace.h>
#include <MFFortran.h>
#include <math.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

#ifndef MAX
#define MAX(X, Y) ((X) < (Y) ? (Y) : (X))
#endif

struct IMFSphereOnEData
    {
     int n;
     int k;
     IMFExpansion E;
     IMFFlow F;
     double eps;
     double R;
    };

void MFFreeIMFSphereOnEData(void *d)
 {
  struct IMFSphereOnEData *data;
  data=(struct IMFSphereOnEData*)d;
  IMFFreeExpansion(data->E);
  IMFFreeFlow(data->F);
  free(data);
 }

int MFProjectIMFSphereOnE(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentIMFSphereOnE(int,int,MFNVector,MFNKMatrix,void*);
int MFIMFSphereOnEProjectForBB(MFNVector,double*,void*);
double MFScaleIMFSphereOnE(int,int,MFNVector,MFNKMatrix,void*);

MFImplicitMF IMFCreateSphereOnExpansion(IMFExpansion E, IMFFlow F, double eps, double R, double r)
 {
/*   Manifold c with
          c-E(s)   =0
         |c-E(0)|^2=R^2               */

  MFImplicitMF this;
  int n,k;
  struct IMFSphereOnEData *data;
  MFNSpace space;

  n=IMFExpansionN(E);
  k=IMFExpansionK(E);
  this=MFIMFCreateBaseClass(n+k,k-1,"InvariantManifoldInitialCurve");
  space=InvMFCreateNSpace(n+k,n);
  MFIMFSetSpace(this,space);
  MFFreeNSpace(space);

  data=(struct IMFSphereOnEData*)malloc(sizeof(struct IMFSphereOnEData));if(data==NULL)abort();
  data->E=E;IMFRefExpansion(E);
  data->F=F;IMFRefFlow(F);
  data->eps=eps;
  data->R=R;
  data->n=n;
  data->k=k;
  MFIMFSetData(this,(void*)data);

  MFIMFSetFreeData(this,MFFreeIMFSphereOnEData);
  MFIMFSetProject(this,MFProjectIMFSphereOnE);
  MFIMFSetTangent(this,MFTangentIMFSphereOnE);
  MFIMFSetScale(this,MFScaleIMFSphereOnE);
  MFIMFSetR(this,r);

  MFIMFSetProjectForBB(this,MFIMFSphereOnEProjectForBB);
  MFIMFSetProjectForDraw(this,MFIMFSphereOnEProjectForBB);

  return this;
 }

int MFProjectIMFSphereOnE(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectIMFSphereOnE"};
  double err;
  double *A=(double*)NULL; /* [n*n];*/
  double *b=(double*)NULL; /* [n];  */
  int *ipvt=(int*)NULL;    /* [n];  */
  int one=1;
  int info=0;
  int itimes;
  int i,j,l,m;
  char trans='N';
  struct IMFSphereOnEData *data;
  double f;
  double *df=(double*)NULL;
  double t,dnorm,unorm;
  double *u0,*Phi,*u,*EPhi;

  double R;
  MFNVector vM0;
  MFNVector vM;
  MFNVector vdM;
  double *M0;
  double *M;
  double *dM;
  double *s;
  MFKVector vs;

  int verbose;

  data=(struct IMFSphereOnEData*)d;

  verbose=0;

  m=IMFExpansionN(data->E);
  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);
  if(verbose){printf("In %s. n=%d, m=%d, k=%d\n",RoutineName,n,m,k);fflush(stdout);}

  A=malloc(n*n*sizeof(double));
  b=malloc(n*sizeof(double));
  ipvt=malloc(n*sizeof(int));
  df=malloc(n*(n-k)*sizeof(double));

  for(i=0;i<n;i++)u[i]=u0[i];

  vs=MFCreateKVector(k+1);
  s=MFKV_CStar(vs);

  vM0=MFCreateNVector(m);
  M0=MFNV_CStar(vM0);

  vdM=MFCreateNVector(m);
  dM=MFNV_CStar(vdM);

  R=data->R;
  for(i=0;i<k+1;i++)s[i]=0.;
  IMFEvaluateExpansion(data->E,vs,vM0);

  for(i=0;i<k+1;i++)s[i]=u[m+i];

  err=1.;
  itimes=0;
  while(err>1.e-10)
   {
    for(i=0;i<k+1;i++)u[m+i]=s[i];
    IMFEvaluateExpansion(data->E,vs,vu);

    f=-R*R;for(i=0;i<m;i++)f+=(u[i]-M0[i])*(u[i]-M0[i]);

    b[0]=-f;
    for(j=1;j<k+1;j++)
      {b[j]=0.;for(i=0;i<k+1;i++)b[j]-=Phi[i+m+n*(j-1)]*(u[i+m]-u0[i+m]);}

    for(j=0;j<k+1;j++)
     {
      IMFEvaluateExpansionDerivative(data->E,vs,j,vdM);
      A[0+(k+1)*j]=0.;for(i=0;i<m;i++)A[0+(k+1)*j]+=2*(u[i]-M0[i])*dM[i];
  
      for(l=1;l<k+1;l++)
        A[l+(k+1)*j]=Phi[j+m+n*(l-1)];
     }

    if(verbose)
     {
      printf("System for Newton Correction\n");
      IMFPrintFull(k+1,A,b);
     }

    err=0.;
    for(i=0;i<k+1;i++)err+=b[i]*b[i];
    err=sqrt(err);
    if(verbose)printf("  %d  %le\n",itimes,err);fflush(stdout);

    MFSolveFull(k+1,A,b);

    t=1.;
    dnorm=fabs(b[0]);
    for(i=1;i<k+1;i++)
      if(fabs(b[i])>dnorm)dnorm=fabs(b[i]);
    unorm=fabs(s[0]);
    for(i=1;i<k+1;i++)
      if(fabs(s[i])>unorm)unorm=fabs(s[i]);
    if(dnorm>.1*(unorm+1))t=.1*(unorm+1)/dnorm; /* Damping */

    for(i=0;i<k+1;i++){s[i]=s[i]+t*b[i];u[i+m]=s[i];}

    itimes++;
    if(itimes>10)
     {
      free(A);
      free(b);
      free(ipvt);
      free(df);
      MFFreeKVector(vs);
      MFFreeNVector(vM0);
      MFFreeNVector(vdM);

      if(verbose){printf("Done %s -- too many iterations\n",RoutineName );fflush(stdout);}
      return 0;
     }
   }
  *index=0;
  for(i=0;i<k+1;i++)u[i+m]=s[i];
  if(verbose){printf("Result is ");MFPrintNVector(stdout,vu);printf("\n");fflush(stdout);}
  free(A);
  free(b);
  free(ipvt);
  free(df);
  MFFreeKVector(vs);
  MFFreeNVector(vM0);
  MFFreeNVector(vdM);

  if(verbose){printf("Done %s\n",RoutineName);fflush(stdout);}

  return 1;
 }

int MFTangentIMFSphereOnE(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentIMFSphereOnE"};
  double *A=(double*)NULL; /* [n*n];*/
  double *b=(double*)NULL; /* [n];  */
  int *ipvt=(int*)NULL;    /* [n];  */
  int one=1;
  int info=0;
  int itimes;
  int i,j,l,m;
  char trans='N';
  struct IMFSphereOnEData *data;
  double t;
  double *u0,*Phi,*u;

  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='A';                   /*    right eigenvectors */
  int lda;                        /* Leading dimension of A */
  double *sv;    /* n */
  double *U;    /* n*n */
  int ldU;
  double *vr;    /* n*k */
  double *V;    /* n*n */
  int ldV;
  double *work;                     /* Work array */
  int lwork;                     /* Length of work array */
  int msv;

  double R;
  MFNVector vM0;
  MFNVector vM;
  MFNVector vdM;
  MFKVector vs;
  double *M0;
  double *M;
  double *dM;
  double *s;

  int verbose;

  verbose=0;

  data=(struct IMFSphereOnEData*)d;

  if(verbose){printf("In %s\n",RoutineName);fflush(stdout);}

  m=IMFExpansionN(data->E);

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  A=malloc(n*n*sizeof(double));
  b=malloc(n*sizeof(double));
  ipvt=malloc(n*sizeof(int));
  vs=MFCreateKVector(k+1);
  s=MFKV_CStar(vs);
  vM0=MFCreateNVector(m);
  M0=MFNV_CStar(vM0);
  vdM=MFCreateNVector(m);
  dM=MFNV_CStar(vdM);

  R=data->R;
  for(i=0;i<k+1;i++)s[i]=0.;
  IMFEvaluateExpansion(data->E,vs,vM0);

  for(i=0;i<k+1;i++)s[i]=u[m+i];

  for(i=0;i<n;i++)b[i]=0.;

  for(j=0;j<k+1;j++)
   {
    IMFEvaluateExpansionDerivative(data->E,vs,j,vdM);
    A[0+(k+1)*j]=0.;for(i=0;i<m;i++)A[0+(k+1)*j]+=2*dM[i]*(u[i]-M0[i]);
 
    for(l=1;l<k+1;l++)A[l+(k+1)*j]=0.;
   }

  if(verbose)
   {
    printf("  u: (%lf",u[0]);for(i=1;i<m+k+1;i++)printf(",%lf",u[i]);printf(")\n");fflush(stdout);
    printf("System for Tangent\n");
    IMFPrintFull(k+1,A,b);
   }

  lda=k+1;
  ldU=k+1;

  U=(double*)malloc((k+1)*(k+1)*sizeof(double));
  if(U==(double*)NULL)abort();

  vr=(double*)malloc((k+1)*k*sizeof(double));
  if(vr==(double*)NULL)abort();

  ldV=(k+1);
  V=malloc((k+1)*(k+1)*sizeof(double));
  if(V==(double*)NULL)abort();

  sv=(double*)malloc((k+1)*sizeof(double));
  if(sv==(double*)NULL)abort();

  i=-1;
  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&lda,&lda,A,&lda,sv,U,&ldU,V,&ldV,&t,&i,&info);
  lwork=MAX(4*(k-1),5*(k-1)-4);
  lwork=t;

  work=malloc(lwork*sizeof(double));
  if(work==(double*)NULL)abort();

  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&lda,&lda,A,&lda,sv,U,&ldU,V,&ldV,work,&lwork,&info);

  msv=0;
  for(i=0;i<k+1;i++)
   {
    if(verbose){printf("Singular value %d is %lf\n",i,sv[i]);fflush(stdout);}
    if(fabs(sv[i])<1.e-7)
     {
      for(j=0;j<k+1;j++)vr[j+(k+1)*msv]=V[i+(k+1)*j];
      if(verbose){printf(" null vector (%lf",vr[0+(k+1)*msv]);for(j=1;j<k+1;j++)printf(",%lf",vr[j+(k+1)*msv]);printf(")\n");fflush(stdout);}
      msv++;
     }
   }

  t=0.;
  for(j=0;j<k+1;j++)t+=vr[j]*vr[j];
  if(fabs(t)>1.e-15)
   {
    t=1./sqrt(t);
    for(j=0;j<k+1;j++){vr[j]=vr[j]*t;}
   }

  for(i=1;i<msv;i++)
   {
    for(l=0;l<i;l++)
     {
      t=0.;
      for(j=0;j<k+1;j++)t+=vr[j+(k+1)*l]*vr[j+(k+1)*i];
      for(j=0;j<k+1;j++)vr[j+(k+1)*i]=vr[j+(k+1)*i]-t*vr[j+(k+1)*l];
     }

    t=0.;
    for(j=0;j<k+1;j++)t+=vr[j+(k+1)*i]*vr[j+(k+1)*i];
    if(fabs(t)>1.e-15)
     {
      t=1./sqrt(t);
      for(j=0;j<k+1;j++){vr[j+(k+1)*i]=vr[j+(k+1)*i]*t;}
     }
   }

  for(j=0;j<n*k;j++)Phi[j]=0.;
  for(j=0;j<k;j++)
   {
    for(l=0;l<k+1;l++)
     {
      IMFEvaluateExpansionDerivative(data->E,vs,l,vdM);
      for(i=0;i<m;i++)
        Phi[i+n*j]+=dM[i]*vr[l+(k+1)*j];
     }
    for(i=0;i<k+1;i++)Phi[i+m+n*j]=vr[i+(k+1)*j];
   }
  if(verbose){printf("Tangent is\n");MFPrintNKMatrix(stdout,mPhi);fflush(stdout);}

  free(A);
  free(b);
  free(ipvt);
  MFFreeKVector(vs);
  MFFreeNVector(vM0);
  MFFreeNVector(vdM);
  free(U);
  free(V);
  free(sv);
  free(vr);
  free(work);

  return 1;
 }

int MFIMFSphereOnEProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;
  struct IMFSphereOnEData *data;

  data=(struct IMFSphereOnEData*)d;

 if(x==(double*)NULL)return IMFExpansionN(data->E);
  for(i=0;i<IMFExpansionN(data->E);i++)x[i]=MFNV_C(u,i);

  return 0;
 }

IMFExpansion IMFSphereOnExpansionGetLocal(MFAtlas c, MFNVector vu)
 {
  static char RoutineName[]={"IMFSphereOnExpansionGetLocal"};
  double *u;
  double *du;
  double *ddu;
  double *Du;
  double *DDu;
  MFNKMatrix mPhi;
  IMFExpansion result;
  int i,j,l,n,k,m;
  double *A=(double*)NULL;
  double *b=(double*)NULL;
  MFNVector vM0,vdM,vdMi,vdMj;
  MFKVector vs;
  double *s,*M0,*dM,*dMi,*dMj;
  int ii,jj,ll,kk;

  struct IMFSphereOnEData *data;

  data=(struct IMFSphereOnEData*)MFIMFGetData(MFAtlasMF(c));

/*
  second derivative is?
  third derivative is?
*/

  n=MFAtlasN(c);
  k=MFAtlasK(c);
  m=IMFExpansionN(data->E);

  result=IMFCreateExpansion(m,k);
  u=MFNV_CStar(vu);
  mPhi=MFIMFTangentSpace(MFAtlasMF(c),vu);

  du=MFNKM_CStar(mPhi);
  ddu=malloc(n*n*k*sizeof(double));
  Du=malloc(m*k*sizeof(double));
  DDu=malloc(m*m*k*sizeof(double));

  A=malloc(n*n*sizeof(double));
  b=malloc(n*sizeof(double));

  vs=MFCreateKVector(k+1);
  s=MFKV_CStar(vs);
  vM0=MFCreateNVector(m);
  M0=MFNV_CStar(vM0);
  vdM=MFCreateNVector(m);
  dM=MFNV_CStar(vdM);
  vdMi=MFCreateNVector(m);
  dMi=MFNV_CStar(vdMi);
  vdMj=MFCreateNVector(m);
  dMj=MFNV_CStar(vdMj);

  for(i=0;i<k+1;i++)s[i]=0.;
  IMFEvaluateExpansion(data->E,vs,vM0);

  for(i=0;i<k+1;i++)s[i]=u[m+i];

  for(ii=0;ii<k;ii++)
   {
    for(jj=0;jj<k;jj++)
     {
      b[0]=0.;
      for(ll=0;ll<k+1;ll++)
       {
        for(kk=0;kk<k+1;kk++)
         {
          IMFEvaluateExpansionDerivative(data->E,vs,ll,vdMi);
          IMFEvaluateExpansionDerivative(data->E,vs,kk,vdMj);
          for(i=0;i<m;i++)b[0]-=2*dMi[i]*du[ll+m+n*ii]*dMj[i]*du[kk+m+n*jj];
         }
       }
      for(i=1;i<k+1;i++)b[i]=0.;

      for(j=0;j<k+1;j++)
       {
        A[0+(k+1)*j]=0.;
        IMFEvaluateExpansionDerivative(data->E,vs,j,vdM);
        for(i=0;i<m;i++)A[0+(k+1)*j]+=2*dM[i]*(u[i]-M0[i]);
    
        for(l=1;l<k+1;l++)A[l+(k+1)*j]=du[j+m+n*(l-1)];
       }

      MFSolveFull(k+1,A,b);
      for(i=0;i<k+1;i++)ddu[i+m+n*(ii+k*jj)]=b[i];
     }
   }

  for(ii=0;ii<k;ii++)
   {
    for(jj=0;jj<k;jj++)
     {
      for(i=0;i<m;i++)
       {
        ddu[i+n*(ii+k*jj)]=0.;
        for(j=0;j<k+1;j++)
         {
          for(ll=0;ll<k+1;ll++)
           {
            IMFEvaluateExpansionDerivative(data->E,vs,ll,vdM);
            ddu[i+n*(ii+k*jj)]+=dM[i]*ddu[ll+m+n*(ii+k*jj)];
            for(kk=0;kk<k+1;kk++)
             {
              IMFEvaluateExpansionSecondDerivative(data->E,vs,ll,kk,vdM);
              ddu[i+n*(ii+k*jj)]+=dM[i]*du[ll+m+n*ii]*du[kk+m+n*jj];
             }
           }
         }
       }
     }
   }

  for(j=0;j<k;j++)
   {
    for(i=0;i<m;i++)
      Du[i+m*j]=du[i+n*j];
    for(l=0;l<k;l++)
     {
      for(i=0;i<m;i++)
        DDu[i+m*(j+k*l)]=du[i+n*(j+k*l)];
     }
   }

  IMFExpansionSetDerivatives(result,u,Du,DDu,(double*)NULL);

  MFFreeNKMatrix(mPhi);
  MFFreeNVector(vM0);
  MFFreeNVector(vdM);
  MFFreeNVector(vdMi);
  MFFreeNVector(vdMj);
  MFFreeKVector(vs);
  free(Du);
  free(DDu);
  free(ddu);
  free(A);
  free(b);
  return result;
 }

double MFScaleIMFSphereOnE(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  struct IMFSphereOnEData *data;
  double R,Rf;

  data=(struct IMFSphereOnEData*)d;

  R=IMFExpansionR(data->E,data->eps);
  Rf=IMFFlowR(data->F,data->eps,vu,mPhi);
  if(Rf<R)R=Rf;

  if(R>data->R||R!=R)R=data->R;

  return R;
 }

