/* 
    @(#)MF3dEdgeMF.c	1.2
    03/07/11 17:06:41
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   Sept. 27, 2002  modified MF3dPolygon   */ 

#include <MFImplicitMF.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static char MFEdgeMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFFreeEdgeIn3SpaceData(void*);
int MFProjectEdgeIn3Space(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentEdgeIn3Space(int,int,MFNVector,MFNKMatrix,void*);
double MFScaleEdgeIn3Space(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteEdgeIn3SpaceData(FILE*,void*);
MFImplicitMF MFReadEdgeIn3Space(FILE*);
int MFEdgeIn3SpaceProjectForBB(MFNVector,double*,void*);

struct MFEdgeIn3SpaceData
 {
  double r;
  double ox,oy,oz;
  double ax,ay,az;
 };

MFImplicitMF MFIMFCreateEdgeIn3SpaceWithRadius(double *v0, double *v1,double R)
 {
  static char RoutineName[]={"MFIMFCreateEdgeIn3SpaceWithRadius"};
  MFImplicitMF edge;
  MFNSpace space;
  struct MFEdgeIn3SpaceData *p;
  double d;
  int verbose;
  int i;

  verbose=0;
  if(verbose)printf("%s\n",RoutineName);

  edge=MFIMFCreateBaseClass(3,1,"EdgeIn3Space");

  space=MFCreateNSpace(3);
  MFIMFSetSpace(edge,space);
  MFFreeNSpace(space);

  p=(struct MFEdgeIn3SpaceData*)malloc(sizeof(struct MFEdgeIn3SpaceData)); /*done*/
  p->r=R;
  p->ox=v0[0];
  p->oy=v0[1];
  p->oz=v0[2];
  p->ax=v1[0]-v0[0];
  p->ay=v1[1]-v0[1];
  p->az=v1[2]-v0[2];
  d=1./sqrt(p->ax*p->ax+p->ay*p->ay+p->az*p->az);
  p->ax=p->ax*d;
  p->ay=p->ay*d;
  p->az=p->az*d;
  if(verbose){printf("%s, data=0x%8.8x, o=(%lf,%lf,%lf), n=(%lf,%lf,%lf)\n",RoutineName,p,p->ox,p->oy,p->oz,p->ax,p->ay,p->az);fflush(stdout);}

  MFIMFSetData(edge,(void*)p);
  MFIMFSetFreeData(edge,MFFreeEdgeIn3SpaceData);
  MFIMFSetProject(edge,MFProjectEdgeIn3Space);
  MFIMFSetProjectForBB(edge,MFEdgeIn3SpaceProjectForBB);
  MFIMFSetProjectForDraw(edge,MFEdgeIn3SpaceProjectForBB);
  MFIMFSetTangent(edge,MFTangentEdgeIn3Space);
  MFIMFSetScale(edge,MFScaleEdgeIn3Space);
  MFIMFSetR(edge,R);
  MFIMFSetWriteData(edge,MFWriteEdgeIn3SpaceData);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  return edge;
 }

void MFFreeEdgeIn3SpaceData(void *d)
 {
  static char RoutineName[]={"MFFreeEdgeIn3SpaceData"};
  struct MFEdgeIn3SpaceData *p;

  p=(struct MFEdgeIn3SpaceData*)d;
  free(p);
  return;
 }

int MFProjectEdgeIn3Space(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectEdgeIn3Space"};
  static int i;
  double a,b,nm;
  struct MFEdgeIn3SpaceData *p;
  double *u0,*Phi,*u;
  double s;
  int verbose;

  verbose=0;

  u0=MFNV_CStar(vu0);
  u=MFNV_CStar(vu);

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}
  p=(struct MFEdgeIn3SpaceData*)d;

  if(verbose){printf("%s, data=0x%8.8x, o=(%lf,%lf,%lf), n=(%lf,%lf,%lf)\n",RoutineName,p,p->ox,p->oy,p->oz,p->ax,p->ay,p->az);fflush(stdout);}

  s=(u0[0]-p->ox)*p->ax+(u0[1]-p->oy)*p->ay+(u0[2]-p->oz)*p->az;
  if(verbose){ printf("    u=");MFPrintNVector(stdout,vu);fflush(stdout);
  printf("    u=(%lf,%lf,%lf), s=%lf\n",u0[0],u0[1],u0[2],s);fflush(stdout);}

  u[0]=p->ox+s*p->ax;
  u[1]=p->oy+s*p->ay;
  u[2]=p->oz+s*p->az;

  *index=0;
  return 1;
 }

int MFTangentEdgeIn3Space(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentEdgeIn3Space"};
  static int i;
  struct MFEdgeIn3SpaceData *p;
  double *Phi,*u;

  Phi=MFNKM_CStar(mPhi);

  p=(struct MFEdgeIn3SpaceData*)d;
  Phi[0]=p->ax;
  Phi[1]=p->ay;
  Phi[2]=p->az;

  return 0;
 }

double MFScaleEdgeIn3Space(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFScaleEdgeIn3Space"};
  struct MFEdgeIn3SpaceData *p;

  p=(struct MFEdgeIn3SpaceData*)d;
  return p->r;
 }

void MFWriteEdgeIn3SpaceData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWriteEdgeIn3SpaceData"};
  struct MFEdgeIn3SpaceData *data;

  data=(struct MFEdgeIn3SpaceData*)d;
  fprintf(fid,"%lf %lf %lf %lf %lf %lf %lf\n",data->r,data->ox,data->oy,data->oz,data->ax,data->ay,data->az);
  fflush(fid);
  return;
 }

MFImplicitMF MFReadEdgeIn3Space(FILE *fid)
 {
  static char RoutineName[]={"MFReadEdgeIn3Space"};
  MFImplicitMF edge;
  struct MFEdgeIn3SpaceData *p;
  double d;
  int verbose;
  int i;
  double r=0.;
  double ox=0.;
  double oy=0.;
  double oz=0.;
  double ax=0.;
  double ay=0.;
  double az=0.;
  double v0[3];
  double v1[3];

  verbose=0;
  if(verbose)
   {
    printf("%s\n",RoutineName);
    fflush(stdout);
   }

  fscanf(fid,"%lf %lf %lf %lf %lf %lf %lf\n",&r,&ox,&oy,&oz,&ax,&ay,&az);

  v0[0]=ox;
  v0[1]=oy;
  v0[2]=oz;
  v1[0]=ox+ax;
  v1[1]=oy+ay;
  v1[2]=oz+az;

  edge=MFIMFCreateEdgeIn3SpaceWithRadius(v0,v1,r);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  return edge;
 }

int MFEdgeIn3SpaceProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;

  if(x==(double*)NULL)return 3;

  for(i=0;i<3;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

