/* 
    @(#)MF3dPolygonMF.c	1.11
    03/07/11 17:06:41
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory
*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */

#include <MFImplicitMF.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static char MFPolygonMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFFreePolygonIn3SpaceData(void*);
int MFProjectPolygonIn3Space(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentPolygonIn3Space(int,int,MFNVector,MFNKMatrix,void*);
double MFScalePolygonIn3Space(int,int,MFNVector,MFNKMatrix,void*);
void MFWritePolygonIn3SpaceData(FILE*,void*);
MFImplicitMF MFReadPolygonIn3Space(FILE*);
int MFPolygonIn3SpaceProjectForBB(MFNVector,double*,void*);

struct MFPolygonIn3SpaceData
 {
  double r;
  double ox,oy,oz;
  double ax,ay,az;
  double bx,by,bz;
  double nx,ny,nz;
 };

MFImplicitMF MFIMFCreatePolygonIn3SpaceWithRadius(int n,double *v,double R)
 {
  static char RoutineName[]={"MFIMFCreatePolygonIn3SpaceWithRadius"};
  MFImplicitMF polygon;
  MFNSpace space;
  struct MFPolygonIn3SpaceData *p;
  double d;
  int verbose;
  int i;

  verbose=0;
  if(verbose)
   {
    printf("%s\n",RoutineName);
    printf("%d\n",n);
    for(i=0;i<n;i++)
     printf("   %d (%lf,%lf,%lf)\n",i,v[3*i],v[3*i+1],v[3*i+2]);
    fflush(stdout);
   }

  polygon=MFIMFCreateBaseClass(3,2,"PolygonIn3Space");

  space=MFCreateNSpace(3);
  MFIMFSetSpace(polygon,space);
  MFFreeNSpace(space);

  p=(struct MFPolygonIn3SpaceData*)malloc(sizeof(struct MFPolygonIn3SpaceData)); /*done*/
  p->r=R;
  p->ox=v[0];
  p->oy=v[1];
  p->oz=v[2];
  if(verbose){printf("  o=(%lf,%lf,%lf)\n",p->ox,p->oy,p->oz);fflush(stdout);}
  p->ax=v[3]-v[0];
  p->ay=v[4]-v[1];
  p->az=v[5]-v[2];
  d=1./sqrt(p->ax*p->ax+p->ay*p->ay+p->az*p->az);
  p->ax=p->ax*d;
  p->ay=p->ay*d;
  p->az=p->az*d;
  if(verbose){printf("  a=(%lf,%lf,%lf)\n",p->ax,p->ay,p->az);fflush(stdout);}
  p->bx=v[6]-v[0];
  p->by=v[7]-v[1];
  p->bz=v[8]-v[2];
  d=p->ax*p->bx+p->ay*p->by+p->az*p->bz;
  p->bx-=p->ax*d;
  p->by-=p->ay*d;
  p->bz-=p->az*d;
  d=1./sqrt(p->bx*p->bx+p->by*p->by+p->bz*p->bz);
  p->bx=p->bx*d;
  p->by=p->by*d;
  p->bz=p->bz*d;
  if(verbose){printf("  b=(%lf,%lf,%lf)\n",p->bx,p->by,p->bz);fflush(stdout);}
  p->nx=p->ay*p->bz-p->az*p->by;
  p->ny=p->az*p->bx-p->ax*p->bz;
  p->nz=p->ax*p->by-p->ay*p->bx;
  if(verbose){printf("  n=(%lf,%lf,%lf)\n",p->nx,p->ny,p->nz);fflush(stdout);}

  MFIMFSetData(polygon,(void*)p);
  MFIMFSetFreeData(polygon,MFFreePolygonIn3SpaceData);
  MFIMFSetProject(polygon,MFProjectPolygonIn3Space);
  MFIMFSetTangent(polygon,MFTangentPolygonIn3Space);
  MFIMFSetScale(polygon,MFScalePolygonIn3Space);
  MFIMFSetR(polygon,R);
  MFIMFSetProjectForBB(polygon,MFPolygonIn3SpaceProjectForBB);
  MFIMFSetProjectForDraw(polygon,MFPolygonIn3SpaceProjectForBB);
  MFIMFSetWriteData(polygon,MFWritePolygonIn3SpaceData);

/* Test */

  if(verbose)
   {
    printf("Tangent [ %lf %lf %lf ]\n",
             (p->ax*p->ax+p->ay*p->ay+p->az*p->az),
             (p->ax*p->bx+p->ay*p->by+p->az*p->bz),
             (p->ax*p->nx+p->ay*p->ny+p->az*p->nz));
    printf("        [ %lf %lf %lf ]\n",
             (p->bx*p->ax+p->by*p->ay+p->bz*p->az),
             (p->bx*p->bx+p->by*p->by+p->bz*p->bz),
             (p->bx*p->nx+p->by*p->ny+p->bz*p->nz));
    printf("        [ %lf %lf %lf ]\n",
             (p->nx*p->ax+p->ny*p->ay+p->nz*p->az),
             (p->nx*p->bx+p->ny*p->by+p->nz*p->bz),
             (p->nx*p->nx+p->ny*p->ny+p->nz*p->nz));
    for(i=1;i<n;i++)
     {
      d=(p->nx*(v[3*i]-v[0])+p->ny*(v[3*i+1]-v[1])+p->nz*(v[3*i+2]-v[2]));
      printf("  (x[%d]-x[0]).n=%lf\n",i,d);
     }
   }

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  return polygon;
 }

void MFFreePolygonIn3SpaceData(void *d)
 {
  static char RoutineName[]={"MFFreePolygonIn3SpaceData"};
  struct MFPolygonIn3SpaceData *p;

  p=(struct MFPolygonIn3SpaceData*)d;
  free(p);
  return;
 }

int MFProjectPolygonIn3Space(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectPolygonIn3Space"};
  static int i;
  double a,b,nm;
  struct MFPolygonIn3SpaceData *p;
  double *u0,*Phi,*u;
  int verbose;

  verbose=0;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}
  p=(struct MFPolygonIn3SpaceData*)d;

  a=(u0[0]-p->ox)*p->ax+(u0[1]-p->oy)*p->ay+(u0[2]-p->oz)*p->az;
  b=(u0[0]-p->ox)*p->bx+(u0[1]-p->oy)*p->by+(u0[2]-p->oz)*p->bz;
  nm=(u0[0]-p->ox)*p->nx+(u0[1]-p->oy)*p->ny+(u0[2]-p->oz)*p->nz;

  u[0]=p->ox+a*p->ax+b*p->bx;
  u[1]=p->oy+a*p->ay+b*p->by;
  u[2]=p->oz+a*p->az+b*p->bz;

  if(verbose)
   {
    printf("  project (%lf,%lf,%lf)=(%lf,%lf,%lf)\n"
           "                       +%lf*(%lf,%lf,%lf)\n"
           "                       +%lf*(%lf,%lf,%lf)\n"
           "                       +%lf*(%lf,%lf,%lf)\n",u0[0],u0[1],u0[2],p->ox,p->oy,p->oz,a,p->ax,p->ay,p->az,b,p->bx,p->by,p->bz,nm,p->nx,p->ny,p->nz);
    printf("        =(%lf,%lf,%lf)\n",p->ox+a*p->ax+b*p->bx+nm*p->nx,p->oy+a*p->ay+b*p->by+nm*p->ny,p->oz+a*p->az+b*p->bz+nm*p->nz);
    printf("        =(%lf,%lf,%lf)\n",u[0],u[1],u[2]);
    printf("done %s\n",RoutineName);
    fflush(stdout);
   }
  *index=0;
  return 1;
 }

int MFTangentPolygonIn3Space(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentPolygonIn3Space"};
  static int i;
  struct MFPolygonIn3SpaceData *p;
  double *Phi,*u;

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  p=(struct MFPolygonIn3SpaceData*)d;
  Phi[0]=p->ax;
  Phi[1]=p->ay;
  Phi[2]=p->az;
  Phi[3]=p->bx;
  Phi[4]=p->by;
  Phi[5]=p->bz;
  if(0)
   {
    printf("%s, \n",RoutineName);
    printf("    [%lf,%lf]\n",Phi[0],Phi[3]);
    printf("    [%lf,%lf]\n",Phi[1],Phi[4]);
    printf("    [%lf,%lf]\n",Phi[2],Phi[5]);
   }

  return 1;
 }

double MFScalePolygonIn3Space(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFScalePolygonIn3Space"};
  struct MFPolygonIn3SpaceData *p;

  p=(struct MFPolygonIn3SpaceData*)d;
  return p->r;
 }

void MFWritePolygonIn3SpaceData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWritePolygonIn3SpaceData"};
  struct MFPolygonIn3SpaceData *data;

  data=(struct MFPolygonIn3SpaceData*)d;
  fprintf(fid,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",data->r,data->ox,data->oy,data->oz,data->ax,data->ay,data->az,data->bx,data->by,data->bz,data->nx,data->ny,data->nz);
  fflush(fid);
  return;
 }

MFImplicitMF MFReadPolygonIn3Space(FILE *fid)
 {
  static char RoutineName[]={"MFReadPolygonIn3Space"};
  MFImplicitMF polygon;
  struct MFPolygonIn3SpaceData *p;
  double d;
  int verbose;
  int i;
  double r=0.;
  double ox=0.;
  double oy=0.;
  double oz=0.;
  double ax=0.;
  double ay=0.;
  double az=0.;
  double bx=0.;
  double by=0.;
  double bz=0.;
  double nx=0.;
  double ny=0.;
  double nz=0.;
  double v[9];

  verbose=0;
  if(verbose)
   {
    printf("%s\n",RoutineName);
    fflush(stdout);
   }

  fscanf(fid,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",&r,&ox,&oy,&oz,&ax,&ay,&az,&bx,&by,&bz,&nx,&ny,&nz);

  v[0]=ox;
  v[1]=oy;
  v[2]=oz;
  v[3]=ox+ax;
  v[4]=oy+ay;
  v[5]=oz+az;
  v[6]=ox+bx;
  v[7]=oy+by;
  v[8]=oz+bz;

  polygon= MFIMFCreatePolygonIn3SpaceWithRadius(3,v,r);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  return polygon;
 }

int MFPolygonIn3SpaceProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;

  if(x==(double*)NULL)return 3;

  for(i=0;i<3;i++)x[i]=MFNV_C(u,i);

  return 0;
 }
