/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              October 6, 2000   Added Polyhedron    */
/*              September 22, 2005   Ported from MFCSGBallsNRegion    */

#include <stdio.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <stdlib.h>
#include <math.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNRegionErrorMsg[256]="";

double *MFNV_CStar(MFNVector);

int CSGBallsTest(MFNVector,void*);
void CSGBallsFree(void*);

struct MFCSGBallData
 {
  int n;
  int nc;
  double *x0;
  double *R;
  int *dir;
 };

MFNRegion MFNRegionCreateCSGBalls(int n,int nc,double *x0,double *R,int *dir)
 {
  static char RoutineName[]={"MFNRegionCreateCSGBalls"};
  MFNRegion csg;
  int i;
  struct MFCSGBallData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreateCSGBalls++;
  starttime=clock();
#endif

  csg=MFNRegionCreateBaseClass("CSGBalls");
  MFNRegionSetTest(csg,CSGBallsTest);
  MFNRegionSetFreeData(csg,CSGBallsFree);

  data=(struct MFCSGBallData*)malloc(sizeof(struct MFCSGBallData));
  if(data==(struct MFCSGBallData*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCSGBallData));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(csg);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreateCSGBalls+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data->n=n;
  data->nc=nc;
  data->x0=(double*)malloc(n*nc*sizeof(double));
  for(i=0;i<n*nc;i++)(data->x0)[i]=x0[i];
  data->R=(double*)malloc(nc*sizeof(double));
  for(i=0;i<nc;i++)(data->R)[i]=R[i];
  data->dir=(int*)malloc(nc*sizeof(int));
  for(i=0;i<nc;i++)(data->dir)[i]=dir[i];

  MFNRegionSetData(csg,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreateCSGBalls+=clock()-starttime;
#endif
  return(csg);
 }

int CSGBallsTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"CSGBallsTest"};
  struct MFCSGBallData *d;
  double r;
  double *x;
  int rc;
  int i,j;
  static int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledCSGBallsTest++;
  starttime=clock();
#endif
  d=(struct MFCSGBallData*)data;
  x=MFNV_CStar(v);

  rc=1;
  for(i=0;i<d->nc;i++)
   {
    r=0.;
    for(j=0;j<d->n;j++)r+=(x[j]-d->x0[j+d->n*i])*(x[j]-d->x0[j+d->n*i]);
    rc=rc&&d->dir[i]*(d->R[i]*d->R[i]-r*r)<0.;
   }

#ifdef MFTIMINGS
    MFTimeCSGBallsTest+=clock()-starttime;
#endif
  return rc;
 }

void CSGBallsFree(void *data)
 {
  static char RoutineName[]={"CSGBallsFree"};
  struct MFCSGBallData *d;

  d=(struct MFCSGBallData*)data;

  free(d->x0);
  free(d->R);
  free(d->dir);
  free(d);

  return;
 }
