/* 
    %W%
    %D% %T%
   

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFAtlasFriends.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <MFEnumDualPolytope.h>
#include <math.h>
#include <sh.h>
#include <MFDraw.h>
#include <MFPrint.h>

int MFComputePointsOn3dPolygon(double r,int nV,double *V,int nPt0,double *Pt0,double *R0,double **Pt,double **R)
 {
  MFImplicitMF M;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  MFNVector u;
  MFNKMatrix Phi;
  double Rad;
  int n;
  int i,j;
  double delta=0;
  int chart;
  int kkk;
  int verbose;
  MFPolytope P;
  double rho,Ri,Rj;
  MFEnumDualPolytope Q;

  verbose=0;

  if(verbose)
   {
    printf("---begin-------------------------------------------------------------------\n");
    printf("MFComputePointsOn3dPolygon(r=%lf,nV=%d,V,nPt0=%d,Pt0,Pt);\n",r,nV,nPt0);
   }

  Omega=MFNRegionCreatePolygonal3dRegion(nV,V);

  M=MFIMFCreatePolygonIn3SpaceWithRadius(nV,V,r);

  S=MFCreateAtlas(M);

  kkk=0;
  for(i=0;i<nPt0;i++)
   {
    u0=MFCreateNVector(3);
    MFNVSetC(u0,0,Pt0[3*i  ]);
    MFNVSetC(u0,1,Pt0[3*i+1]);
    MFNVSetC(u0,2,Pt0[3*i+2]);

    if(verbose){printf("%d) Add initial point to Atlas ",kkk);MFPrintNVector(stdout,u0);printf("\n");fflush(stdout);}

/*  MFAtlasAddChart(S,u0); */
    MFAtlasAddChartWithAll(S,u0,(MFNKMatrix)NULL,R0[i]); 
    kkk++;
    MFFreeNVector(u0);
   }
  if(verbose){printf("Done adding inital points\n");fflush(stdout);}

  u=MFCreateNVector(3);
  while((chart=MFAtlasPointOnBoundaryInsideRegion(S,Omega,u,&Phi,&delta))>-1)
   {
    if(verbose){printf("%d) Adding point ",kkk);MFPrintNVector(stdout,MFAtlasCenterOfChart(S,chart));printf("\n");fflush(stdout);}
    kkk++;
    Rad=MFIMFScale(MFAtlasMF(S),u,Phi);
    MFAtlasAddChartWithAll(S,u,Phi,Rad);
    MFFreeNVector(u);
    MFFreeNKMatrix(Phi);
    u=MFCreateNVector(3);
   }
  if(verbose){printf("No more points on boundary\n");fflush(stdout);}

/*MFDrawClear();
  MFDrawAtlasOnce(S);
  Q=MFEnumDualOfAtlas(S);
  MFDrawEnumDualPolytope(Q);
  MFFreeEnumDualPolytope(Q);*/

/*DrawEdges();
  shSetOutputFormat("tiff");
  shSetOutputFilename("Filled");
  MFDrawDisplay();MFDrawClear();*/

  MFFreeNVector(u);

  n=MFAtlasNumberOfCharts(S)-nPt0;
  if(n>0)
   {
  (*Pt)=(double*)realloc((void*)(*Pt),3*n*sizeof(double));
  (*R)=(double*)realloc((void*)(*R),n*sizeof(double));
    for(i=0;i<n;i++)
     {
      u=MFAtlasCenterOfChart(S,i+nPt0);
      (*Pt)[3*i  ]=MFNV_C(u,0);
      (*Pt)[3*i+1]=MFNV_C(u,1);
      (*Pt)[3*i+2]=MFNV_C(u,2);
      (*R)[i]=0.;
      P=MFChartPolytope(MFAtlasChart(S,i+nPt0));
      Ri=MFAtlasChartRadius(S,i);
      for(j=0;j<MFPolytopeNumberOfVertices(P);j++)
       {
        Rj=MFPolytopeRadiusOfVertex(P,j);
        rho=sqrt(2*Rj*Rj-Ri*Ri);
        if(rho>(*R)[i])(*R)[i]=rho;
       }
     }
   }

  MFFreeAtlas(S);
  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  if(verbose)printf("----end--------------------------------------------------------------------\n");

  return(n);
 }
