/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */

#include <MFError.h>
#include <MFNVector.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>

#ifndef DBL_QNAN
#define DBL_QNAN 1.e300
#endif

struct MFDenseNVectorData
 {
  int nC;
  double *data;
  int wrapped;
 };

void MFFreeDenseNVectorData(void*);
int MFDenseNVGetNC(void*);
double MFDenseNVGetC(int,void*);
void MFDenseNVSetC(int,double,void*);
void MFDenseNVDiff(void*,void*,void*);
void MFDenseNVAdd(void*,void*,void*);
void MFWriteDenseNVectorData(FILE*,void*);
MFNVector MFReadDenseNVector(FILE*);
void MFPrintDenseNVector(FILE*,void*);
MFNVector MFCloneDenseNVector(void*);

static char MFNVectorErrorMsg[256]="";

MFNVector MFCreateNVectorWithData(int n,double *vl)
 {
  static char RoutineName[]={"MFCreateNVectorWithData"};
  struct MFDenseNVectorData *data;
  int i;
  MFNVector this;

  if(n<1)
   {
    sprintf(MFNVectorErrorMsg,"Length of Vector %d (argument 1) is Illegal. Must be positive.",n);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNVector+=clock()-starttime;
#endif
    return((MFNVector)NULL);
   }

  data=(struct MFDenseNVectorData*)malloc(sizeof(struct MFDenseNVectorData)); /*done*/
  if(data==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFDenseNVectorData));
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return(this);
   }

  data->nC=n;
  data->data=(double*)malloc(n*sizeof(double)); /*done*/
  if(n>0&&data->data==(double*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return(this);
   }

  if(n>0&&vl!=(double*)NULL)
   {
    for(i=0;i<n;i++)(data->data)[i]=vl[i];
   }else if(vl==(double*)NULL)
   {
    for(i=0;i<n;i++)(data->data)[i]=0.;
   }

  data->wrapped=0;

  this=MFCreateNVectorBaseClass("DENSE");

  MFNVectorSetData(this,data);
  MFNVectorSetFreeData(this,MFFreeDenseNVectorData);
  MFNVectorSetWriteData(this,MFWriteDenseNVectorData);
  MFNVectorSetGetNC(this,MFDenseNVGetNC);
  MFNVectorSetGetC(this,MFDenseNVGetC);
  MFNVectorSetSetC(this,MFDenseNVSetC);
  MFNVectorSetDiff(this,MFDenseNVDiff);
  MFNVectorSetAdd(this,MFDenseNVAdd);
  MFNVectorSetClone(this,MFCloneDenseNVector);
  MFNVectorSetPrint(this,MFPrintDenseNVector);

  return this;
 }

MFNVector MFCreateWrappedNVector(int n,double *vl)
 {
  static char RoutineName[]={"MFCreateWrappedNVector"};
  struct MFDenseNVectorData *data;
  int i;
  MFNVector this;

  this=(MFNVector)NULL;

  if(n<1)
   {
    sprintf(MFNVectorErrorMsg,"Length of Vector %d (argument 1) is Illegal. Must be positive.",n);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNVector+=clock()-starttime;
#endif
    return((MFNVector)NULL);
   }

  if(vl==(double*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to data (argument 2) is Illegal. Must be non-NULL.");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return(this);
   }

  data=(struct MFDenseNVectorData*)malloc(sizeof(struct MFDenseNVectorData)); /*done*/
  if(data==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFDenseNVectorData));
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return(this);
   }

  data->nC=n;
  data->data=vl;
  data->wrapped=1;

  this=MFCreateNVectorBaseClass("DENSE");

  MFNVectorSetData(this,data);
  MFNVectorSetFreeData(this,MFFreeDenseNVectorData);
  MFNVectorSetWriteData(this,MFWriteDenseNVectorData);
  MFNVectorSetGetNC(this,MFDenseNVGetNC);
  MFNVectorSetGetC(this,MFDenseNVGetC);
  MFNVectorSetSetC(this,MFDenseNVSetC);
  MFNVectorSetDiff(this,MFDenseNVDiff);
  MFNVectorSetAdd(this,MFDenseNVAdd);
  MFNVectorSetClone(this,MFCloneDenseNVector);
  MFNVectorSetPrint(this,MFPrintDenseNVector);

  return this;
 }

MFNVector MFCreateNVector(int n)
 {
  static char RoutineName[]={"MFCreateNVector"};
  int i;
  MFNVector this;

  this=MFCreateNVectorWithData(n,(double*)NULL);

  return this;
 }

void MFFreeDenseNVectorData(void *data)
 {
  static char RoutineName[]={"MFFreeDenseNVectorData"};
  struct MFDenseNVectorData *this;

  this=(struct MFDenseNVectorData*)data;

  if(this==(struct MFDenseNVectorData*)NULL)return;
  if(!this->wrapped&&this->data!=(double*)NULL)free(this->data);
  free(this);

  return;
 }

int MFDenseNVGetNC(void *data)
 {
  static char RoutineName[]={"MFNV_NC"};
  struct MFDenseNVectorData *this;

  this=(struct MFDenseNVectorData*)data;

  if(this==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return(this->nC);
 }

double MFDenseNVGetC(int i,void *data)
 {
  static char RoutineName[]={"MFDenseNVGetC"};
  struct MFDenseNVectorData *this;
  double result;

  this=(struct MFDenseNVectorData*)data;

  if(this==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return DBL_QNAN;
   }

  if(i<0|| !(i<this->nC))
   {
    sprintf(MFNVectorErrorMsg,"Coordinate %d (argument 1) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return DBL_QNAN;
   }

  result=this->data[i];

  return result;
 }

void MFDenseNVSetC(int i,double vl,void *data)
 {
  static char RoutineName[]={"MFDenseNVSetC"};
  struct MFDenseNVectorData *this;

  this=(struct MFDenseNVectorData*)data;

  if(this==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0|| !(i<this->nC))
   {
    sprintf(MFNVectorErrorMsg,"Coordinate %d (argument 1) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->data[i]=vl;

  return;
 }

void MFDenseNVDiff(void *adata,void *bdata, void *cdata)
 {
  static char RoutineName[]={"MFDenseNVDiff"};
  int i,n;
  struct MFDenseNVectorData *a;
  struct MFDenseNVectorData *b;
  struct MFDenseNVectorData *c;

  a=(struct MFDenseNVectorData*)adata;
  if(a==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  b=(struct MFDenseNVectorData*)bdata;
  if(b==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  c=(struct MFDenseNVectorData*)cdata;
  if(c==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFNVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
   c->data[i]=a->data[i]-b->data[i];

  return;
 }

void MFDenseNVAdd(void *adata,void *bdata,void *cdata)
 {
  static char RoutineName[]={"MFDenseNVAdd"};
  int i,n;
  struct MFDenseNVectorData *a;
  struct MFDenseNVectorData *b;
  struct MFDenseNVectorData *c;

  a=(struct MFDenseNVectorData*)adata;
  if(a==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  b=(struct MFDenseNVectorData*)bdata;
  if(b==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  c=(struct MFDenseNVectorData*)cdata;
  if(c==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFNVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNVDiff+=clock()-starttime;
#endif
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
   c->data[i]=a->data[i]+b->data[i];

  return;
 }

double *MFNV_CStar(MFNVector this)
 {
  static char RoutineName[]={"MFNV_CStar"};
  struct MFDenseNVectorData *u;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNV_CStar++;
  starttime=clock();
#endif
  if(this==(MFNVector)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector (argument 1) is NULL");
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNV_CStar+=clock()-starttime;
#endif
    return (double*)NULL;
   }

  if(strcmp(MFNVGetId(this),"DENSE"))
   {
    sprintf(MFNVectorErrorMsg,"Trying to get dense arrays from non-dense Vector type \"%s\"",MFNVGetId(this));
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNV_CStar+=clock()-starttime;
#endif
    return (double*)NULL;
   }

  u=(struct MFDenseNVectorData*)MFNVectorGetData(this);

#ifdef MFTIMINGS
    MFTimeMFNV_CStar+=clock()-starttime;
#endif
  return(u->data);
 }

void MFWriteDenseNVectorData(FILE *fid,void *data)
 {
  static char RoutineName[]={"MFWriteDenseNVectorData"};
  int i;
  struct MFDenseNVectorData *u;

  u=(struct MFDenseNVectorData*)data;
  if(u==(struct MFDenseNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  fprintf(fid,"%d\n",u->nC);
  for(i=0;i<u->nC;i++)
   {
    if(i>0)fprintf(fid," ");
    fprintf(fid,"%lf",u->data[i]);
   }
  fprintf(fid,"\n");

  return;
 }

MFNVector MFReadDenseNVector(FILE *fid)
 {
  static char RoutineName[]={"MFReadDenseNVector"};
  int n;
  double *vl;
  MFNVector this;
  int i;

  fscanf(fid,"%d\n",&n);

  vl=(double*)malloc(n*sizeof(double)); /*done*/
  if(n>0&&vl==(double*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  for(i=0;i<n;i++)
   {
    if(i>0)fscanf(fid," ");
    fscanf(fid,"%lf",&(vl[i]));
   }
  fscanf(fid,"\n");

  this=MFCreateNVectorWithData(n,vl);

  free(vl);

  return this;
 }

MFNVector MFCloneDenseNVector(void *data)
 {
  static char RoutineName[]={"MFCloneDenseNVector"};
  struct MFDenseNVectorData *u;

  u=(struct MFDenseNVectorData*)data;

  return MFCreateNVectorWithData(u->nC,u->data);
 }

void MFPrintDenseNVector(FILE *fid,void *data)
 {
  static char RoutineName[]={"MFPrintDenseNVector"};
  int i;
  struct MFDenseNVectorData *u;

  u=(struct MFDenseNVectorData*)data;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(data==(void*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"data (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(u->nC<1)return;

  if(u->nC<10)
   {
    fprintf(fid,"(%lf",(u->data)[0]);
    for(i=1;i<u->nC;i++)
      fprintf(fid,",%lf",(u->data)[i]);
    fprintf(fid,")");
   }else{
    fprintf(fid,"(%lf",(u->data)[0]);
    for(i=1;i<5;i++)
      fprintf(fid,",%lf",(u->data)[i]);
    fprintf(fid,",...");
    for(i=u->nC-5;i<u->nC;i++)
      fprintf(fid,",%lf",(u->data)[i]);
    fprintf(fid,")");
   }

  return;
 }
