/* 
    @(#)MFDrawChartState.c	1.5
    02/04/19 16:29:40
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFDraw.h>
#include <MFDrawChartState.h>
#include <MFChart.h>
#include <MFPolytope.h>
#include <sh.h>
#include <math.h>
#include <MFError.h>
#include <stdlib.h>


void MFSetError(int,char*,char*,int,char*);
static char MFChartStateErrorMsg[256]="";

struct MFChartStateSt
 {
  int r[128],g[128],b[128];
  MFChart chart;
  int nP;
  int mP;
  int *nV;
  float **pX;
  float **pY;
  float **pZ;
  int nNP;
  int mNP;
  int *nNV;
  float **pNX;
  float **pNY;
  float **pNZ;
  float **pNNrm;
  int *pc;
  int nL;
  float *lX;
  float *lY;
  float *lZ;
  int nPlns;
  int *gInd;
  int nVert;
 };

int MFDrawNCharts=0;
int MFDrawMCharts=0;
MFChartState *MFDrawChartList=(MFChartState*)NULL;

void MFFreeChartState(MFChartState);

MFChartState MFCreateChartState(MFChart chart)
 {
  static char RoutineName[]={"MFCreateChartState"};
  MFChartState this;
  MFPolytope P;
  int color;
  int zero=0;
  int half=128;
  int full=255;
  int i;

  this=(MFChartState)malloc(sizeof(struct MFChartStateSt)); /*done*/
  if(this==(MFChartState)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFChartStateSt));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return (MFChartState)NULL;
   }

  this->chart=chart;
  MFRefChart(chart);

  this->nP=0;
  this->mP=0;
  this->nV=(int*)NULL;
  this->pX=(float**)NULL;
  this->pY=(float**)NULL;
  this->pZ=(float**)NULL;

  this->nNP=0;
  this->mNP=0;
  this->nNV=(int*)NULL;
  this->pNX=(float**)NULL;
  this->pNY=(float**)NULL;
  this->pNZ=(float**)NULL;
  this->pNNrm=(float**)NULL;
  this->pc=(int*)NULL;

  this->nL=0;
  this->lX=(float*)NULL;
  this->lY=(float*)NULL;
  this->lZ=(float*)NULL;

  this->r[0]=half;
  this->g[0]=half;
  this->b[0]=half;

  P=MFChartPolytope(chart);
  this->nVert=MFPolytopeNumberOfVertices(P);
  this->nPlns=MFPolytopeNumberOfFaces(P);
  this->gInd=(int*)malloc(this->nPlns*sizeof(int)); /*done*/
  if(this->nPlns>0&&this->gInd==(int*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",this->nPlns*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return (MFChartState)NULL;
   }
  for(i=0;i<this->nPlns;i++)
   this->gInd[i]=MFPolytopeFaceIndex(P,i);

  return(this);
 }

MFChartState MFAddChartState(MFChart chart)
 {
  static char RoutineName[]={"MFAddChartState"};
  if(MFDrawNCharts==MFDrawMCharts)
   {
    MFDrawMCharts+=10;
    MFDrawChartList=(MFChartState*)realloc((void*)MFDrawChartList,MFDrawMCharts*sizeof(MFChartState)); /*done*/
    if(MFDrawChartList==(MFChartState*)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",MFDrawMCharts*sizeof(MFChartState));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return (MFChartState)NULL;
     }
   }
  MFDrawChartList[MFDrawNCharts]=MFCreateChartState(chart);
  MFDrawNCharts++;

  return MFDrawChartList[MFDrawNCharts-1];
 }

int MFNChartStates()
 {
  static char RoutineName[]={"MFNChartStates"};
  return MFDrawNCharts;
 }

void MFClearChartState(MFChartState this)
 {
  static char RoutineName[]={"MFClearChartState"};
  MFPolytope P;
  int i;

  if(this->nP>0)
   {
    if(this->pX!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pX)[i]!=(float*)NULL)free((this->pX)[i]);
      free(this->pX);this->pX=(float**)NULL;
     }
    if(this->pY!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pY)[i]!=(float*)NULL)free((this->pY)[i]);
      free(this->pY);this->pY=(float**)NULL;
     }
    if(this->pZ!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pZ)[i]!=(float*)NULL)free((this->pZ)[i]);
      free(this->pZ);this->pZ=(float**)NULL;
     }
    if(this->nV!=(int*)NULL){free(this->nV);this->nV=(int*)NULL;}
   }
  this->nP=0;
  this->mP=0;

  if(this->nNP>0)
   {
    if(this->pNX!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNX)[i]!=(float*)NULL)free((this->pNX)[i]);
      free(this->pNX);this->pNX=(float**)NULL;
     }
    if(this->pNY!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNY)[i]!=(float*)NULL)free((this->pNY)[i]);
      free(this->pNY);this->pNY=(float**)NULL;
     }
    if(this->pNZ!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNZ)[i]!=(float*)NULL)free((this->pNZ)[i]);
      free(this->pNZ);this->pNZ=(float**)NULL;
     }
    if(this->pNNrm!=(float**)NULL)
     {
      for(i=0;i<this->nNP;i++)
       if((this->pNNrm)[i]!=(float*)NULL)free((this->pNNrm)[i]);
      free(this->pNNrm);this->pNNrm=(float**)NULL;
     }
    if(this->nNV!=(int*)NULL){free(this->nNV);this->nNV=(int*)NULL;}
    if(this->pc!=(int*)NULL){free(this->pc);this->pc=(int*)NULL;}
   }
  this->nNP=0;
  this->mNP=0;

  this->nL=0;
  if(this->lX!=(float*)NULL){free(this->lX);this->lX=(float*)NULL;}
  if(this->lY!=(float*)NULL){free(this->lY);this->lY=(float*)NULL;}
  if(this->lZ!=(float*)NULL){free(this->lZ);this->lZ=(float*)NULL;}


  P=MFChartPolytope(this->chart);
  this->nVert=MFPolytopeNumberOfVertices(P);
  this->nPlns=MFPolytopeNumberOfFaces(P);
  if(this->gInd!=(int*)NULL)free(this->gInd);
  this->gInd=(int*)malloc(this->nPlns*sizeof(int)); /*done*/
  if(this->gInd==(int*)NULL&&this->nPlns>0)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",this->nPlns*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  for(i=0;i<this->nPlns;i++)
   this->gInd[i]=MFPolytopeFaceIndex(P,i);

  return;
 }

void MFFreeChartState(MFChartState this)
 {
  static char RoutineName[]={"MFFreeChartState"};
  MFPolytope P;
  int i;

  if(this->nP>0)
   {
    if(this->pX!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pX)[i]!=(float*)NULL)free((this->pX)[i]);
      free(this->pX);this->pX=(float**)NULL;
     }
    if(this->pY!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pY)[i]!=(float*)NULL)free((this->pY)[i]);
      free(this->pY);this->pY=(float**)NULL;
     }
    if(this->pZ!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pZ)[i]!=(float*)NULL)free((this->pZ)[i]);
      free(this->pZ);this->pZ=(float**)NULL;
     }
    if(this->nV!=(int*)NULL){free(this->nV);this->nV=(int*)NULL;}
   }
  this->nP=0;
  this->mP=0;

  if(this->nNP>0)
   {
    if(this->pNX!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNX)[i]!=(float*)NULL)free((this->pNX)[i]);
      free(this->pNX);this->pNX=(float**)NULL;
     }
    if(this->pNY!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNY)[i]!=(float*)NULL)free((this->pNY)[i]);
      free(this->pNY);this->pNY=(float**)NULL;
     }
    if(this->pNZ!=(float**)NULL)
     {
      for(i=0;i<this->nP;i++)
       if((this->pNZ)[i]!=(float*)NULL)free((this->pNZ)[i]);
      free(this->pNZ);this->pNZ=(float**)NULL;
     }
    if(this->pNNrm!=(float**)NULL)
     {
      for(i=0;i<this->nNP;i++)
       if((this->pNNrm)[i]!=(float*)NULL)free((this->pNNrm)[i]);
      free(this->pNNrm);this->pNNrm=(float**)NULL;
     }
    if(this->nNV!=(int*)NULL){free(this->nNV);this->nNV=(int*)NULL;}
    if(this->pc!=(int*)NULL){free(this->pc);this->pc=(int*)NULL;}
   }
  this->nNP=0;
  this->mNP=0;

  this->nL=0;
  if(this->lX!=(float*)NULL){free(this->lX);this->lX=(float*)NULL;}
  if(this->lY!=(float*)NULL){free(this->lY);this->lY=(float*)NULL;}
  if(this->lZ!=(float*)NULL){free(this->lZ);this->lZ=(float*)NULL;}

  this->nVert=0;
  this->nPlns=0;
  if(this->gInd!=(int*)NULL)free(this->gInd);

  free(this);

  return;
 }

void MFSetChartStateColor(MFChartState this,int c,int r,int g,int b)
 {
  static char RoutineName[]={"MFSetChartStateColor"};
  this->r[c]=r;
  this->g[c]=g;
  this->b[c]=b;
  return;
 }

void MFGetChartStateColor(MFChartState this,int c,int *r,int *g,int *b)
 {
  static char RoutineName[]={"MFGetChartStateColor"};
  *r=this->r[c];
  *g=this->g[c];
  *b=this->b[c];
  return;
 }

void MFChartStateAddPolygon(MFChartState this,int n,float *x,float *y,float *z,int c)
 {
  static char RoutineName[]={"MFChartStateAddPolygon"};
  int i;

  for(i=0;i<n-1;i++)
   if(0 && sqrt((x[i+1]-x[i])*(x[i+1]-x[i])
          +(y[i+1]-y[i])*(y[i+1]-y[i])
          +(z[i+1]-z[i])*(z[i+1]-z[i]))>1.)return;

  if(this->nP+1>this->mP)
   {
    this->mP+=1000;
    this->nV=(int*)realloc((void*)(this->nV),this->mP*sizeof(int)); /*done*/
    if(this->nV==(int*)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mP*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pX=(float**)realloc((void*)(this->pX),this->mP*sizeof(float*)); /*done*/
    if(this->pX==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pY=(float**)realloc((void*)(this->pY),this->mP*sizeof(float*)); /*done*/
    if(this->pY==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pZ=(float**)realloc((void*)(this->pZ),this->mP*sizeof(float*)); /*done*/
    if(this->pZ==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pc=(int*)malloc(this->mP*sizeof(int)); /*done*/
    if(this->pc==(int*)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
   }

  this->nV[this->nP]=n;
  this->pX[this->nP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pX[this->nP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->pY[this->nP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pY[this->nP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->pZ[this->nP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pZ[this->nP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  for(i=0;i<n;i++)
   {
    (this->pX[this->nP])[i]=x[i];
    (this->pY[this->nP])[i]=y[i];
    (this->pZ[this->nP])[i]=z[i];
   }
  this->pc[this->nP]=c;
  this->nP++;
  return;
 }

void MFChartStateAddPolygonWithNormal(MFChartState this,int n,float *x,float *y,float *z,float *nrm, int c)
 {
  static char RoutineName[]={"MFChartStateAddPolygonWithNormal"};
  int i;

  for(i=0;i<n-1;i++)
   if(0 && sqrt((x[i+1]-x[i])*(x[i+1]-x[i])
          +(y[i+1]-y[i])*(y[i+1]-y[i])
          +(z[i+1]-z[i])*(z[i+1]-z[i]))>1.)return;

  if(this->nNP+1>this->mNP)
   {
    this->mNP+=1000;
    this->nNV=(int*)realloc((void*)(this->nNV),this->mNP*sizeof(int)); /*done*/
    if(this->nNV==(int*)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pNX=(float**)realloc((void*)(this->pNX),this->mNP*sizeof(float*)); /*done*/
    if(this->pNX==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pNY=(float**)realloc((void*)(this->pNY),this->mNP*sizeof(float*)); /*done*/
    if(this->pNY==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pNZ=(float**)realloc((void*)(this->pNZ),this->mNP*sizeof(float*)); /*done*/
    if(this->pNZ==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pNNrm=(float**)realloc((void*)(this->pNNrm),this->mNP*sizeof(float*)); /*done*/
    if(this->pNNrm==(float**)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(float*));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
    this->pc=(int*)realloc((void*)(this->pc),this->mNP*sizeof(int)); /*done*/
    if(this->pc==(int*)NULL)
     {
      sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",this->mNP*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
      return;
     }
   }

  this->nNV[this->nNP]=n;
  this->pNX[this->nNP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pNX[this->nNP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->pNY[this->nNP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pNY[this->nNP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->pNZ[this->nNP]=(float*)malloc(n*sizeof(float)); /*done*/
  if(this->pNZ[this->nNP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->pNNrm[this->nNP]=(float*)malloc(3*n*sizeof(float)); /*done*/
  if(this->pNNrm[this->nNP]==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to allocate %d bytes",3*n*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  for(i=0;i<n;i++)
   {
    (this->pNX[this->nNP])[i]=x[i];
    (this->pNY[this->nNP])[i]=y[i];
    (this->pNZ[this->nNP])[i]=z[i];
    (this->pNNrm[this->nNP])[3*i]=nrm[3*i];
    (this->pNNrm[this->nNP])[3*i+1]=nrm[3*i+1];
    (this->pNNrm[this->nNP])[3*i+1]=nrm[3*i+2];
   }
  this->pc[this->nP]=c;
  this->nNP++;
  return;
 }

void MFChartStateAddLine(MFChartState this,float x0,float y0,float z0,float x1,float y1,float z1)
 {
  static char RoutineName[]={"MFChartStateAddLine"};
  if(0 && sqrt((x1-x0)*(x1-x0)+(y1-y0)*(y1-y0)+(z1-z0)*(z1-z0))>1.)return;

  this->lX=(float*)realloc((void*)(this->lX),2*(this->nL+1)*sizeof(float)); /*done*/
  if(this->lX==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",2*(this->nL+1)*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->lY=(float*)realloc((void*)(this->lY),2*(this->nL+1)*sizeof(float)); /*done*/
  if(this->lY==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",2*(this->nL+1)*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->lZ=(float*)realloc((void*)(this->lZ),2*(this->nL+1)*sizeof(float)); /*done*/
  if(this->lZ==(float*)NULL)
   {
    sprintf(MFChartStateErrorMsg,"Out of memory, trying to reallocate %d bytes",2*(this->nL+1)*sizeof(float));
    printf("%s -- %s\n",RoutineName,MFChartStateErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartStateErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->lX[2*this->nL]=x0;
  this->lY[2*this->nL]=y0;
  this->lZ[2*this->nL]=z0;
  this->lX[2*this->nL+1]=x1;
  this->lY[2*this->nL+1]=y1;
  this->lZ[2*this->nL+1]=z1;
  this->nL++;
  return;
 }

MFChartState MFGetChartState(int i)
 {
  static char RoutineName[]={"MFGetChartState"};
  return MFDrawChartList[i];
 }

int MFChartStateChanged(MFChartState this)
 {
  static char RoutineName[]={"MFChartStateChanged"};
  MFPolytope P;
  int i;

  P=MFChartPolytope(this->chart);

  if(MFPolytopeNumberOfVertices(P)!=this->nVert)return 1;
  if(MFPolytopeNumberOfFaces(P)!=this->nPlns)return 1;

  for(i=0;i<this->nPlns;i++)
    if(MFPolytopeFaceIndex(P,i)!=this->gInd[i])return 1;

  return 0;
 }

void MFDrawChartFromState(MFChartState this)
 {
  static char RoutineName[]={"MFDrawChartFromState"};
  int i,n;
  int full=255;
  int zero=0;

  for(i=0;i<this->nP;i++)
   {
    shtric(&(this->r[this->pc[i]]),&(this->g[this->pc[i]]),&(this->b[this->pc[i]]));
    shpg(this->nV+i,(this->pX)[i],(this->pY)[i],(this->pZ)[i]);
   }

  shpgc(&full,&full,&zero);
  shpec(&full,&full,&zero);
  shbgc(&full,&full,&zero);
  shbec(&full,&full,&zero);
  for(i=0;i<this->nNP;i++)
   shpg(this->nNV+i,(this->pNX)[i],(this->pNY)[i],(this->pNZ)[i]);
/* shpgnrm(this->nNV+i,(this->pNX)[i],(this->pNY)[i],(this->pNZ)[i],(this->pNNrm)[i]);*/

  shlinc(&zero,&zero,&zero);
  for(i=0;i<this->nL;i++)
   shline(this->lX+2*i,this->lY+2*i,this->lZ+2*i,this->lX+2*i+1,this->lY+2*i+1,this->lZ+2*i+1);

  return;
 }

void MFFreeDrawCharts()
 {
  static char RoutineName[]={"MFFreeDrawCharts"};
  int i;

  for(i=0;i<MFDrawNCharts;i++)MFFreeChartState(MFDrawChartList[i]);
  if(MFDrawChartList!=(MFChartState*)NULL)free(MFDrawChartList);
  MFDrawNCharts=0;
  MFDrawMCharts=0;
  MFDrawChartList=(MFChartState*)NULL;

  return;
 }
