/* 
    @(#)MFFlat.c	1.14
    03/07/11 17:06:41
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   April 27, 1999                     */

#include <MFImplicitMF.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static char MFFlatMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFFreeFlatData(void*);
int MFProjectFlat(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentFlat(int,int,MFNVector,MFNKMatrix,void*);
double MFScaleFlat(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteFlatData(FILE*,void*);
MFImplicitMF MFReadFlat(FILE*);

int MFFlatProjectToSave(MFNVector,double*,void *d);
int MFFlatProjectToDraw(MFNVector,double*,void *d);
int MFFlatProjectForBB(MFNVector,double*,void *d);

struct MFFlatData
 {
  int n;
  int k;
  double *o;
  double *v;
 };

MFImplicitMF MFIMFCreateFlat(int n, int k, double *o,double *v)
 {
  static char RoutineName[]={"MFIMFCreateFlat"};
  MFImplicitMF flat;
  MFNSpace space;
  struct MFFlatData *data;
  int i;

  flat=MFIMFCreateBaseClass(n,k,"Flat");

  space=MFCreateNSpace(n);
  MFIMFSetSpace(flat,space);
  MFFreeNSpace(space);

  data=(struct MFFlatData*)malloc(sizeof(struct MFFlatData)); /*done*/
  if(data==(struct MFFlatData*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFFlatData));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    free(flat);
    return (MFImplicitMF)NULL;
   }
  data->n=n;
  data->k=k;
  data->o=(double*)malloc(n*sizeof(double)); /*done*/
  if(data->o==(double*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    free(flat);
    free(data);
    return (MFImplicitMF)NULL;
   }
  for(i=0;i<n;i++)data->o[i]=o[i];
  data->v=(double*)malloc(k*n*sizeof(double)); /*done*/
  if(data->v==(double*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*k*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    free(flat);
    free(data->o);
    free(data);
    return (MFImplicitMF)NULL;
   }
  for(i=0;i<k*n;i++)data->v[i]=v[i];

  MFIMFSetData(flat,(void*)data);
  MFIMFSetFreeData(flat,MFFreeFlatData);
  MFIMFSetProject(flat,MFProjectFlat);
  MFIMFSetTangent(flat,MFTangentFlat);
  MFIMFSetScale(flat,MFScaleFlat);
  MFIMFSetWriteData(flat,MFWriteFlatData);
  MFIMFSetProjectForSave(flat,MFFlatProjectToSave);
  MFIMFSetProjectForDraw(flat,MFFlatProjectToDraw);
  MFIMFSetProjectForBB(flat,MFFlatProjectForBB);

  return flat;
 }

void MFFreeFlatData(void *d)
 {
  static char RoutineName[]={"MFFreeFlatData"};
  struct MFFlatData *data;

  if(d==(void*)NULL)return;
  data=(struct MFFlatData*)d;
  if(data->o!=(double*)NULL)free(data->o);
  if(data->v!=(double*)NULL)free(data->v);
  free(data);
  return;
 }

int MFProjectFlat(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectFlat"};
  int i;
  struct MFFlatData *data;
  double *u0, *u;

  u0=MFNV_CStar(vu0);
  u=MFNV_CStar(vu);

  data=(struct MFFlatData*)d;

  for(i=0;i<data->n;i++)u[i]=u0[i];
  *index=0;
  return 1;
 }

int MFTangentFlat(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentFlat"};
  int i;
  struct MFFlatData *data;
  double *Phi;

  data=(struct MFFlatData*)d;

  Phi=MFNKM_CStar(mPhi);

  for(i=0;i<n;i++)Phi[i]=data->v[i];

  return 1;
 }

double MFScaleFlat(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  return .3;
 }

void MFWriteFlatData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWriteFlatData"};
  struct MFFlatData *data;
  int i;

  data=(struct MFFlatData*)d;

  fprintf(fid,"%d %d\n",data->n,data->k);
  for(i=0;i<data->n;i++)fprintf(fid,"%lf",data->o[i]);
  for(i=0;i<data->k*data->n;i++)fprintf(fid,"%lf",data->v[i]);

  fflush(fid);
  return;
 }

MFImplicitMF MFReadFlat(FILE *fid)
 {
  static char RoutineName[]={"MFReadFlat"};
  int n=0;
  int k=0;
  double *o;
  double *v;
  MFImplicitMF flat;
  struct MFFlatData *data;
  int i;

  fscanf(fid,"%d %d\n",&n,&k);

  o=(double*)malloc(n*sizeof(double)); /*done*/
  if(o==(double*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    free(flat);
    free(data);
    return (MFImplicitMF)NULL;
   }
  for(i=0;i<n;i++)fscanf(fid,"%lf",o+i);
  v=(double*)malloc(k*n*sizeof(double)); /*done*/
  if(v==(double*)NULL)
   {
    sprintf(MFFlatMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*k*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFFlatMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFFlatMFErrorMsg,__LINE__,__FILE__);
    free(o);
    return (MFImplicitMF)NULL;
   }
  for(i=0;i<k*n;i++)fscanf(fid,"%lf",v+i);

  flat=MFIMFCreateFlat(n,k,o,v);
  free(o);
  free(v);

  return flat;
 }

int MFFlatProjectToSave(MFNVector u, double *x, void *d)
 {
  struct MFFlatData *data;
  int i;

  data=(struct MFFlatData*)d;
  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFFlatProjectToDraw(MFNVector u, double *x, void *d)
 {
  struct MFFlatData *data;
  int i;

  data=(struct MFFlatData*)d;
  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFFlatProjectForBB(MFNVector u, double *x, void *d)
 {
  struct MFFlatData *data;
  int i;

  data=(struct MFFlatData*)d;
  if(x==(double*)NULL)return data->n;

  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }
