/* 
    %W%
    %D% %T%
  
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   February 22, 1999                     */
/*              October 6, 2004   Added ProjectFromCenter */

#include <MFImplicitMF.h>
#include <MFKVector.h>
#include <MFNVector.h>
#include <MFNKMatrix.h>
#include <MFNSpace.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>
#ifdef MFDEBUG
#include <umalloc.h>
#endif

#include <MFNVector.h>

static char MFImplicitMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);
void MFIMFSetK(MFImplicitMF,int);

struct MFImplicitMFSt {
                       MFNSpace space;
                       int n;
                       int k;

                       void *data;
                       void (*writedata)(FILE*,void*);
                       void (*freedata)(void*);
                       int (*projectFromCenter)(int,int,MFNVector,MFNKMatrix,MFKVector,MFNVector,void*,int*);
                       int (*project)(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
                       int (*tangent)(int,int,MFNVector,MFNKMatrix,void*);
                       int (*tangentWithGuess)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*);
                       void (*evaluate)(int,MFNVector,MFNVector,void*);
                       void (*applyJacobian)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*);
                       void (*applySecDer)(int,int,MFNVector,MFNVector,MFNVector,MFNVector,void*);
                       double (*scale)(int,int,MFNVector,MFNKMatrix,void*);
                       int (*stop)(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*);
                       int (*singular)(int,int,MFNVector,MFNKMatrix,MFNVector,void*);
                       void (*stability)(MFImplicitMF,MFNVector,MFNKMatrix,void*);
                       double R;
                       double RMin;

                       char *id;
                       int nRefs;
                       int (*saver)(MFNVector,double*,void*);
                       int (*drawer)(MFNVector,double*,void*);
                       int (*bbprojector)(MFNVector,double*,void*);
                      };

void MFRefImplicitMF(MFImplicitMF this)
 {
  static char RoutineName[]={"MFRefImplicitMF"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefImplicitMF++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
/*  sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
*/
#ifdef MFTIMINGS
    MFTimeMFRefImplicitMF+=clock()-starttime;
#endif
    return;
   }

  this->nRefs++;
#ifdef MFTIMINGS
    MFTimeMFRefImplicitMF+=clock()-starttime;
#endif
  return;
 }

char *MFImplicitMFId(MFImplicitMF this)
 {
  static char RoutineName[]={"MFImplicitMFId"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFImplicitMFId++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFImplicitMFId+=clock()-starttime;
#endif
    return (char*)NULL;
   }

#ifdef MFTIMINGS
    MFTimeMFImplicitMFId+=clock()-starttime;
#endif
  return this->id;
 }

void MFFreeImplicitMF(MFImplicitMF this)
 {
  static char RoutineName[]={"MFFreeImplicitMF"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeImplicitMF++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeImplicitMF+=clock()-starttime;
#endif
    return;
   }

  this->nRefs--;

  if(this->nRefs<1)
   {
    if(this->space!=(MFNSpace)NULL)MFFreeNSpace(this->space);
    if(this->freedata!=NULL && this->data!=(void*)NULL)(this->freedata)(this->data);
    if(this->id!=(char*)NULL)free(this->id);
    free(this);
   }
#ifdef MFTIMINGS
    MFTimeMFFreeImplicitMF+=clock()-starttime;
#endif
  return;
 }

int MFIMF_N(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMF_N"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMF_N++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFIMF_N+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
    MFTimeMFIMF_N+=clock()-starttime;
#endif
  return this->n;
 }

int MFIMF_K(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMF_K"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMF_K++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFIMF_K+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
    MFTimeMFIMF_K+=clock()-starttime;
#endif
  return this->k;
 }

MFNSpace MFIMFNSpace(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMFNSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFNSpace++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFIMFNSpace+=clock()-starttime;
#endif
    return (MFNSpace)NULL;
   }

#ifdef MFTIMINGS
    MFTimeMFIMFNSpace+=clock()-starttime;
#endif
  return this->space;
 }

int MFIMFProjectFromCenter(MFImplicitMF this,MFNVector u0,MFNKMatrix Phi,MFKVector s, MFNVector u)
 {
  static char RoutineName[]={"MFIMFProjectFromCenter"};
  int rc;
  int index;
  if(this->projectFromCenter==NULL)
   {
    MFNVector up;
    MFNVector v;
    up=MFCloneNVector(u0);
    v=MFCloneNVector(u);
    MFMVMul(MFIMFNSpace(this),Phi,s,v);
    MFNSpaceAdd(MFIMFNSpace(this),u0,v,up);
    rc=this->project(this->n,this->k,up,Phi,u,this->data,&index);
    MFNVSetIndex(u,index);
    MFFreeNVector(up);
    MFFreeNVector(v);
    return rc;
   }

  rc=this->projectFromCenter(this->n,this->k,u0,Phi,s,u,this->data,&index);
  MFNVSetIndex(u,index);

  return(rc);
 }

int MFIMFProject(MFImplicitMF this,MFNVector u0,MFNKMatrix Phi,MFNVector u)
 {
  static char RoutineName[]={"MFIMFProject"};
  int rc;
  int index;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFProject++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif
  if(this->project==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"No Project Routine has been supplied");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return 0;
   }

  rc=this->project(this->n,this->k,u0,Phi,u,this->data,&index);
  MFNVSetIndex(u,index);

#ifdef MFTIMINGS
    MFTimeMFIMFProject+=clock()-starttime;
#endif
  return(rc);
 }

MFNKMatrix MFIMFTangentSpace(MFImplicitMF this,MFNVector u)
 {
  static char RoutineName[]={"MFIMFTangentSpace"};
  MFNVector *col;
  MFNKMatrix Phi;
  int i;
  int rc;
  int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFTangentSpace++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  col=(MFNVector*)malloc(this->k*sizeof(MFNVector));
  if(col==(MFNVector*)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Out of memory, trying to allocate %d bytes",this->k*sizeof(MFNVector));
    printf("%s -- %s\n",RoutineName,MFImplicitMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return (MFNKMatrix)NULL;
   }

  if(this->tangent==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"No Tangent Routine has been supplied");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return 0;
   }

  for(i=0;i<this->k;i++)col[i]=MFCloneNVector(u);
  Phi=MFCreateNKMatrix(this->k,col);
  for(i=0;i<this->k;i++)MFFreeNVector(col[i]);
  free(col);

  rc=this->tangent(this->n,this->k,u,Phi,this->data);
  if(!rc)
   {
    MFFreeNKMatrix(Phi);
    return (MFNKMatrix)NULL;
   }

  if(verbose)
   {
    int j,jj;
    MFNVector phi0;
    MFNVector phi1;
    {printf("       check basis for tangent space\n");fflush(stdout);}
    for(jj=0;jj<this->k;jj++)
     for(j=0;j<this->k;j++)
      {
       phi0=MFMColumn(Phi,jj);
       phi1=MFMColumn(Phi,j);
       printf(" <phi_%d,phi_%d>=%le\n",jj,j,MFNSpaceInner(this->space,phi0,phi1));
       MFFreeNVector(phi0);
       MFFreeNVector(phi1);
      }
    fflush(stdout);
   }


#ifdef MFTIMINGS
    MFTimeMFIMFTangentSpace+=clock()-starttime;
#endif
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return Phi;
 }

MFNKMatrix MFIMFTangentSpaceWithGuess(MFImplicitMF this,MFNVector u,MFNKMatrix Phi0)
 {
  static char RoutineName[]={"MFIMFTangentSpaceWithGuess"};
  MFNKMatrix Phi;
  int verbose=0;
  int rc;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFTangentSpaceWithGuess++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->tangentWithGuess==NULL)
   {
#ifdef MFTIMINGS
    MFTimeMFIMFTangentSpaceWithGuess+=clock()-starttime;
#endif
    return MFIMFTangentSpace(this,u);
   }

  Phi=MFCloneNKMatrix(Phi0);
  rc=this->tangentWithGuess(this->n,this->k,u,Phi0,Phi,this->data);
  if(!rc)
   {
    MFFreeNKMatrix(Phi);
    return (MFNKMatrix)NULL;
   }


  if(verbose)
   {
    int j,jj;
    MFNVector phi0;
    MFNVector phi1;
    printf("%s\n",RoutineName);
    {printf("       check basis for tangent space\n");fflush(stdout);}
    for(jj=0;jj<this->k;jj++)
     for(j=0;j<this->k;j++)
      {
       phi0=MFMColumn(Phi,jj);
       phi1=MFMColumn(Phi,j);
       printf(" <phi_%d,phi_%d>=%le\n",jj,j,MFNSpaceInner(this->space,phi0,phi1));
       MFFreeNVector(phi0);
       MFFreeNVector(phi1);
      }
    fflush(stdout);
   }


#ifdef MFTIMINGS
    MFTimeMFIMFTangentSpaceWithGuess+=clock()-starttime;
#endif
  return Phi;
 }

double MFIMFScale(MFImplicitMF this,MFNVector u, MFNKMatrix Phi)
 {
  static char RoutineName[]={"MFIMFScale"};
  double result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFScale++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->scale==NULL&&this->R>0.)
    result=this->R;
   else if(this->scale==NULL&&this->R<0.)
    result=1.;
   else{
    result=this->scale(this->n,this->k,u,Phi,this->data);
    if(this->R>0.&&result>this->R)result=this->R;
   }
#ifdef MFTIMINGS
    MFTimeMFIMFScale+=clock()-starttime;
#endif
  return result;
 }

void MFWriteImplicitMF(FILE *fid,MFImplicitMF u)
 {
  static char RoutineName[]={"MFWriteImplicitMF"};
  int i;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteImplicitMF++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  fprintf(fid,"%s\n","ImplicitMF");
  fprintf(fid,"%d\n",strlen(u->id));
  fprintf(fid,"%s\n",u->id);fflush(fid);
  if(verbose){printf(" tag is -->%s<--\n","ImplicitMF");}
  if(verbose){printf(" strlen(id) is %d\n",strlen(u->id));}
  if(verbose){printf(" id is -->%s<--\n",u->id);fflush(fid);}
  u->writedata(fid,u->data);
  fprintf(fid,"%lf %d\n",u->R,u->nRefs);
  if(verbose){printf(" R is %lf\n",u->R);fflush(fid);}
  if(verbose){printf(" nRefs is %d\n",u->nRefs);fflush(fid);}

#ifdef MFTIMINGS
    MFTimeMFWriteImplicitMF+=clock()-starttime;
#endif
  return;
 }

#ifdef TPBVPIMF
MFImplicitMF MFReadTPBVP(FILE*);
#endif
#ifdef PENDULAIMF
MFImplicitMF MFReadPendula(FILE*);
#endif
#ifdef POLYGONIN3SPACEIMF
MFImplicitMF MFReadPolygonIn3Space(FILE*);
#endif
#ifdef CSTRIMF
MFImplicitMF MFReadCSTR(FILE*);
#endif
#ifdef FLATIMF
MFImplicitMF MFReadFlat(FILE*);
#endif
#ifdef NSPACEIMF
MFImplicitMF MFReadNSpaceMF(FILE*);
#endif
#ifdef PLANEIMF
MFImplicitMF MFReadPlane(FILE*);
#endif
#ifdef SPHEREIMF
MFImplicitMF MFReadSphere(FILE*);
#endif
#ifdef CIRCLEIMF
MFImplicitMF MFReadCircle(FILE*);
#endif
#ifdef SWALLOWIMF
MFImplicitMF MFReadSwallow(FILE*);
#endif
#ifdef EXPRESSIONIMF
MFImplicitMF MFReadExpression(FILE*);
#endif
#ifdef INVARIANTMF
MFImplicitMF MFReadInvariantMF(FILE*);
#endif
#ifdef PEITGENIMF
MFImplicitMF MFReadPeitgenSystem(FILE*);
#endif
#ifdef TORUSIMF
MFImplicitMF MFReadTorus(FILE*);
#endif

MFImplicitMF MFReadImplicitMF(FILE *fid)
 {
  static char RoutineName[]={"MFReadImplicitMF"};
  int i;
  MFImplicitMF M;
  char tag[100]="";
  char *id;
  int n=0;
  int nRefs=0;
  double R=0.;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadImplicitMF++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  fscanf(fid,"%s\n",tag);
  if(verbose){printf(" tag is -->%s<--\n",tag);fflush(stdout);}
  if(strcmp(tag,"ImplicitMF"))
   {
    sprintf(MFImplicitMFErrorMsg,"Next Object is not a ImplicitMF! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadImplicitMF+=clock()-starttime;
#endif
    return;
   }
  M=(MFImplicitMF)NULL;
  fscanf(fid,"%d\n",&n);
  if(verbose){printf(" strlen(id) is %d\n",n);fflush(stdout);}
  id=(char*)malloc((n+1)*sizeof(char));
  if(id==(char*)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Out of memory, trying to allocate %d bytes",(n+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFImplicitMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadImplicitMF+=clock()-starttime;
#endif
    return (MFImplicitMF)NULL;
   }
  fscanf(fid,"%s\n",id);
  if(verbose){printf(" id is -->%s<--, length %d\n",id,strlen(id));fflush(stdout);}
  if(0) {}
#ifdef TORUSIMF
   else if(!strcmp(id,"Torus"))
    M=MFReadTorus(fid);
#endif
#ifdef PENDULAIMF
   else if(!strcmp(id,"Pendula"))
    M=MFReadPendula(fid);
#endif
#ifdef POLYGONIN3SPACEIMF
   else if(!strcmp(id,"PolygonIn3Space"))
    M=MFReadPolygonIn3Space(fid);
#endif
#ifdef CSTRIMF
   else if(!strcmp(id,"CSTR"))
    M=MFReadCSTR(fid);
#endif
#ifdef FLATIMF
   else if(!strcmp(id,"Flat"))
    M=MFReadFlat(fid);
#endif
#ifdef NSPACEIMF
   else if(!strcmp(id,"NSpaceMF"))
    M=MFReadNSpaceMF(fid);
#endif
#ifdef PLANEIMF
   else if(!strcmp(id,"Plane"))
    M=MFReadPlane(fid);
#endif
#ifdef SPHEREIMF
   else if(!strcmp(id,"Sphere"))
    M=MFReadSphere(fid);
#endif
#ifdef CIRCLEIMF
   else if(!strcmp(id,"Circle"))
    M=MFReadCircle(fid);
#endif
#ifdef SWALLOWIMF
   else if(!strcmp(id,"Swallow"))
    M=MFReadSwallow(fid);
#endif
#ifdef EXPRESSIONIMF
   else if(!strcmp(id,"Expression"))
    M=MFReadExpression(fid);
#endif
#ifdef INVARIANTMF
   else if(!strcmp(id,"InvariantMF"))
    M=MFReadInvariantMF(fid);
#endif
#ifdef PEITGENMF
   else if(!strcmp(id,"PeitgenSystem"))
    M=MFReadPeitgenSystem(fid);
#endif
#ifdef TPBVPIMF
   else if(!strcmp(id,"TPBVP"))
    M=MFReadTPBVP(fid);
#endif
   else 
    {
     free(id);
#ifdef MFTIMINGS
    MFTimeMFReadImplicitMF+=clock()-starttime;
#endif
     return(M);
    }
  free(id);

  if(M!=(MFImplicitMF)NULL)
   {
    M->space=(MFNSpace)NULL;
    fscanf(fid,"%lf %d\n",&(M->R),&(M->nRefs));
    if(verbose){printf(" R is %lf\n",M->R);fflush(stdout);}
    if(verbose){printf(" nRefs is %d\n",M->nRefs);fflush(stdout);}
   }else
    fscanf(fid,"%lf %d\n",&R,&nRefs);
  
#ifdef MFTIMINGS
    MFTimeMFReadImplicitMF+=clock()-starttime;
#endif
  return M;
 }

void MFIMFSetSpace(MFImplicitMF this,MFNSpace space)
 {
  static char RoutineName[]={"MFIMFSetSpace"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)return;
  if(space!=(MFNSpace)NULL)MFRefNSpace(space);
  if(this->space!=(MFNSpace)NULL)MFFreeNSpace(this->space);
  this->space=space; 
  return;
 }

void MFIMFEvaluate(MFImplicitMF this,MFNVector u,MFNVector f)
 {
  static char RoutineName[]={"MFIMFEvaluate"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFEvaluate++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->evaluate==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"No Routine to evaluate the function");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->evaluate(this->n,u,f,this->data);

#ifdef MFTIMINGS
    MFTimeMFIMFEvaluate+=clock()-starttime;
#endif
  return;
 }

void MFIMFApplyJacobian(MFImplicitMF this,MFNVector u,MFNKMatrix Phi0,MFNKMatrix Phi1)
 {
  static char RoutineName[]={"MFIMFApplyJacobian"};
  int n ,k ;
  int n0,k0;
  int n1,k1;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFApplyJacobian++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  n=this->n;
  k=this->k;
  n0=MFNKMatrixN(Phi0);
  k0=MFNKMatrixK(Phi0);
  n1=MFNKMatrixN(Phi1);
  k1=MFNKMatrixK(Phi1);

  if(n0!=n||n1!=n||k1!=k0)
   {
    sprintf(MFImplicitMFErrorMsg,"Incompatible matrices for apply jacobian (%d,%d) = (%d,%d).(%d,%d)",n1,k1,n,k,n0,k0);
    printf("%s -- %s\n",RoutineName,MFImplicitMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFIMFApplyJacobian+=clock()-starttime;
#endif
    return;
   }

  if(this->applyJacobian==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"No Routine to apply the Jacobian has been supplied");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->applyJacobian(this->n,k0,u,Phi0,Phi1,this->data);

#ifdef MFTIMINGS
  MFTimeMFIMFApplyJacobian+=clock()-starttime;
#endif
  return;
 }

void MFIMFSetK(MFImplicitMF this,int k)
 {
  static char RoutineName[]={"MFIMFSetK"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->k=k;

  return;
 }

int MFIMFStop(MFImplicitMF this, MFNVector u0, MFNKMatrix Phi0, MFNVector u1, MFNKMatrix Phi1)
 {
  static char RoutineName[]={"MFIMFStop"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->stop!=(NULL))
    return this->stop(this,u0,Phi0,u1,Phi1,this->data);
   else
    return 0;
 }

int MFIMFProjectToSave(MFImplicitMF this, MFNVector u, double *x)
 {
  static char RoutineName[]={"MFIMFProjectToSave"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(this->saver==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Project for Save is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return (this->saver)(u,x,this->data);
 }

int MFIMFProjectToDraw(MFImplicitMF this, MFNVector u, double *x)
 {
  static char RoutineName[]={"MFIMFProjectToDraw"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(this->drawer==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Project for Draw is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return (this->drawer)(u,x,this->data);
 }

int MFIMFProjectToBB(MFImplicitMF this, MFNVector u, double *x)
 {
  static char RoutineName[]={"MFIMFProjectToBB"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Pointer to ImplicitMF (argument 1) is NULL");
    MFSetError(4,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(this->bbprojector==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Project for BB is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return (this->bbprojector)(u,x,this->data);
 }

MFImplicitMF MFIMFCreateBaseClass(int n, int k, char *id)
 {
  static char RoutineName[]={"MFIMFCreateBaseClass"};
  MFImplicitMF this;

  if(id==(char*)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Id (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }

  if(k<0)
   {
    sprintf(MFImplicitMFErrorMsg,"k (%d) (argument 3) is invalid, must be positive",k);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }

  this=(MFImplicitMF)malloc(sizeof(struct MFImplicitMFSt)); /*done*/
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFImplicitMFSt));
    printf("%s -- %s\n",RoutineName,MFImplicitMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }

  this->id=(char*)malloc((strlen(id)+1)*sizeof(char)); /*done*/
  if(this->id==(char*)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(id)+1)*sizeof(char));
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }
  strcpy(this->id,id);

  this->space=(MFNSpace)NULL;
  this->n=n;
  this->k=k;
  this->nRefs=1;
  this->data=(void*)NULL;
  this->writedata=NULL;
  this->freedata=NULL;
  this->project=NULL;
  this->projectFromCenter=NULL;
  this->tangent=NULL;
  this->tangentWithGuess=NULL;
  this->evaluate=NULL;
  this->applyJacobian=NULL;
  this->applySecDer=NULL;
  this->scale=NULL;
  this->stop=NULL;
  this->stability=NULL;
  this->singular=NULL;
  this->R=-1;
  this->RMin=-1;
  this->saver=NULL;
  this->drawer=NULL;
  this->bbprojector=NULL;
  this->data=(void*)NULL;

  return this;
 }

void MFIMFSetData(MFImplicitMF this,void *data)
 {
  static char RoutineName[]={"MFIMFSetData"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->data=data;
  return;
 }

void *MFIMFGetData(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMFGetData"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  return this->data;
  return;
 }

void MFIMFSetWriteData(MFImplicitMF this,void (*writedata)(FILE*,void*))
 {
  static char RoutineName[]={"MFIMFSetWriteData"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->writedata=writedata;
  return;
 }

void MFIMFSetFreeData(MFImplicitMF this,void (*freedata)(void*))
 {
  static char RoutineName[]={"MFIMFSetFreeData"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->freedata=freedata;
  return;
 }

void MFIMFSetProjectFromCenter(MFImplicitMF this,int (*projectFromCenter)(int,int,MFNVector,MFNKMatrix,MFKVector,MFNVector,void*,int*))
 {
  static char RoutineName[]={"MFIMFSetProjectFromCenter"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->projectFromCenter=projectFromCenter;
  return;
 }

void MFIMFSetProject(MFImplicitMF this,int (*project)(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*))
 {
  static char RoutineName[]={"MFIMFSetProject"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->project=project;
  return;
 }

void MFIMFSetTangent(MFImplicitMF this,int (*tangent)(int,int,MFNVector,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetTangent"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->tangent=tangent;
  return;
 }

void MFIMFSetTangentWithGuess(MFImplicitMF this,int (*tangentWithGuess)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetTangentWithGuess"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->tangentWithGuess=tangentWithGuess;
  return;
 }

void MFIMFSetEvaluate(MFImplicitMF this,void (*evaluate)(int,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFIMFSetEvaluate"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->evaluate=evaluate;
  return;
 }

void MFIMFSetApplyJacobian(MFImplicitMF this,void (*applyJacobian)(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetApplyJacobian"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->applyJacobian=applyJacobian;
  return;
 }

void MFIMFSetScale(MFImplicitMF this,double (*scale)(int,int,MFNVector,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetScale"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->scale=scale;
  return;
 }

void MFIMFSetStop(MFImplicitMF this,int (*stop)(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetStop"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->stop=stop;
  return;
 }

void MFIMFSetSingular(MFImplicitMF this,int (*singular)(int,int,MFNVector,MFNKMatrix,MFNVector,void*))
 {
  static char RoutineName[]={"MFIMFSetSingular"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->singular=singular;
  return;
 }

double MFIMFGetR(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMFGetR"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  return this->R;
 }

void MFIMFSetR(MFImplicitMF this,double R)
 {
  static char RoutineName[]={"MFIMFSetR"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->R=R;
  return;
 }

double MFIMFGetRMin(MFImplicitMF this)
 {
  static char RoutineName[]={"MFIMFGetRMin"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  return this->RMin;
 }

void MFIMFSetRMin(MFImplicitMF this,double R)
 {
  static char RoutineName[]={"MFIMFSetRMin"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->RMin=R;
  return;
 }

void MFIMFSetProjectForSave(MFImplicitMF this,int (*saver)(MFNVector,double*,void*))
 {
  static char RoutineName[]={"MFIMFSetProjectForSave"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->saver=saver;
  return;
 }

void MFIMFSetProjectForDraw(MFImplicitMF this,int (*drawer)(MFNVector,double*,void*))
 {
  static char RoutineName[]={"MFIMFSetProjectForDraw"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->drawer=drawer;
  return;
 }

void MFIMFSetProjectForBB(MFImplicitMF this,int (*bbprojector)(MFNVector,double*,void*))
 {
  static char RoutineName[]={"MFIMFSetProjectForBB"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->bbprojector=bbprojector;
  return;
 }

int MFIMFSingular(MFImplicitMF this,MFNVector u,MFNKMatrix Phi,MFNVector w)
 {
  static char RoutineName[]={"MFIMFSingular"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->singular!=(NULL))
    return this->singular(this->n,this->k,u,Phi,w,this->data);
   else
    return 0;
 }

void MFIMFApplySecDer(MFImplicitMF this,MFNVector u,MFNVector v0,MFNVector v1,MFNVector w)
 {
  static char RoutineName[]={"MFIMFApplySecDer"};
  int n ,k ;
  int n0,k0;
  int n1,k1;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIMFApplySecDer++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  n=this->n;
  k=this->k;
  n0=MFNV_NC(v0);
  n1=MFNV_NC(v1);

  if(n0!=n||n1!=n)
   {
    sprintf(MFImplicitMFErrorMsg,"Incompatible vectors for apply scond derivative %d = %d,%d",n,n0,n1);
    printf("%s -- %s\n",RoutineName,MFImplicitMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFIMFApplySecDer+=clock()-starttime;
#endif
    return;
   }

  if(this->applySecDer==NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"No Routine to apply the second derivative has been supplied");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->applySecDer(this->n,k0,u,v0,v1,w,this->data);

#ifdef MFTIMINGS
  MFTimeMFIMFApplySecDer+=clock()-starttime;
#endif
  return;
 }

void MFIMFSetApplySecDer(MFImplicitMF this,void (*applySecDer)(int,int,MFNVector,MFNVector,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFIMFSetApplyJacobian"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->applySecDer=applySecDer;
  return;
 }


void MFIMFSetStability(MFImplicitMF this, MFNVector u0, MFNKMatrix Phi0)
 {
  static char RoutineName[]={"MFIMFSetStability"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this->stability!=(NULL)) this->stability(this,u0,Phi0,this->data);

  return;
 }

void MFIMFSetSetStability(MFImplicitMF this,void (*stability)(MFImplicitMF,MFNVector,MFNKMatrix,void*))
 {
  static char RoutineName[]={"MFIMFSetSetStability"};
  if(this==(MFImplicitMF)NULL)
   {
    sprintf(MFImplicitMFErrorMsg,"Implicitly Defined Manifold (argument 1) is NULL");
    MFSetError(12,RoutineName,MFImplicitMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  this->stability=stability;
  return;
 }
