/* 
    @(#)MFLOCA.c	1.4
    03/07/11 16:58:25
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   April 22, 2002 Modified MFTorus       */

#include <MFImplicitMF.h>
#include <MFLOCA.h>
#include <MFError.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static char MFLOCAMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

extern int MFProjectLOCA(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
extern int MFTangentLOCA(int,int,MFNVector,MFNKMatrix,void*);
extern double MFScaleLOCA(int,int,MFNVector,MFNKMatrix,void*);

int MFLOCAProjectToSave(MFNVector,double*,void*);
int MFLOCAProjectToDraw(MFNVector,double*,void*);
int MFLOCAProjectForBB(MFNVector,double*,void*);

double *MFLOCANVP(MFNVector);
double *MFLOCANVX(MFNVector);
double MFPrintMetricLOCA(double*,double*);

MFImplicitMF MFIMFCreateLOCA(int np, void *con)
 {
  static char RoutineName[]={"MFIMFCreateLOCA"};
  MFImplicitMF loca;
  MFNSpace space;

  loca=MFIMFCreateBaseClass(-1,np,"LOCA");

  space=MFCreateLOCANSpace(con);
  MFIMFSetSpace(loca,space);
  MFFreeNSpace(space);

  MFIMFSetData(loca,(void*)con);
  MFIMFSetProject(loca,MFProjectLOCA);
  MFIMFSetTangent(loca,MFTangentLOCA);
  MFIMFSetScale(loca,MFScaleLOCA);
  MFIMFSetProjectForSave(loca,MFLOCAProjectToSave);
  MFIMFSetProjectForDraw(loca,MFLOCAProjectToDraw);
  MFIMFSetProjectForBB(loca,MFLOCAProjectForBB);

  return loca;
 }

int MFLOCAProjectToSave(MFNVector u, double *x, void *d)
 {
  int i,np;
  double *p;

  np=MFLOCANVNP(u);
  if(x==(double*)NULL)return np;

  p=MFLOCANVP(u);
  for(i=0;i<np;i++)x[i]=p[i];

  return 0;
 }

int MFLOCAProjectToDraw(MFNVector u, double *x, void *d)
 {
  int i,np;
  double *p;
  double *y;

  np=MFLOCANVNP(u);
  if(x==(double*)NULL)return np+1;

  p=MFLOCANVP(u);
  y=MFLOCANVX(u);
  x[0]=MFPrintMetricLOCA(y,p);
  for(i=0;i<np;i++)x[i+1]=p[i];

  return 0;
 }

int MFLOCAProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;
  int np;
  double *p;

  np=MFLOCANVNP(u);
  if(x==(double*)NULL)return np;

  p=MFLOCANVP(u);
  for(i=0;i<np;i++)x[i]=p[i];

  return 0;
 }
