/* 
    @(#)MFListOfCharts.c	1.5
    02/04/19 16:30:44
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFListOfCharts.h>
#include <stdio.h>
#include <stdlib.h>
#include <MFTime.h>

void MFSetError(int,char*,char*,int,char*);
static char MFListOfChartsErrorMsg[256]="";

struct MFListOfChartsSt {
                       int n;
                       int *charts;
                      };

MFListOfCharts MFCreateListOfCharts(int n,int *charts)
 {
  static char RoutineName[]={"MFCreateListOfCharts"};
  MFListOfCharts this;
  int i,j,k;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateListOfCharts++;
  starttime=clock();
#endif

  this=(MFListOfCharts)malloc(sizeof(struct MFListOfChartsSt)); /*done*/
  if(this==(MFListOfCharts)NULL)
   {
    sprintf(MFListOfChartsErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFListOfChartsSt));
    printf("%s -- %s\n",RoutineName,MFListOfChartsErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFListOfChartsErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateListOfCharts+=clock()-starttime;
#endif
    return (MFListOfCharts)NULL;
   }

  this->n=n;
  this->charts=charts;

  for(i=0;i<n-1;i++)
   for(j=i+1;j<n;j++)
    {
     if(this->charts[i]>this->charts[j])
      {
       k=this->charts[j];
       this->charts[j]=this->charts[i];
       this->charts[i]=k;
      }
    }

#ifdef MFTIMINGS
    MFTimeMFCreateListOfCharts+=clock()-starttime;
#endif
  return(this);
 }

int MFNumberOfIntersectingCharts(MFListOfCharts L)
 {
  static char RoutineName[]={"MFNumberOfIntersectingCharts"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNumberOfIntersectingCharts++;
  starttime=clock();
#endif
#ifdef MFTIMINGS
    MFTimeMFNumberOfIntersectingCharts+=clock()-starttime;
#endif
  return(L->n);
 }

int MFIntersectingChart(MFListOfCharts L,int chart)
 {
  static char RoutineName[]={"MFIntersectingChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFIntersectingChart++;
  starttime=clock();
#endif
#ifdef MFTIMINGS
    MFTimeMFIntersectingChart+=clock()-starttime;
#endif
  return(L->charts[chart]);
 }

void MFFreeListOfIntersectingCharts(MFListOfCharts L)
 {
  static char RoutineName[]={"MFFreeListOfIntersectingCharts"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeListOfIntersectingCharts++;
  starttime=clock();
#endif
  if(L->charts!=(int*)NULL)free(L->charts);
  free(L);
#ifdef MFTIMINGS
    MFTimeMFFreeListOfIntersectingCharts+=clock()-starttime;
#endif
  return;
 }
