/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              October 6, 2000   Added Polyhedron    */

#include <stdio.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <stdlib.h>
#include <math.h>
#include <MFTime.h>

struct MFNRegionSt
 {
  char *id;
  int (*test)(MFNVector,void*);
  void *data;
  void (*freedata)(void*);
  void (*writedata)(FILE*,void*);
 };

void MFSetError(int,char*,char*,int,char*);
static char MFNRegionErrorMsg[256]="";

double *MFNV_CStar(MFNVector);

int MFNRegionInterior(MFNRegion Omega,MFNVector v)
 {
  static char RoutineName[]={"MFRegionInterior"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRegionInterior++;
  starttime=clock();
#endif

  result=Omega->test(v,Omega->data);

#ifdef MFTIMINGS
    MFTimeMFRegionInterior+=clock()-starttime;
#endif

  return result;
 }

void MFFreeNRegion(MFNRegion Omega)
 {
  char RoutineName[]={"MFFreeNRegion"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeNRegion++;
  starttime=clock();
#endif 

  if(Omega->id!=(char*)NULL)free(Omega->id);
  if(Omega->freedata!=NULL)Omega->freedata(Omega->data);
  free(Omega);

#ifdef MFTIMINGS
    MFTimeMFFreeNRegion+=clock()-starttime;
#endif
  return;
 }

MFNRegion MFNRegionCreateBaseClass(char *id)
 {
  MFNRegion result;
  static char RoutineName[]={"MFNRegionCreateBaseClass"};

  result=(MFNRegion)malloc(sizeof(struct MFNRegionSt)); /*done*/
  if(result==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNRegionSt));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return (MFNRegion)NULL;
   }

  result->id=(char*)malloc((strlen(id)+1)*sizeof(char)); /*done*/
  if(result->id==(char*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(id)+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return (MFNRegion)NULL;
   }

  strcpy(result->id,id);

  return result;
 }

void MFNRegionSetTest(MFNRegion Omega,int (*test)(MFNVector,void*))
 {
  static char RoutineName[]={"MFNRegionSetTest"};

  if(Omega==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Region (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return;
   }

  Omega->test=test;

  return;
 }

void MFNRegionSetData(MFNRegion Omega,void *data)
 {
  static char RoutineName[]={"MFNRegionSetData"};

  if(Omega==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Region (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return;
   }

  Omega->data=data;

  return;
 }

void *MFNRegionGetData(MFNRegion Omega)
 {
  static char RoutineName[]={"MFNRegionGetData"};

  if(Omega==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Region (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return (void*)NULL;
   }

  return Omega->data;
 }

void MFNRegionSetFreeData(MFNRegion Omega,void (*freedata)(void*))
 {
  static char RoutineName[]={"MFNRegionSetFreeData"};

  if(Omega==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Region (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return;
   }

  Omega->freedata=freedata;
  return;
 }

void MFNRegionSetWriteData(MFNRegion Omega,void (*writedata)(FILE*,void*))
 {
  static char RoutineName[]={"MFNRegionSetWriteData"};

  if(Omega==(MFNRegion)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Region (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    return;
   }

  Omega->writedata=writedata;
  return;
 }
