/* 
    @(#)MFPendula.c	1.17
    03/08/07 16:05:05
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   February 22, 1999                     */

#include <MFImplicitMF.h>
#include <MFNKMatrix.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFFortran.h>

#define GU GuPendula
#define GUU GuuPendula

static char MFPendulaMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

extern double MFEpsilon;

/* Space for LU decompositions */

double *A,*B,*C,*D,*R,*S,*X,*Y;
int   *Pivots;

int nBorders,nBandsU,nBandsL;
int ldA,ldB,ldC,ldD;

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFFreePendulaData(void*);
int MFProjectPendula(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentPendula(int,int,MFNVector,MFNKMatrix,void*);
int MFTangentPendulaWithGuess(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*);
double MFScalePendula(int,int,MFNVector,MFNKMatrix,void*);
void MFPendulaInitializeSolver(int);
double GPendula(int,double*,int,double,double,double);
double GuPendula(int,int,double*,int,double,double,double);
double GuuPendula(int,int,int,double*,int,double,double,double);
double TGuPendula(int,int,double*,int,double,double,double);
double TGuuPendula(int,int,int,double*,int,double,double,double);
int MFTangentPendulaSingle(int,int,int,double*,double*,double*,double*,void*);
void MFCurvaturePendula(int,int,double*,double*,double*,double*,double*,double*,void*);
void MFWritePendulaData(FILE*,void*);
MFImplicitMF MFReadPendula(FILE*);

int MFPendulaProjectToSave(MFNVector,double*,void*);
int MFPendulaProjectToDraw(MFNVector,double*,void*);
int MFPendulaProjectForBB(MFNVector,double*,void*);

struct MFPendulaData
 {
  int nt;
  double kappa;
  double gamma;
  double R;
 };

MFImplicitMF MFIMFCreatePendula(int nt,double kappa,double gamma,int windingno)
 {
  static char RoutineName[]={"MFIMFCreatePendula"};
  MFImplicitMF pendula;
  int *e;
  int i;
  MFNSpace space;
  struct MFPendulaData *data;

  pendula=MFIMFCreateBaseClass(2*nt+3,2,"Pendula");

  space=MFCreateWeightedNSpaceWithExponents(2*nt+3,2*nt,1./nt,e);
  MFIMFSetSpace(pendula,space);
  MFFreeNSpace(space);

  data=(struct MFPendulaData*)malloc(sizeof(struct MFPendulaData)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFPendulaData));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }
  data->nt=nt;
  data->kappa=kappa;
  data->gamma=gamma;
  data->R=(double)windingno;

  MFIMFSetData(pendula,(void*)data);
  MFIMFSetFreeData(pendula,MFFreePendulaData);
  MFIMFSetProject(pendula,MFProjectPendula);
  MFIMFSetTangent(pendula,MFTangentPendula);
  MFIMFSetTangentWithGuess(pendula,MFTangentPendulaWithGuess);
  MFIMFSetScale(pendula,MFScalePendula);
  MFIMFSetWriteData(pendula,MFWritePendulaData);
  MFIMFSetProjectForSave(pendula,MFPendulaProjectToSave);
  MFIMFSetProjectForDraw(pendula,MFPendulaProjectToDraw);
  MFIMFSetProjectForBB(pendula,MFPendulaProjectForBB);

  MFPendulaInitializeSolver(nt);

  return pendula;
 }

void MFFreePendulaData(void *d)
 {
  static char RoutineName[]={"MFFreePendulaData"};
  free((struct MFPendulaData*)d);
  return;
 }

extern double MFEpsilon;

double MFScalePendula(int n,int k,MFNVector vu,MFNKMatrix mPhi, void *d)
 {
  static char RoutineName[]={"MFScalePendula"};
  double *a00,*a01,*a11;
  double A,B,C,l;
  double r;
  int i,nt;
  double *u,*Phi;
  int verbose;

  verbose=0;

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  nt=(n-3)/2;
  if(1)
   {
    if(verbose){printf("  MFScalePendula \n");fflush(stdout);}
    a00=(double*)malloc(n*sizeof(double)); /*done*/
    if(a00==(double*)NULL)
     {
      sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
      printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
      return -1.;
     }
    a01=(double*)malloc(n*sizeof(double)); /*done*/
    if(a01==(double*)NULL)
     {
      sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
      printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
      return -1.;
     }
    a11=(double*)malloc(n*sizeof(double)); /*done*/
    if(a11==(double*)NULL)
     {
      sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(double));
      printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
      return -1.;
     }

    MFCurvaturePendula(n,k,u,Phi,Phi+n,a00,a01,a11,d);

    A=0.;
    for(i=0;i<n-3;i++)A+=a00[i]*a00[i]/nt;
    for(i=n-3;i<n;i++)A+=a00[i]*a00[i];
    A=sqrt(A);
    if(verbose){printf("  |a_00|=%lf\n",A);fflush(stdout);}

    B=0.;
    for(i=0;i<n-3;i++)B+=a01[i]*a01[i]/nt;
    for(i=n-3;i<n;i++)B+=a01[i]*a01[i];
    B=sqrt(B);
    if(verbose){printf("  |a_01|=|a_10|=%lf\n",B);fflush(stdout);}

    C=0.;
    for(i=0;i<n-3;i++)C+=a11[i]*a11[i]/nt;
    for(i=n-3;i<n;i++)C+=a11[i]*a11[i];
    C=sqrt(C);
    if(verbose){printf("  |a_11|=%lf\n",C);fflush(stdout);}

    free(a00);
    free(a01);
    free(a11);

    if(A+C>0)
      l=.5*(A+C)+.5*sqrt((A-C)*(A-C)+4*B*B);
     else
      l=-.5*(A+C)+.5*sqrt((A-C)*(A-C)+4*B*B);
    r=.6*sqrt(2*MFEpsilon/l);

    if(verbose){printf("%s r=%lf\n",RoutineName,r);fflush(stdout);}
    if(r>.4)r=.4;
   }else{
    r=.4;
   }

  return r;
 }

double GPendula(int i,double *point,int nt, double kappa, double gamma, double r)
 {
  static char RoutineName[]={"GPendula"};
  double w,Ic,In;
  double x,h;
  double W;
  double result;
  int n;
  int j,m;
  int jm,j0,jp;
  double f,fm,f0,fp;

/*printf("GPendula(%d,0x%8.8x,%d,%lf,%lf,%lf)\n",i,point,nt,kappa,gamma,r);fflush(stdout);*/

  if(i<0 || i>2*nt+1)
   {
    printf("i<0 || i>n-k, in G i=%d, n-k=%d\n",i,2*nt+1);
    return(0.);
   }

  w=2*3.1415926/(10*point[2*nt]);
  In=point[2*nt+1];
  Ic=3.1415926*kappa*point[2*nt+2];
  Ic=point[2*nt+2];

/*
 
   EVALUATES G (U,L)
                                 I
      U IS:
             1     2     3        N-4  N-3   N-2  N-1 N
 
           P1(1),P2(1),P1(2),...,P1(M),P2(M), T , In, Ic
 
*/

  h=r*2.*3.1415926/nt;

  if(i<0 || i>=2*nt+1)
   {
    printf("i<0 || i>=n in GPendula i=%d, n=%d\n",i,2*nt+1);fflush(stdout);
    abort();
    return(0.);
   }

  j=i/2;
  m=i%2;

  x=j*h;

  jm=j-1;
  j0=j;
  jp=j+1;

  if(j0== 0)jm=nt-1;
  if(j0==nt-1)jp=0;

  result=0.;
  if(i<2*nt)
   {
    switch(m)
     {
      case 0:
       fp=w+sin(w*(x+h)+point[2*jp])+kappa*(point[2*jp]-point[2*jp+1])-In-Ic;
       f0=w+sin(w*x  +point[2*j0])+kappa*(point[2*j0]-point[2*j0+1])-In-Ic;
       fm=w+sin(w*(x-h)+point[2*jm])+kappa*(point[2*jm]-point[2*jm+1])-In-Ic;
       f=.25*(fp+2*f0+fm);
       result=w*w*(point[2*jp]-2.*point[2*j0]+point[2*jm])+gamma*w*(point[2*jp]-point[2*jm])*h+f*h*h;
       break;
      case 1:
       fp=w+sin(w*(x+h)+point[2*jp+1])+kappa*(point[2*jp+1]-point[2*jp])-In+Ic;
       f0=w+sin(w*(x  )+point[2*j0+1])+kappa*(point[2*j0+1]-point[2*j0])-In+Ic;
       fm=w+sin(w*(x-h)+point[2*jm+1])+kappa*(point[2*jm+1]-point[2*jm])-In+Ic;
       f=.25*(fp+2*f0+fm);
       result=w*w*(point[2*jp+1]-2.*point[2*j0+1]+point[2*jm+1])+gamma*w*(point[2*jp+1]-point[2*jm+1])*h+f*h*h;
       break;
     }
   }else{
    result=.5*(point[2*(nt-1)]+point[2*(nt-1)+1]);    /* Phase Constraint */
   }

  return(result);
 }

void MFPendulaInitializeSolver(int nt)
 {
  static char RoutineName[]={"MFPendulaInitializeSolver"};
  int nEqs,n;
  int k;

  nEqs=2;
  k=2;
  n=nEqs*nt+3;
  nBorders=nEqs+1;
  nBandsU=2*nEqs-1;
  nBandsL=2*nEqs-1;

  ldA=2*nBandsL+nBandsU+1;
  ldB=n-k-nBorders;
  ldC=nBorders+k;
  ldD=nBorders+k;

  A=(double*)malloc(ldA*(n-k-nBorders)*sizeof(double)); /*done*/
  if(A==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",ldA*(n-k-nBorders)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  B=(double*)malloc(ldB*(nBorders+k)*sizeof(double)); /*done*/
  if(B==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",ldB*(nBorders+k)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  C=(double*)malloc(ldC*(n-k-nBorders)*sizeof(double)); /*done*/
  if(C==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",ldC*(n-k-nBorders)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  D=(double*)malloc(ldD*(nBorders+k)*sizeof(double)); /*done*/
  if(D==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",ldD*(nBorders+k)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  R=(double*)malloc((n-k-nBorders)*sizeof(double)); /*done*/
  if(R==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",(n-k-nBorders)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  S=(double*)malloc((nBorders+k)*sizeof(double)); /*done*/
  if(S==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",(nBorders+k)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  X=(double*)malloc((nBorders+k+nBandsL+1)*(nBorders+k+nBandsL+1)*sizeof(double)); /*done*/
  if(X==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",(nBorders+k+nBandsL+1)*(nBorders+k+nBandsL+1)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  Y=(double*)malloc((nBorders+k+nBandsL+1)*sizeof(double)); /*done*/
  if(Y==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",(nBorders+k+nBandsL+1)*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  Pivots=(int*)malloc(n*sizeof(int)); /*done*/
  if(Pivots==(int*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(A==(double*)NULL || B==(double*)NULL || C==(double*)NULL || D==(double*)NULL || 
     R==(double*)NULL || S==(double*)NULL || X==(double*)NULL || Y==(double*)NULL || 
     Pivots==(int*)NULL)
   {
    printf(" %s: out of space (malloc returned NULL)\n",RoutineName);
    exit(8);
   }

  return;
 }

int MFProjectPendula(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectPendula"};
  static double eps=.001;
  static double e;
  static double emax;
  static double error;
  static double delta;
  static double factor;
  static int i,j;
  static int i0,i1;
  static int itimes;
  static int verbose,test;
  static int nBands,bandW;
  static int ierr;
  static double *tangentA;
  static double *tangentB;
  double *u,*u0,*Phi;

  static int nt;
  static int na,nb;
  static double kappa,gamma,r;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  nt=((struct MFPendulaData*)d)->nt;
  kappa=((struct MFPendulaData*)d)->kappa;
  gamma=((struct MFPendulaData*)d)->gamma;
  r=((struct MFPendulaData*)d)->R;

  verbose=0;
  test=0;

  tangentA=Phi;
  tangentB=Phi+n;

  nBands=nBandsL+nBandsU+1;

  for(i=0;i<n;i++)u[i]=u0[i];

/*printf("MFProjectPendula, nt=%d, kappa=%lf, gamma=%lf, R=%lf\n",nt,kappa,gamma,r);fflush(stdout);*/

/* Test basis for the tangent Space */
  if(test)
   {
    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentA[i]*tangentA[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentA[i]*tangentA[i];
    printf("direction A.direction A: %e\n",e);

    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentA[i]*tangentB[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentA[i]*tangentB[i];
    printf("direction A.direction B: %e\n",e);

    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentB[i]*tangentB[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentB[i]*tangentB[i];
    printf("direction B.direction B: %e\n",e);
   }

  error=1.;
  itimes=0;
  while(error>1.e-10)
   {
    for(j=0;j<n-k-nBorders;j++)
     {
      i0=j-nBandsU;
      if(i0<0)i0=0;
      i1=j+nBandsL;
      if(i1>n-k-nBorders-1)i1=n-k-nBorders-1;
      for(i=i0;i<i1+1;i++)
        A[i-j+nBands-1+ldA*j]=GU(i,j,u,nt,kappa,gamma,r);
     }

    for(i=0;i<n-k-nBorders;i++)
     {
      for(j=0;j<nBorders+k;j++)
        B[i+ldB*j]=GU(i,n-k-nBorders+j,u,nt,kappa,gamma,r);
     }

    for(i=0;i<nBorders;i++)
     {
      for(j=0;j<n-k-nBorders;j++)
        C[i+ldC*j]=GU(n-k-nBorders+i,j,u,nt,kappa,gamma,r);
     }
    for(j=0;j<n-k-nBorders;j++)
     {
      if(j<2*nt)
       {
        C[nBorders+0+ldC*j]=tangentA[j]/nt;
        C[nBorders+1+ldC*j]=tangentB[j]/nt;
       }else{
        C[nBorders+0+ldC*j]=tangentA[j];
        C[nBorders+1+ldC*j]=tangentB[j];
       }
     }

    for(i=0;i<nBorders;i++)
     {
      for(j=0;j<nBorders+k;j++)
        D[i+ldD*j]=GU(n-k-nBorders+i,n-k-nBorders+j,u,nt,kappa,gamma,r);
     }
    for(j=0;j<nBorders+k;j++)
     {
      if(n-k-nBorders+j<2*nt)
       {
        D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j]/nt;
        D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j]/nt;
       }else{
        D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j];
        D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j];
       }
     }

    for(i=0;i<n-k-nBorders;i++)
      R[i]=-GPendula(i,u,nt,kappa,gamma,r);

    for(i=0;i<nBorders;i++)
      S[i]=-GPendula(n-k-nBorders+i,u,nt,kappa,gamma,r);

    S[nBorders+0]=0.;
    S[nBorders+1]=0.;
    for(i=0;i<n;i++)
     {
      if(i<2*nt)
       {
        S[nBorders+0]+=tangentA[i]*(u0[i]-u[i])/nt;
        S[nBorders+1]+=tangentB[i]*(u0[i]-u[i])/nt;
       }else{
        S[nBorders+0]+=tangentA[i]*(u0[i]-u[i]);
        S[nBorders+1]+=tangentB[i]*(u0[i]-u[i]);
       }
     }

/* Calculate the size of the right hand side. */

    error=0.;
    for(i=0;i<n-k-nBorders;i++)
     {
      if(i<2*nt)
       {
        error+=R[i]*R[i]/nt;
       }else{
        error+=R[i]*R[i];
       }
     }
    for(i=0;i<nBorders+k;i++)
     {
      if(i+n-k-nBorders<2*nt)
       {
        error+=S[i]*S[i]/nt;
       }else{
        error+=S[i]*S[i];
       }
     }
    error=sqrt(error);

/* Compute the correction. */
    if(0&&verbose)
     {
      printf(" System:\n\n");
      for(i=0;i<n-k-nBorders;i++)
       {
        printf("[");
        for(j=0;j<n-k-nBorders;j++)
         {
          if(j>i-nBandsL-1 && j<i+nBandsU+1)
           {
            printf("%7.3lf",A[i-j+nBands-1+ldA*j]);
           }else{
            printf("       ");
           }
         }
        printf(" | ");
        for(j=0;j<nBorders+k;j++)printf("%7.3lf",B[i+ldB*j]);
        printf("]");
        printf(" [%7.3lf]\n",R[i]);
       }
  
      printf("[");
      for(j=0;j<n-k-nBorders;j++)printf("--------");
      printf(" + ");
      for(j=0;j<nBorders+k;j++)printf("--------");
      printf("]");
      printf(" [-------]\n");

      for(i=0;i<nBorders+k;i++)
       {
        printf("[");
        for(j=0;j<n-k-nBorders;j++)printf("%7.3lf",C[i+ldC*j]);
        printf(" | ");
        for(j=0;j<nBorders+k;j++)printf("%7.3lf",D[i+ldC*j]);
        printf("]");
        printf(" [%7.3lf]\n",S[i]);
       }
     }

    if(verbose)printf("\n                   Size of rhs: %e\n",error);

    na=n-k-nBorders;
    nb=nBorders+k;
    F77_FUNC(dbofa,DBOFA)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,(double*)NULL,Pivots,&ierr);
    if(ierr!=0)
     {
      printf(" Problem with factor, zero on diagonal %d\n",ierr);
      return 0;
     }
    ierr=0;
    F77_FUNC(dbosl,DBOSL)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,Y,R,S,(double*)NULL,Pivots,&ierr);

/* Is GPendula(u+eps*delta)-GPendula(u) ~ -eps*GPendula(u)  */
/*    P(u+eps*delta)-P(u) ~ -eps*P(u)  */
/*    N(u+eps*delta)-N(u) ~ -eps*N(u)? */

/* Calculate the size of the Correction. */

    delta=0.;
    for(i=0;i<n-k-nBorders;i++)
     {
      if(i<2*nt)
       {
        delta+=R[i]*R[i]/nt;
       }else{
        delta+=R[i]*R[i];
       }
     }
    for(i=0;i<nBorders+k;i++)
     {
      if(i+n-k-nBorders<2*nt)
       {
        delta+=S[i]*S[i]/nt;
       }else{
        delta+=S[i]*S[i];
       }
     }
    delta=sqrt(delta);

    if(verbose)printf("\n                   Size of correction: %e\n",delta);

/* Apply the Correction */

    factor=1.;
/*  if(delta>MFEpsilon)return 0;*/

    for(i=0;i<n-k-nBorders;i++)
      u[i]+=factor*R[i];
    for(i=0;i<nBorders+k;i++)
      u[n-k-nBorders+i]+=factor*S[i];

    if(FALSE||verbose)printf(" %d error %e size of correction %e\n",itimes,error,delta);

    itimes++;
    if(itimes>20)
     {
      printf(" Too many refine iterations %d\n",itimes);fflush(stdout);
      return 0;
     }
   }
  if(verbose)printf("\n");

  *index=0;
  if(FALSE||verbose)printf("Done refine\n");
  return 1;
 }

int MFTangentPendula(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentPendula"};
  int i;
  double *Phi0;
  double *u,*Phi;

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  Phi0=(double*)malloc(n*k*sizeof(double)); /*done*/
  if(Phi0==(double*)NULL)
   {
    sprintf(MFPendulaMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*k*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFPendulaMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPendulaMFErrorMsg,__LINE__,__FILE__);
    return;
   }
  for(i=0;i<n*k;i++)Phi0[i]=0.;
  Phi0[n-2]=1.;
  Phi0[n-1+n]=1.;
   
  MFTangentPendulaSingle(n,k,0,u,Phi0,Phi0+n,Phi,d);
  MFTangentPendulaSingle(n,k,1,u,Phi ,Phi0+n,Phi,d);
  free(Phi0);
  return 1;
 }

int MFTangentPendulaWithGuess(int n,int k,MFNVector vu,MFNKMatrix mPhi0,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentPendulaWithGuess"};
  double *u,*Phi,*Phi0;

  Phi=MFNKM_CStar(mPhi);
  Phi0=MFNKM_CStar(mPhi0);
  u=MFNV_CStar(vu);

  MFTangentPendulaSingle(n,k,0,u,Phi0,Phi0+n,Phi,d);
  MFTangentPendulaSingle(n,k,1,u,Phi ,Phi0+n,Phi,d);
  return 1;
 }

int MFTangentPendulaSingle(int n,int k,int t, double *u,double *tangentA,double *tangentB,double *Phi, void *d)
 {
  static double *result;
  static int i,j;
  static int i0,i1;
  static int verbose,test;
  static int nBands,bandW;
  static int na,nb;
  static int ierr;
  static double s;
  static double e;
  static int nt;
  static double kappa,gamma,r;

  nt=((struct MFPendulaData*)d)->nt;
  kappa=((struct MFPendulaData*)d)->kappa;
  gamma=((struct MFPendulaData*)d)->gamma;
  r=((struct MFPendulaData*)d)->R;

  verbose=0;
  test=0;
  if(verbose||test)printf("surfaceTangentBordered direction %d\n",t);

  if(verbose)
   {
    printf("TangentA:\n");
    printf("[");
    if(n<10)
     {
      for(i=0;i<n;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",tangentA[i]);
       }
     }else{
      for(i=0;i<4;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",tangentA[i]);
       }
      printf(",...");
      for(i=n-3;i<n;i++)
       {
        printf(",%7.3lf",tangentA[i]);
       }
     }
    printf("]\n");fflush(stdout);
   
    printf("TangentB:\n");
    printf("[");
    if(n<10)
     {
      for(i=0;i<n;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",tangentB[i]);
       }
     }else{
      for(i=0;i<4;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",tangentB[i]);
       }
      printf(",...");
      for(i=n-3;i<n;i++)
       {
        printf(",%7.3lf",tangentB[i]);
       }
     }
    printf("]\n");fflush(stdout);
   }

  if(test)
   {
    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentA[i]*tangentA[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentA[i]*tangentA[i];
    printf("direction A.direction A: %e\n",e);

    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentA[i]*tangentB[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentA[i]*tangentB[i];
    printf("direction A.direction B: %e\n",e);

    e=0.;
    for(i=0;i<2*nt;i++)
      e+=tangentB[i]*tangentB[i]/nt;
    for(i=2*nt;i<n;i++)
      e+=tangentB[i]*tangentB[i];
    printf("direction B.direction B: %e\n",e);
   }

  nBands=nBandsL+nBandsU+1;

  for(j=0;j<n-k-nBorders;j++)
   {
    i0=j-nBandsU;
    if(i0<0)i0=0;
    i1=j+nBandsL;
    if(i1>n-k-nBorders-1)i1=n-k-nBorders-1;
    for(i=i0;i<i1+1;i++)
      A[i-j+nBands-1+ldA*j]=GU(i,j,u,nt,kappa,gamma,r);
   }

  for(i=0;i<n-k-nBorders;i++)
   {
    for(j=0;j<nBorders+k;j++)
      B[i+ldB*j]=GU(i,n-k-nBorders+j,u,nt,kappa,gamma,r);
   }

  for(i=0;i<nBorders;i++)
   {
    for(j=0;j<n-k-nBorders;j++)
      C[i+ldC*j]=GU(n-k-nBorders+i,j,u,nt,kappa,gamma,r);
   }
  for(j=0;j<n-k-nBorders;j++)
   {
    if(j<2*nt)
     {
      C[nBorders+0+ldC*j]=tangentA[j]/nt;
      C[nBorders+1+ldC*j]=tangentB[j]/nt;
     }else{
      C[nBorders+0+ldC*j]=tangentA[j];
      C[nBorders+1+ldC*j]=tangentB[j];
     }
   }

  for(i=0;i<nBorders;i++)
   {
    for(j=0;j<nBorders+k;j++)
      D[i+ldD*j]=GU(n-k-nBorders+i,n-k-nBorders+j,u,nt,kappa,gamma,r);
   }
  for(j=0;j<nBorders+k;j++)
   {
    if(n-k-nBorders+j<2*nt)
     {
      D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j]/nt;
      D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j]/nt;
     }else{
      D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j];
      D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j];
     }
   }
  for(i=0;i<n-k-nBorders;i++)R[i]=0.;

  for(i=0;i<nBorders+k;i++)S[i]=0.;
  S[nBorders+t]=1.;


/* Calculate the size of the right hand side. */

/* Compute the tangent. */
  if(verbose)
   {
    printf(" System:\n\n");
    for(i=0;i<n-k-nBorders;i++)
     {
      printf("[");
      for(j=0;j<n-k-nBorders;j++)
       {
        if(j>i-nBandsL-1 && j<i+nBandsU+1)
         {
          printf("%7.3lf",A[i-j+nBands-1+ldA*j]);
         }else{
          printf("       ");
         }
       }
      printf(" | ");
      for(j=0;j<nBorders+k;j++)printf("%7.3lf",B[i+ldB*j]);
      printf("]");
      printf(" [%7.3lf]\n",R[i]);
     }
  
    printf("[");
    for(j=0;j<n-k-nBorders;j++)printf("--------");
    printf(" + ");
    for(j=0;j<nBorders+k;j++)printf("--------");
    printf("]");
    printf(" [-------]\n");

    for(i=0;i<nBorders+k;i++)
     {
      printf("[");
      for(j=0;j<n-k-nBorders;j++)printf("%7.3lf",C[i+ldC*j]);
      printf(" | ");
      for(j=0;j<nBorders+k;j++)printf("%7.3lf",D[i+ldC*j]);
      printf("]");
      printf(" [%7.3lf]\n",S[i]);
     }
   }

  na=n-k-nBorders;
  nb=nBorders+k;
  F77_FUNC(dbofa,DBOFA)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,(double*)NULL,Pivots,&ierr);
  if(ierr!=0)
   {
    printf(" Problem with factor, zero on diagonal %d\n",ierr);
    exit(8);
   }
  ierr=0;
  F77_FUNC(dbosl,DBOSL)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,Y,R,S,(double*)NULL,Pivots,&ierr);

/* Extract the result */

  for(i=0;i<n-k-nBorders;i++)
    Phi[i+n*t]=R[i];
  for(i=0;i<nBorders+k;i++)
    Phi[n-k-nBorders+i+n*t]=S[i];

/* Normalize */

  s=0.;
  for(i=0;i<2*nt;i++)s+=Phi[i+n*t]*Phi[i+n*t]/nt;
  for(i=2*nt;i<n;i++)s+=Phi[i+n*t]*Phi[i+n*t];
  if(fabs(s)<.00001){printf("Warning, result small in norm -- MFPendula Single Tangent\n");fflush(stdout);}
  s=1./sqrt(s);

  for(i=0;i<n;i++)Phi[i+n*t]=s*Phi[i+n*t];

  if(verbose)
   {
    printf("Tangent (t=%d):\n",t);
    printf("[");
    if(n<10)
     {
      for(i=0;i<n;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",Phi[i+n*t]);
       }
     }else{
      for(i=0;i<4;i++)
       {
        if(i>0)printf(",");
        printf("%7.3lf",Phi[i+n*t]);
       }
      printf(",...");
      for(i=n-3;i<n;i++)
       {
        printf(",%7.3lf",Phi[i+n*t]);
       }
     }
    printf("]\n");fflush(stdout);
    printf("done surfaceTangentBordered\n");
   }

  return 1;
 }

double GuPendula(int i,int j, double *point,int nt, double kappa, double gamma, double r)
 {
  static char RoutineName[]={"GuPendula"};
  double w,Ic,In;
  double dwdT;
  double x,h;
  double W;
  double result;
  int n;
  int I,m;
  int Im,I0,Ip;
  double f,fm,f0,fp;

/*printf("GuPendula(%d,%d,0x%8.8x,%d,%lf,%lf,%lf)\n",i,j,point,nt,kappa,gamma,r);fflush(stdout);*/

  if(i<0 || i>2*nt+1)
   {
    printf("i<0 || i>n-k, in Gu i=%d, n-k=%d\n",i,2*nt+1);
    return(0.);
   }

  if(j<0 || j>2*nt+3)
   {
    printf("j<0 || j>n, in Gu i=%d, j=%d, n=%d\n",i,j,2*nt+3);
    return(0.);
   }

  w=2*3.1415926/(10*point[2*nt]);
  dwdT=-10*w*w/(2*3.1415926);
  In=point[2*nt+1];
  Ic=3.1415926*kappa*point[2*nt+2];
  Ic=point[2*nt+2];

/*
 
   EVALUATES GU(U,L)
                                 I
      U IS:
             1     2     3        N-4  N-3   N-2  N-1 N
 
           P1(1),P2(1),P1(2),...,P1(M),P2(M), T , In, Ic
 
*/

  h=r*2.*3.1415926/nt;

  I=i/2;
  m=i%2;

  x=I*h;

  Im=I-1;
  I0=I;
  Ip=I+1;

  if(I0== 0)Im=nt-1;
  if(I0==nt-1)Ip=0;

  result=0.;
  if(i<2*nt)
   {
    result=0.;
    switch(m)
     {
      case 0:
       if(j==2*Ip)
        {
         fp=cos(w*(x+h)+point[2*Ip])+kappa;
         f0=0.;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=w*w+gamma*w*h+f*h*h;
        }else if(j== 2*Ip+1)
        {
         fp=-kappa;
         f0=0.;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*I0)
        {
         fp=0.;
         f0=cos(w*x  +point[2*I0])+kappa;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=-2.*w*w+f*h*h;
        }else if(j== 2*I0+1)
        {
         fp=0.;
         f0=-kappa;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*Im)
        {
         fp=0.;
         f0=0.;
         fm=cos(w*(x-h)+point[2*Im])+kappa;
         f=.25*(fp+2*f0+fm);
         result=w*w-gamma*w*h+f*h*h;
        }else if(j== 2*Im+1)
        {
         fp=0.;
         f0=0.;
         fm=-kappa;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*nt)
        {   /* w */
         fp=1.+(x+h)*cos(w*(x+h)+point[2*Ip]);
         f0=1.+ x   *cos(w*x  +point[2*I0]);
         fm=1.+(x-h)*cos(w*(x-h)+point[2*Im]);
         f=.25*(fp+2*f0+fm);
         result=2.*w*(point[2*Ip]-2.*point[2*I0]+point[2*Im])+gamma*(point[2*Ip]-point[2*Im])*h+f*h*h;
         result=result*dwdT;
        }else if(j== 2*nt+1)
        {   /* In */
         fp=-1.;
         f0=-1.;
         fm=-1.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*nt+2)
        {   /* Ic */
         fp=-3.1415926*kappa;
         f0=-3.1415926*kappa;
         fm=-3.1415926*kappa;
         fp=-1.;
         f0=-1.;
         fm=-1.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else
         result=0.;
       break;
      case 1:
       if(j== 2*Ip)
        {
         fp=-kappa;
         f0=0.;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*Ip+1)
        {
         fp=cos(w*(x+h)+point[2*Ip+1])+kappa;
         f0=0.;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=w*w+gamma*w*h+f*h*h;
        }else if(j== 2*I0)
        {
         fp=0.;
         f0=-kappa;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*I0+1)
        {
         fp=0.;
         f0=cos(w*(x  )+point[2*I0+1])+kappa;
         fm=0.;
         f=.25*(fp+2*f0+fm);
         result=-2.*w*w+f*h*h;
        }else if(j== 2*Im)
        {
         fp=0.;
         f0=0.;
         fm=-kappa;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*Im+1)
        {
         fp=0.;
         f0=0.;
         fm=cos(w*(x-h)+point[2*Im+1])+kappa;
         f=.25*(fp+2*f0+fm);
         result=w*w-gamma*w*h+f*h*h;
        }else if(j== 2*nt)
        {   /* w  */
         fp=1.+(x+h)*cos(w*(x+h)+point[2*Ip+1]);
         f0=1.+ x   *cos(w*(x  )+point[2*I0+1]);
         fm=1.+(x-h)*cos(w*(x-h)+point[2*Im+1]);
         f=.25*(fp+2*f0+fm);
         result=2*w*(point[2*Ip+1]-2.*point[2*I0+1]+point[2*Im+1])+gamma*(point[2*Ip+1]-point[2*Im+1])*h+f*h*h;
         result=result*dwdT;
        }else if(j== 2*nt+1)
        {   /* In */
         fp=-1.;
         f0=-1.;
         fm=-1.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else if(j== 2*nt+2)
        {   /* Ic */
/*       fp=3.1415926*kappa;
         f0=3.1415926*kappa;
         fm=3.1415926*kappa;*/
         fp=1.;
         f0=1.;
         fm=1.;
         f=.25*(fp+2*f0+fm);
         result=f*h*h;
        }else
         result=0.;
       break;
     }
   }else{
    /* Phase Constraint */
    if(j==2*(nt-1))result=.5;
     else if(j==2*(nt-1)+1)result=.5;
     else result=0.;
   }

  return(result);
 }

double TGuuPendula(int i,int j,int k, double *point,int nt, double kappa, double gamma, double r)
 {
  static char RoutineName[]={"TGuuPendula"};
  double eps=.001;
  double t;
  double gm,gp;
  double test;
  double result;
  printf("%s\n",RoutineName);fflush(stdout);

  t=point[k];

  point[k]=t+eps/2;
  gp=TGuPendula(i,j,point,nt,kappa,gamma,r);

  point[k]=t-eps/2;
  gm=TGuPendula(i,j,point,nt,kappa,gamma,r);

  point[k]=t;

  result=(gp-gm)/eps;
  test=GuuPendula(i,j,k,point,nt,kappa,gamma,r);
  if(fabs(test-result)>eps)
   {printf("Large error in Guu, %d, %d, %d, diff=%lf, exact=%lf\n",i,j,k,result,test);fflush(stdout);}
  
  return result;
 }

double TGuPendula(int i,int j, double *point,int nt, double kappa, double gamma, double r)
 {
  static char RoutineName[]={"TGuPendula"};
  double eps=.001;
  double t;
  double gm,gp;
  double test;
  double result;
  printf("%s\n",RoutineName);fflush(stdout);

  t=point[j];

  point[j]=t+eps/2;
  gp=GPendula(i,point,nt,kappa,gamma,r);

  point[j]=t-eps/2;
  gm=GPendula(i,point,nt,kappa,gamma,r);

  point[j]=t;

  result=(gp-gm)/eps;
  test=GuPendula(i,j,point,nt,kappa,gamma,r);
  if(fabs(test-result)>eps)
   {printf("Large error in Gu, %d, %d, diff=%lf, exact=%lf\n",i,j,result,test);fflush(stdout);}
  
  return result;
 }

double GuuPendula(int i,int j,int k, double *point,int nt, double kappa, double gamma, double r)
 {
  static char RoutineName[]={"GuuPendula"};
  double w,Ic,In;
  double dwdT,ddwdTdw;
  double x,h;
  double W;
  double result;
  int n;
  int I,m;
  int Im,I0,Ip;
  double f,fm,f0,fp;

/*printf("GuuPendula(%d,%d,%d,0x%8.8x,%d,%lf,%lf,%lf)\n",i,j,k,point,nt,kappa,gamma,r);fflush(stdout);*/

  if(i<0 || i>2*nt+1)
   {
    printf("i<0 || i>n-k, in Guu i=%d, n-k=%d\n",i,2*nt+1);
    abort();
    return(0.);
   }

  if(j<0 || j>2*nt+3)
   {
    printf("j<0 || j>n, in Guu i=%d, j=%d, k=%d, n=%d\n",i,j,k,2*nt+3);
    return(0.);
   }

  if(k<0 || k>2*nt+3)
   {
    printf("k<0 || k>n, in Guu i=%d, j=%d, k=%d, n=%d\n",i,j,k,2*nt+3);
    return(0.);
   }


  if(j>i+nBandsU && j!=2*nt && k!=2*nt)return 0.;
  if(j<i-nBandsL && j!=2*nt && k!=2*nt)return 0.;
  if(j!=k && j!=2*nt && k!=2*nt)return 0.;

  w=2*3.1415926/(10*point[2*nt]);
  dwdT=-10*w*w/(2*3.1415926);
  ddwdTdw=-2*10*w/(2*3.1415926);
  In=point[2*nt+1];
  Ic=3.1415926*kappa*point[2*nt+2];
  Ic=point[2*nt+2];

/*
 
   EVALUATES GUU(U,L)
                                 I
      U IS:
             1     2     3        N-4  N-3   N-2  N-1 N
 
           P1(1),P2(1),P1(2),...,P1(M),P2(M), T , In, Ic
 
*/

  h=r*2.*3.1415926/nt;

  I=i/2;
  m=i%2;

  x=I*h;

  Im=I-1;
  I0=I;
  Ip=I+1;

  if(I0== 0)Im=nt-1;
  if(I0==nt-1)Ip=0;

  result=0.;
  if(i<2*nt)
   {
    result=0.;
    switch(m)
     {
      case 0:
       if(j==2*Ip)
        {
         if(k==2*Ip)
          {
           fp=-sin(w*(x+h)+point[2*Ip]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k==2*nt)
          {
           fp=-(x+h)*sin(w*(x+h)+point[2*Ip]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=2*w+gamma*h+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*I0)
        {
         if(k==2*I0)
          {
           fp=0.;
           f0=-sin(w*x  +point[2*I0]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k==2*nt)
          {
           fp=0.;
           f0=-x*sin(w*x  +point[2*I0]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=-4*w+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*Im)
        {
         if(k==2*Im)
          {
           fp=0.;
           f0=0.;
           fm=-sin(w*(x-h)+point[2*Im]);
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k==2*nt)
          {
           fp=0.;
           f0=0.;
           fm=-(x-h)*sin(w*(x-h)+point[2*Im]);
           f=.25*(fp+2*f0+fm);
           result=2*w-gamma*h+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*nt)
        {   /* w */
         if(k==2*Ip)
          {
           fp=-(x+h)*sin(w*(x+h)+point[2*Ip]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=2.*w+gamma*h+f*h*h;
           result=result*dwdT;
          }else if(k==2*I0)
          {
           fp=0.;
           f0=-x*sin(w*x  +point[2*I0]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=-4*w+f*h*h;
           result=result*dwdT;
          }else if(k==2*Im)
          {
           fp=0.;
           f0=0.;
           fm=-(x-h)*sin(w*(x-h)+point[2*Im]);
           f=.25*(fp+2*f0+fm);
           result=2.*w-gamma*h+f*h*h;
           result=result*dwdT;
          }else if(k==2*nt)
          {
           fp=-(x+h)*(x+h)*sin(w*(x+h)+point[2*Ip]);
           f0=- x   * x   *sin(w*x  +point[2*I0]);
           fm=-(x-h)*(x-h)*sin(w*(x-h)+point[2*Im]);
           f=.25*(fp+2*f0+fm);
           result=2*(point[2*Ip]-2.*point[2*I0]+point[2*Im])+f*h*h;
           result=result*dwdT*dwdT;
           fp=1.+(x+h)*cos(w*(x+h)+point[2*Ip]);
           f0=1.+ x   *cos(w*x  +point[2*I0]);
           fm=1.+(x-h)*cos(w*(x-h)+point[2*Im]);
           f=.25*(fp+2*f0+fm);
/*         result+=2*w*w*w/6.283185/6.283185*(2.*w*(point[2*Ip]-2.*point[2*I0]+point[2*Im])+gamma*(point[2*Ip]-point[2*Im])*h+f*h*h);*/
           result+=dwdT*ddwdTdw*(2.*w*(point[2*Ip]-2.*point[2*I0]+point[2*Im])+gamma*(point[2*Ip]-point[2*Im])*h+f*h*h);
          }else result=0.;
        }else result=0.;
       break;
      case 1:
       if(j== 2*Ip+1)
        {
         if(k== 2*Ip+1)
          {
           fp=-sin(w*(x+h)+point[2*Ip+1]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k== 2*nt)
          {
           fp=-(x+h)*sin(w*(x+h)+point[2*Ip+1]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=2*w+gamma*h+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*I0+1)
        {
         if(k== 2*I0+1)
          {
           fp=0.;
           f0=-sin(w*(x  )+point[2*I0+1]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k== 2*nt)
          {
           fp=0.;
           f0=-x*sin(w*(x  )+point[2*I0+1]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=-4*w+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*Im+1)
        {
         if(k== 2*Im+1)
          {
           fp=0.;
           f0=0.;
           fm=-sin(w*(x-h)+point[2*Im+1]);
           f=.25*(fp+2*f0+fm);
           result=f*h*h;
          }else if(k== 2*nt)
          {
           fp=0.;
           f0=0.;
           fm=-(x-h)*sin(w*(x-h)+point[2*Im+1]);
           f=.25*(fp+2*f0+fm);
           result=2*w-gamma*h+f*h*h;
           result=result*dwdT;
          }else result=0.;
        }else if(j== 2*nt)
        {   /* w  */
         if(k== 2*Ip+1)
          {
           fp=-(x+h)*sin(w*(x+h)+point[2*Ip+1]);
           f0=0.;
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=2*w+gamma*h+f*h*h;
           result=result*dwdT;
          }else if(k== 2*I0+1)
          {
           fp=0.;
           f0=-x*sin(w*(x  )+point[2*I0+1]);
           fm=0.;
           f=.25*(fp+2*f0+fm);
           result=-4*w+f*h*h;
           result=result*dwdT;
          }else if(k== 2*Im+1)
          {
           fp=0.;
           f0=0.;
           fm=-(x-h)*sin(w*(x-h)+point[2*Im+1]);
           f=.25*(fp+2*f0+fm);
           result=2*w-gamma*h+f*h*h;
           result=result*dwdT;
          }else if(k== 2*nt)
          {
           fp=-(x+h)*(x+h)*sin(w*(x+h)+point[2*Ip+1]);
           f0=- x   * x   *sin(w*(x  )+point[2*I0+1]);
           fm=-(x-h)*(x-h)*sin(w*(x-h)+point[2*Im+1]);
           f=.25*(fp+2*f0+fm);
           result=2*(point[2*Ip+1]-2.*point[2*I0+1]+point[2*Im+1])+f*h*h;
           result=result*dwdT*dwdT;
           fp=1.+(x+h)*cos(w*(x+h)+point[2*Ip+1]);
           f0=1.+ x   *cos(w*x  +point[2*I0+1]);
           fm=1.+(x-h)*cos(w*(x-h)+point[2*Im+1]);
           f=.25*(fp+2*f0+fm);
/*         result+=2*w*w*w/6.283185/6.283185*(2.*w*(point[2*Ip+1]-2.*point[2*I0+1]+point[2*Im+1])+gamma*(point[2*Ip+1]-point[2*Im+1])*h+f*h*h);*/
           result+=dwdT*ddwdTdw*(2.*w*(point[2*Ip+1]-2.*point[2*I0+1]+point[2*Im+1])+gamma*(point[2*Ip+1]-point[2*Im+1])*h+f*h*h);
          }else result=0.;
        }else
         result=0.;
       break;
     }
   }else{
    /* Phase Constraint */
    result=0.;
   }

  return(result);
 }

void MFCurvaturePendula(int n,int k,double *u,double *tangentA,double *tangentB,double *a00, double *a01, double *a11, void *d)
{
  static double *result;
  static int i,j,l;
  static int i0,i1;
  static int verbose,test;
  static int nBands,bandW;
  static int na,nb;
  static int ierr;
  static double e;
  static double scl;
  static int nt;
  static double kappa,gamma,r;

  nt=((struct MFPendulaData*)d)->nt;
  kappa=((struct MFPendulaData*)d)->kappa;
  gamma=((struct MFPendulaData*)d)->gamma;
  r=((struct MFPendulaData*)d)->R;

  verbose=0;
  nBands=nBandsL+nBandsU+1;

  for(j=0;j<n-k-nBorders;j++)
   {
    i0=j-nBandsU;
    if(i0<0)i0=0;
    i1=j+nBandsL;
    if(i1>n-k-nBorders-1)i1=n-k-nBorders-1;
    for(i=i0;i<i1+1;i++)
      A[i-j+nBands-1+ldA*j]=GU(i,j,u,nt,kappa,gamma,r);
   }

  for(i=0;i<n-k-nBorders;i++)
   {
    for(j=0;j<nBorders+k;j++)
      B[i+ldB*j]=GU(i,n-k-nBorders+j,u,nt,kappa,gamma,r);
   }

  for(i=0;i<nBorders;i++)
   {
    for(j=0;j<n-k-nBorders;j++)
      C[i+ldC*j]=GU(n-k-nBorders+i,j,u,nt,kappa,gamma,r);
   }
  for(j=0;j<n-k-nBorders;j++)
   {
    if(j<2*nt)
     {
      C[nBorders+0+ldC*j]=tangentA[j]/nt;
      C[nBorders+1+ldC*j]=tangentB[j]/nt;
     }else{
      C[nBorders+0+ldC*j]=tangentA[j];
      C[nBorders+1+ldC*j]=tangentB[j];
     }
   }

  for(i=0;i<nBorders;i++)
   {
    for(j=0;j<nBorders+k;j++)
      D[i+ldD*j]=GU(n-k-nBorders+i,n-k-nBorders+j,u,nt,kappa,gamma,r);
   }
  for(j=0;j<nBorders+k;j++)
   {
    if(n-k-nBorders+j<2*nt)
     {
      D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j]/nt;
      D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j]/nt;
     }else{
      D[nBorders+0+ldD*j]=tangentA[n-k-nBorders+j];
      D[nBorders+1+ldD*j]=tangentB[n-k-nBorders+j];
     }
   }

/* Calculate the size of the right hand side. */

/* Compute the tangent. */
  if(verbose)
   {
    printf(" System:\n\n");
    for(i=0;i<n-k-nBorders;i++)
     {
      printf("[");
      for(j=0;j<n-k-nBorders;j++)
       {
        if(j>i-nBandsL-1 && j<i+nBandsU+1)
         {
          printf("%7.3lf",A[i-j+nBands-1+ldA*j]);
         }else{
          printf("       ");
         }
       }
      printf(" | ");
      for(j=0;j<nBorders+k;j++)printf("%7.3lf",B[i+ldB*j]);
      printf("]");
      printf(" [%7.3lf]\n",R[i]);
     }
  
    printf("[");
    for(j=0;j<n-k-nBorders;j++)printf("--------");
    printf(" + ");
    for(j=0;j<nBorders+k;j++)printf("--------");
    printf("]");
    printf(" [-------]\n");

    for(i=0;i<nBorders+k;i++)
     {
      printf("[");
      for(j=0;j<n-k-nBorders;j++)printf("%7.3lf",C[i+ldC*j]);
      printf(" | ");
      for(j=0;j<nBorders+k;j++)printf("%7.3lf",D[i+ldC*j]);
      printf("]");
      printf(" [%7.3lf]\n",S[i]);
     }
   }

  na=n-k-nBorders;
  nb=nBorders+k;
  F77_FUNC(dbofa,DBOFA)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,(double*)NULL,Pivots,&ierr);
  if(ierr!=0)
   {
    printf(" Problem with factor, zero on diagonal %d\n",ierr);
    exit(8);
   }

/* -GuuPhiSPhiT */

  for(i=0;i<n-k-nBorders;i++)
   {
    R[i]=0.;
    for(j=0;j<n;j++)
     {
      R[i]+=GUU(i,j,j,u,nt,kappa,gamma,r)*tangentA[j]*tangentA[j];
      for(l=j+1;l<n;l++)
        R[i]+=2*GUU(i,j,l,u,nt,kappa,gamma,r)*tangentA[j]*tangentA[l];
     }
   }

  for(i=0;i<nBorders;i++)
   {
    S[i]=0.;
    for(j=0;j<n;j++)
     {
      S[i]+=GUU(n-k-nBorders+i,j,j,u,nt,kappa,gamma,r)*tangentA[j]*tangentA[j];
      for(l=j+1;l<n;l++)
        S[i]+=2*GUU(n-k-nBorders+i,j,l,u,nt,kappa,gamma,r)*tangentA[j]*tangentA[l];
     }
   }
  for(i=nBorders;i<nBorders+k;i++)S[i]=0.;

  e=0.;
  for(i=0;i<n-k-nBorders;i++)e+=R[i]*R[i]/nt;
  for(i=0;i<nBorders;i++)e+=S[i]*S[i]/nt;
  for(i=nBorders;i<nBorders+k;i++){if(i+n-k-nBorders==2*nt)e+=1.*(S[i]*S[i]);else e+=S[i]*S[i];}
  e=sqrt(e);
  if(verbose){printf("Norm of rhs for a00 is %lf\n",e);fflush(stdout);}

  ierr=0;
  F77_FUNC(dbosl,DBOSL)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,Y,R,S,(double*)NULL,Pivots,&ierr);

/* Extract the result */

  for(i=0;i<n-k-nBorders;i++)
    a00[i]=R[i];
  for(i=0;i<nBorders+k;i++)
    a00[n-k-nBorders+i]=S[i];

  for(i=0;i<n-k-nBorders;i++)
   {
    R[i]=0.;
    for(j=0;j<n;j++)
     {
      R[i]+=GUU(i,j,j,u,nt,kappa,gamma,r)*tangentA[j]*tangentB[j];
      for(l=j+1;l<n;l++)
        R[i]+=GUU(i,j,l,u,nt,kappa,gamma,r)*(tangentA[j]*tangentB[l]+tangentA[l]*tangentB[j]);
     }
   }

  for(i=0;i<nBorders;i++)
   {
    S[i]=0.;
    for(j=0;j<n;j++)
     {
      S[i]+=GUU(n-k-nBorders+i,j,j,u,nt,kappa,gamma,r)*tangentA[j]*tangentB[j];
      for(l=j+1;l<n;l++)
        S[i]+=GUU(n-k-nBorders+i,j,l,u,nt,kappa,gamma,r)*(tangentA[j]*tangentB[l]+tangentA[l]*tangentB[j]);
     }
   }
  for(i=nBorders;i<nBorders+k;i++)S[i]=0.;

  e=0.;
  for(i=0;i<n-k-nBorders;i++)e+=R[i]*R[i]/nt;
  for(i=0;i<nBorders;i++)e+=S[i]*S[i]/nt;
  for(i=nBorders;i<nBorders+k;i++){if(i+n-k-nBorders==2*nt)e+=1.*(S[i]*S[i]);else e+=S[i]*S[i];}
  e=sqrt(e);
  if(verbose){printf("Norm of rhs for a01 is %lf\n",e);fflush(stdout);}

  ierr=0;
  F77_FUNC(dbosl,DBOSL)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,Y,R,S,(double*)NULL,Pivots,&ierr);

/* Extract the result */

  for(i=0;i<n-k-nBorders;i++)
    a01[i]=R[i];
  for(i=0;i<nBorders+k;i++)
    a01[n-k-nBorders+i]=S[i];

/* -GuuPhiSPhiT */

  for(i=0;i<n-k-nBorders;i++)
   {
    R[i]=0.;
    for(j=0;j<n;j++)
     {
      R[i]+=GUU(i,j,j,u,nt,kappa,gamma,r)*tangentB[j]*tangentB[j];
      for(l=j+1;l<n;l++)
       R[i]+=2*GUU(i,j,l,u,nt,kappa,gamma,r)*tangentB[j]*tangentB[l];
     }
   }

  for(i=0;i<nBorders;i++)
   {
    S[i]=0.;
    for(j=0;j<n;j++)
     {
      S[i]+=GUU(n-k-nBorders+i,j,j,u,nt,kappa,gamma,r)*tangentB[j]*tangentB[j];
      for(l=j+1;l<n;l++)
       S[i]+=2*GUU(n-k-nBorders+i,j,l,u,nt,kappa,gamma,r)*tangentB[j]*tangentB[l];
     }
   }
  for(i=nBorders;i<nBorders+k;i++)S[i]=0.;

  e=0.;
  for(i=0;i<n-k-nBorders;i++)e+=R[i]*R[i]/nt;
  for(i=0;i<nBorders;i++)e+=S[i]*S[i]/nt;
  for(i=nBorders;i<nBorders+k;i++){if(i+n-k-nBorders==2*nt)e+=1.*(S[i]*S[i]);else e+=S[i]*S[i];}
  e=sqrt(e);
  if(verbose){printf("Norm of rhs for a11 is %lf\n",e);fflush(stdout);}

  ierr=0;
  F77_FUNC(dbosl,DBOSL)(A,&ldA,&na,&nBandsL,&nBandsU,B,&ldB,&nb,C,&ldC,D,&ldD,X,Y,R,S,(double*)NULL,Pivots,&ierr);

/* Extract the result */

  for(i=0;i<n-k-nBorders;i++)
    a11[i]=R[i];
  for(i=0;i<nBorders+k;i++)
    a11[n-k-nBorders+i]=S[i];

  return;
 }

void MFWritePendulaData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWritePendulaData"};
  struct MFPendulaData *data;
  int verbose;

  verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  data=(struct MFPendulaData*)d;
  fprintf(fid,"%d %lf %lf %lf\n",data->nt,data->kappa,data->gamma,data->R);
  fflush(fid);
  if(verbose)
   {
    printf("  nt=%d\n",data->nt);
    printf("  kappa=%lf\n",data->kappa);
    printf("  gamma=%lf\n",data->gamma);
    printf("  R=%lf\n",data->R);fflush(stdout);
   }
  return;
 }

MFImplicitMF MFReadPendula(FILE *fid)
 {
  static char RoutineName[]={"MFReadPendula"};
  MFImplicitMF pendula;
  int nt=0;
  double kappa=0.;
  double gamma=0.;
  double R=0.;
  int verbose;

  verbose=0;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  fscanf(fid,"%d %lf %lf %lf\n",&nt,&kappa,&gamma,&R);
  pendula=MFIMFCreatePendula(nt,kappa,gamma,R);

  return pendula;
 }

int MFPendulaProjectToSave(MFNVector u, double *x, void *d)
 {
  int i,n;
  struct MFPendulaData *data;

  data=(struct MFPendulaData*)d;

  n=2*(data->nt)+3;

  if(x==(double*)NULL)return n;
  for(i=0;i<n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFPendulaProjectToDraw(MFNVector u, double *x, void *d)
 {
  int i,n;
  struct MFPendulaData *data;

  data=(struct MFPendulaData*)d;

  n=2*(data->nt)+3;

  if(x==(double*)NULL)return 3;
  x[0]=MFNV_C(u,n-3);
  x[1]=MFNV_C(u,n-2);
  x[2]=MFNV_C(u,n-1);

  return 0;
 }

int MFPendulaProjectForBB(MFNVector u, double *x, void *d)
 {
  int i,n;
  struct MFPendulaData *data;

  data=(struct MFPendulaData*)d;

  n=2*(data->nt)+3;

  if(x==(double*)NULL)return 3;
  x[0]=MFNV_C(u,n-3);
  x[1]=MFNV_C(u,n-2);
  x[2]=MFNV_C(u,n-1);

  return 0;
 }
