/* 
    @(#)MFPendulaNRegion.c	1.1
    02/07/24 09:05:26
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */
/*              October 6, 2000   Added Polyhedron    */

#include <stdio.h>
#include <MFNVector.h>
#include <MFNRegion.h>
#include <stdlib.h>
#include <math.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNRegionErrorMsg[256]="";

int PendulaTest(MFNVector,void*);
void PendulaFree(void*);
int PendulaFourTest(MFNVector,void*);
void PendulaFourFree(void*);

double *MFNV_CStar(MFNVector);

MFNRegion MFNRegionCreatePendula(int n, double kappa,double gamma,int WN,double IcMin,double IcMax, double InMin,double InMax,double TMax)
 {
  static char RoutineName[]={"MFNRegionCreatePendula"};
  MFNRegion pendula;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreatePendula++;
  starttime=clock();
#endif

  pendula=MFNRegionCreateBaseClass("Pendula");
  MFNRegionSetTest(pendula,PendulaTest);
  MFNRegionSetFreeData(pendula,PendulaFree);

  data=(void*)malloc(6*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(pendula);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePendula+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=kappa;
  data[1]=IcMin;
  data[2]=IcMax;
  data[3]=InMin;
  data[4]=InMax;
  data[5]=TMax;

  MFNRegionSetData(pendula,data);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePendula+=clock()-starttime;
#endif
  return(pendula);
 }

int PendulaTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"PendulaTest"};
  double kappa,T,In,Ic;
  double IcMin,IcMax,InMin,InMax,TMax;
  int i;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledPendulaTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  kappa=((double*)data)[0];
  IcMin=((double*)data)[1];   /* -1. */
  IcMax=((double*)data)[2];   /*  1. */
  InMin=((double*)data)[3];   /*  0. */
  InMax=((double*)data)[4];   /* 10. */
  TMax=((double*)data)[5];    /* 50. */

  T=10*x[n-3];
  In=x[n-2];
  Ic=x[n-1];  /* Ic=3.1415926*kappa*x[n-1] */

  rc=1;
  if(Ic<IcMin)rc=0;
  if(Ic>IcMax)rc=0;
  if(In<InMin)rc=0;
  if(In>InMax)rc=0;
  if(T>TMax)rc=0;

  if(0){printf(" Pendula Test (Ic,In,T)=(%lf,%lf,%lf)  Ic in [%lf,%lf], In in [%lf,%lf], T<%lf, rc=%d\n",Ic,In,T,IcMin,IcMax,InMin,InMax,TMax,rc);fflush(stdout);}

#ifdef MFTIMINGS
    MFTimePendulaTest+=clock()-starttime;
#endif
  return rc;
 }

void PendulaFree(void *data)
 {
  static char RoutineName[]={"PendulaFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledPendulaFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimePendulaFree+=clock()-starttime;
#endif
  return;
 }

MFNRegion MFNRegionCreatePendulaFour(int n, int WN, double kappaMin,double kappaMax, double gammaMin,double gammaMax,double IcMin,double IcMax, double InMin,double InMax,double TMax)
 {
  static char RoutineName[]={"MFNRegionCreatePendulaFour"};
  MFNRegion pendula;
  double *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNRegionCreatePendulaFour++;
  starttime=clock();
#endif

  pendula=MFNRegionCreateBaseClass("PendulaFour");
  MFNRegionSetTest(pendula,PendulaFourTest);
  MFNRegionSetFreeData(pendula,PendulaFourFree);

  data=(void*)malloc(8*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFNRegionErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(double));
    printf("%s -- %s\n",RoutineName,MFNRegionErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNRegionErrorMsg,__LINE__,__FILE__);
    free(pendula);
#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePendulaFour+=clock()-starttime;
#endif
    return (MFNRegion)NULL;
   }

  data[0]=kappaMin;
  data[1]=IcMin;
  data[2]=IcMax;
  data[3]=InMin;
  data[4]=InMax;
  data[5]=TMax;
  data[6]=kappaMax;
  data[7]=gammaMin;
  data[8]=gammaMax;

  MFNRegionSetData(pendula,data);

/*printf("MFNRegion - PendulaFour, n=%d Winding Number=%d\n",n,WN);
  printf("                         kappa in %lf,%lf\n",kappaMin,kappaMax);
  printf("                         gamma in %lf,%lf\n",gammaMin,gammaMax);
  printf("                         Ic    in %lf,%lf\n",IcMin,IcMax);
  printf("                         In    in %lf,%lf\n",InMin,InMax);
  printf("                         T     < %lf\n",TMax);*/
  MFNRegionSetFreeData(pendula,PendulaFourFree);

#ifdef MFTIMINGS
    MFTimeMFNRegionCreatePendulaFour+=clock()-starttime;
#endif
  return(pendula);
 }

int PendulaFourTest(MFNVector v,void *data)
 {
  static char RoutineName[]={"PendulaFourTest"};
  double kappa,gamma,T,In,Ic;
  double IcMin,IcMax,InMin,InMax,TMax;
  double kappaMin,kappaMax,gammaMin,gammaMax;
  int i;
  int rc;
  int n;
  double *x;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledPendulaFourTest++;
  starttime=clock();
#endif
  n=MFNV_NC(v);
  x=MFNV_CStar(v);

  kappaMin=((double*)data)[0];
  IcMin=((double*)data)[1];   /* -1. */
  IcMax=((double*)data)[2];   /*  1. */
  InMin=((double*)data)[3];   /*  0. */
  InMax=((double*)data)[4];   /* 10. */
  TMax=((double*)data)[5];    /* .1  */
  kappaMax=((double*)data)[6];
  gammaMin=((double*)data)[7];
  gammaMax=((double*)data)[8];

  T=x[n-5];
  In=x[n-4];
  Ic=x[n-3];  /* Ic=3.1415926*kappa*x[n-1] */
  kappa=x[n-2];
  gamma=x[n-1];

/*printf("%s\n",RoutineName);
  printf(" %lf < %lf (Ic) < %lf \n",IcMin,Ic,IcMax);
  printf(" %lf < %lf (In) < %lf \n",InMin,In,InMax);
  printf(" %lf > %lf (T) \n",TMax,T);
  printf(" %lf < %lf (kappa) < %lf \n",kappaMin,kappa,kappaMax);
  printf(" %lf < %lf (gamma) < %lf \n",gammaMin,gamma,gammaMax);*/

  rc=1;
  if(Ic<IcMin)rc=0;
  if(Ic>IcMax)rc=0;
  if(In<InMin)rc=0;
  if(In>InMax)rc=0;
  if(T>TMax)rc=0;
  if(kappa<kappaMin)rc=0;
  if(kappa>kappaMax)rc=0;
  if(gamma<gammaMin)rc=0;
  if(gamma>gammaMax)rc=0;

#ifdef MFTIMINGS
    MFTimePendulaFourTest+=clock()-starttime;
#endif
  return rc;
 }

void PendulaFourFree(void *data)
 {
  static char RoutineName[]={"PendulaFourFree"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledPendulaFourFree++;
  starttime=clock();
#endif
  free(data);
#ifdef MFTIMINGS
    MFTimePendulaFourFree+=clock()-starttime;
#endif
  return;
 }
