/* 
    @(#)MFPlane.c	1.14
    03/07/11 17:00:39
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */
#include <MFImplicitMF.h>
#include <stdlib.h>
#include <string.h>

static char MFPlaneMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

int MFProjectPlane(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentPlane(int,int,MFNVector,MFNKMatrix,void*);
double MFScalePlane(int,int,MFNVector,MFNKMatrix,void*);
MFImplicitMF MFReadPlane(FILE*);

int MFPlaneProjectToSave(MFNVector,double*,void*);
int MFPlaneProjectToDraw(MFNVector,double*,void*);
int MFPlaneProjectForBB(MFNVector,double*,void*);
MFImplicitMF MFIMFCreatePlane();

MFImplicitMF MFIMFCreatePlane()
 {
  static char RoutineName[]={"MFIMFCreatePlane"};
  MFImplicitMF plane;
  MFNSpace space;

  plane=MFIMFCreateBaseClass(2,2,"Plane");

  space=MFCreateNSpace(2);
  MFIMFSetSpace(plane,space);
  MFFreeNSpace(space);

  MFIMFSetProject(plane,MFProjectPlane);
  MFIMFSetTangent(plane,MFTangentPlane);
  MFIMFSetScale(plane,MFScalePlane);
  MFIMFSetProjectForSave(plane,MFPlaneProjectToSave);
  MFIMFSetProjectForDraw(plane,MFPlaneProjectToDraw);
  MFIMFSetProjectForBB(plane,MFPlaneProjectForBB);

  return plane;
 }

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

int MFProjectPlane(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectPlane"};
  double *u,*u0,*Phi;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  u[0]=u0[0];
  u[1]=u0[1];
  *index=0;
  return 1;
 }

int MFTangentPlane(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentPlane"};
  double *u,*Phi;

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  Phi[0]=1.;
  Phi[1]=0.;
  Phi[2]=0.;
  Phi[3]=1.;

  return 1;
 }

double MFScalePlane(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFScalePlane"};
  return .1;
 }

MFImplicitMF MFReadPlane(FILE *fid)
 {
  static char RoutineName[]={"MFReadPlane"};
  MFImplicitMF plane;

  plane=MFIMFCreatePlane();

  return plane;
 }

int MFPlaneProjectToSave(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 2;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);

  return 0;
 }

int MFPlaneProjectToDraw(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 2;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);

  return 0;
 }

int MFPlaneProjectForBB(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 2;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);

  return 0;
 }
