/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFPrint.h>
#include <MFError.h>
#include <MFKVector.h>
#include <MFAtlasFriends.h>

void MFSetError(int,char*,char*,int,char*);
static char MFPrintErrorMsg[256]="";

#define DENSE 0
#define VECTOR 1
#define LOCA 2

void MFPrintAtlas(FILE *fid,MFAtlas A)
 {
  static char RoutineName[]={"MFPrintAtlas"};
  int i,j;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFPrintErrorMsg,"Atlas (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  fprintf(fid,"     Atlas: \n\n");
  fprintf(fid,"       Base space dimension: %d \n",MFAtlasK(A));
  fprintf(fid,"       Target space dimension: %d \n",MFAtlasN(A));

  fprintf(fid,"\n       %d Half Spaces \n",MFAtlasNumberOfHalfSpaces(A));
  for(i=0;i<MFAtlasNumberOfHalfSpaces(A);i++)
   {
    fprintf(fid,"        %d (%d), [%d,%d]",i,MFAtlasHalfSpaceIndex(A,i),MFAtlasHalfSpaceLeftChart(A,i),MFAtlasHalfSpaceRightChart(A,i));
    fprintf(fid,"  x.");
    MFPrintKVector(fid,MFAtlasHalfSpaceNormal(A,i));
    fprintf(fid,"-%lf<0\n",MFAtlasHalfSpaceOrigin(A,i));
   }

  fprintf(fid,"\n       %d Charts \n",MFAtlasNumberOfCharts(A));
  for(i=0;i<MFAtlasNumberOfCharts(A);i++)
   {
    fprintf(fid,"         Chart %d\n",i);
    fprintf(fid,"           Center: ");
    MFPrintNVector(fid,MFChartCenter(MFAtlasChart(A,i)));
    fprintf(fid,"           Radius %lf\n",MFAtlasChartRadius(A,i));
    fprintf(fid,"           Basis for tangent space:\n");
    MFPrintNKMatrix(fid,MFChartTangentSpace(MFAtlasChart(A,i)));
    fprintf(fid,"           Polytope:\n");
    MFPrintPolytope(fid,MFChartPolytope(MFAtlasChart(A,i)));
   }

  return;
 }

void MFPrintKVector(FILE *fid,MFKVector s)
 {
  static char RoutineName[]={"MFPrintKVector"};
  int i,n;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(s==(MFKVector)NULL)
   {
    sprintf(MFPrintErrorMsg,"Vector (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  n=MFKV_NC(s);
  fprintf(fid,"(");
  for(i=0;i<n;i++)
   {
    if(i>0)fprintf(fid,",");
    fprintf(fid,"%lf",MFKV_C(s,i));
   }
  fprintf(fid,")");

  return;
 }

void MFPrintChart(FILE *fid,MFChart chart)
 {
  static char RoutineName[]={"MFPrintChart"};
  MFNVector phi;
  int i;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(chart==(MFChart)NULL)
   {
    sprintf(MFPrintErrorMsg,"Chart (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  fprintf(fid,"Chart center ");
  MFPrintNVector(fid,MFChartCenter(chart));
  fprintf(fid,"\n");
  fprintf(fid,"      R %lf\n",MFChartRadius(chart));
  fprintf(fid,"      Basis for TS ");
  for(i=0;i<MFChartK(chart);i++)
   {
    if(i>0)fprintf(fid,"                   ");
    phi=MFMColumn(MFChartTangentSpace(chart),i);
    MFPrintNVector(fid,phi);fprintf(fid,"\n");
    MFFreeNVector(phi);
   }
  fprintf(fid,"      Polytope\n");MFPrintPolytope(fid,MFChartPolytope(chart));
  fflush(fid);
  return;
 }

void MFPrintPolytope(FILE *file,MFPolytope P)
 {
  static char RoutineName[]={"MFPrintPolytope"};
  int i,j;
  MFKVector v;

  if(file==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(P==(MFPolytope)NULL)
   {
    sprintf(MFPrintErrorMsg,"Polytope (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(P==(MFPolytope)NULL)return;
  v=MFCreateKVector(MFPolytopeDimension(P));

  fprintf(file,"               dimension %d\n",MFPolytopeDimension(P));

  fprintf(file,"               %d vertices\n",MFPolytopeNumberOfVertices(P));
  for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
   {
    fprintf(file,"                   %d (",i);fflush(file);
    MFPolytopeVertex(P,i,v);
    MFPrintKVector(file,v);
    fprintf(file,") indices %d (",MFPolytopeNumberOfVertexIndices(P,i));fflush(file);
    for(j=0;j<MFPolytopeNumberOfVertexIndices(P,i);j++)
     {
      if(j>0)fprintf(file,",");
      fprintf(file,"%d",MFPolytopeVertexIndex(P,i,j));fflush(file);
     }
      
    fprintf(file,")");
    fprintf(file," r=%lf",MFPolytopeRadiusOfVertex(P,i));
    fprintf(file,"\n");
    fflush(file);
   }

  fprintf(file,"               %d faces\n",MFPolytopeNumberOfFaces(P));
  fflush(file);
  for(i=0;i<MFPolytopeNumberOfFaces(P);i++)
   {
    fprintf(file,"                   %d, index %d contains %d vertices\n",i,MFPolytopeFaceIndex(P,i),MFPolytopeNumberOfVerticesOnFace(P,i));
    if(MFPolytopeFaceNormal(P,i)!=(MFKVector)NULL)
     {
      fprintf(file,"                          ");
      MFPrintKVector(file,MFPolytopeFaceNormal(P,i));
      fprintf(file,".s<%lf    0x%8.8x\n",MFPolytopeFaceOrigin(P,i),MFPolytopeFaceNormal(P,i));
      fflush(file);
     }
   }

  MFFreeKVector(v);
  return;
 }

void MFPrintNKMatrix(FILE *fid,MFNKMatrix L)
 {
  static char RoutineName[]={"MFPrintNKMatrix"};
  int i,n,k;
  MFKVector s;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(L==(MFNKMatrix)NULL)
   {
    sprintf(MFPrintErrorMsg,"Matrix (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  n=MFNKMatrixN(L);
  k=MFNKMatrixK(L);
  s=MFCreateKVector(k);
  if(n<0)  /* i.e. LOCA */
   {
    MFNVector col;
    printf("[\n");
    for(i=0;i<k;i++)
     {
      col=MFMColumn(L,i);
      MFPrintNVector(fid,col);
      fprintf(fid,"\n");
      MFFreeNVector(col);
     }
    printf("]^T\n");
   }else if(n<16)
   {
    for(i=0;i<n;i++)
     {
      MFMRow(L,i,s);
      MFPrintKVector(fid,s);
      fprintf(fid,"\n");
     }
   }else{
    for(i=0;i<8;i++)
     {
      MFMRow(L,i,s);
      MFPrintKVector(fid,s);
      fprintf(fid,"\n");
     }
    fprintf(fid,"...\n");
    for(i=n-8;i<n;i++)
     {
      MFMRow(L,i,s);
      MFPrintKVector(fid,s);
      fprintf(fid,"\n");
     }
   }
  MFFreeKVector(s);
  return;
 }

void MFPrintPolytopeTerse(FILE *file,MFPolytope P)
 {
  static char RoutineName[]={"MFPrintPolytopeTerse"};
  int i,j;

  if(file==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(P==(MFPolytope)NULL)
   {
    sprintf(MFPrintErrorMsg,"Polytope (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }
  if(P==(MFPolytope)NULL)return;

  fprintf(file,"    Dimension %d\n",MFPolytopeDimension(P));

  fprintf(file,"    Vertices\n",MFPolytopeNumberOfVertices(P));
  for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
   {
    fprintf(file,"         %3d [",i);fflush(file);
    for(j=0;j<MFPolytopeNumberOfVertexIndices(P,i);j++)
     {
      if(j>0)fprintf(file,",");
      fprintf(file,"%3d",MFPolytopeVertexIndex(P,i,j));fflush(file);
     }
      
    fprintf(file,"]");
    fprintf(file,"\n");
    fflush(file);
   }

  fprintf(file,"    Faces\n",MFPolytopeNumberOfFaces(P));
  fflush(file);
  for(i=0;i<MFPolytopeNumberOfFaces(P);i++)
   {
    fprintf(file,"     %3d %3d vertices\n",MFPolytopeFaceIndex(P,i),MFPolytopeNumberOfVerticesOnFace(P,i));
   }

  fflush(file);
  return;
 }

void MFPrintAtlasFaceList(FILE *fid,MFAtlas A)
 {
  static char RoutineName[]={"MFPrintAtlasFaceList"};
  int i,j;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFPrintErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFPrintErrorMsg,"Atlas (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFPrintErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFPrintErrorMsg,__LINE__,__FILE__);
    return;
   }

  fprintf(fid,"\n     %3d Half Spaces \n",MFAtlasNumberOfHalfSpaces(A));
  for(i=0;i<MFAtlasNumberOfHalfSpaces(A);i++)
   {
    if(MFAtlasIsHalfSpaceHyperCube(A,i))
     {
      fprintf(fid,"     %3d, Hypercube face\n",i);
     }else{
      fprintf(fid,"     %3d, [%3d,%3d]\n",i,MFAtlasHalfSpaceLeftChart(A,i),MFAtlasHalfSpaceRightChart(A,i));
     }
   }

  return;
 }
