/* 
    @(#)MFSphere.c	1.16
    03/07/11 17:00:39
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   February 22, 1999                     */

#include <MFImplicitMF.h>
#include <MFPrint.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static char MFSphereMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

void MFFreeSphereData(void*);
int MFProjectSphere(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFTangentSphere(int,int,MFNVector,MFNKMatrix,void*);
double MFScaleSphere(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteSphereData(FILE*,void*);
MFImplicitMF MFReadSphere(FILE*);

int MFSphereProjectToSave(MFNVector,double*,void*);
int MFSphereProjectToDraw(MFNVector,double*,void*);
int MFSphereProjectForBB(MFNVector,double*,void*);

MFImplicitMF MFIMFCreateSphere(double x,double y,double z,double R)
 {
  static char RoutineName[]={"MFIMFCreateSphere"};
  MFImplicitMF sphere;
  MFNSpace space;
  double *data;

  sphere=MFIMFCreateBaseClass(3,2,"Sphere");

  space=MFCreateNSpace(3);
  MFIMFSetSpace(sphere,space);
  MFFreeNSpace(space);

  data=(double*)malloc(4*sizeof(double)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFSphereMFErrorMsg,"Out of memory, trying to allocate %d bytes",4*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFSphereMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFSphereMFErrorMsg,__LINE__,__FILE__);
    free(sphere);
    return (MFImplicitMF)NULL;
   }
  data[0]=x;
  data[1]=y;
  data[2]=z;
  data[3]=R;
  MFIMFSetData(sphere,(void*)data);
  MFIMFSetFreeData(sphere,MFFreeSphereData);
  MFIMFSetProject(sphere,MFProjectSphere);
  MFIMFSetTangent(sphere,MFTangentSphere);
  MFIMFSetR(sphere,-1.);
  MFIMFSetScale(sphere,MFScaleSphere);
  MFIMFSetWriteData(sphere,MFWriteSphereData);
  MFIMFSetProjectForSave(sphere,MFSphereProjectToSave);
  MFIMFSetProjectForDraw(sphere,MFSphereProjectToDraw);
  MFIMFSetProjectForBB(sphere,MFSphereProjectForBB);

  return sphere;
 }

void MFFreeSphereData(void *d)
 {
  static char RoutineName[]={"MFFreeSphereData"};
  free(d);
  return;
 }

int MFProjectSphere(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectSphere"};
  double t,A,B,C,R,nx,ny,nz;
  double u0x,u0y,u0z;
  double v0x,v0y,v0z;
  double v1x,v1y,v1z;
  double err;
  double *u,*u0,*Phi;
  int verbose=0;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  v0x=Phi[0];
  v0y=Phi[1];
  v0z=Phi[2];
  v1x=Phi[0+3];
  v1y=Phi[1+3];
  v1z=Phi[2+3];
  nx=v0y*v1z-v0z*v1y;
  ny=v0z*v1x-v0x*v1z;
  nz=v0x*v1y-v0y*v1x;
  u0x=u0[0]-((double*)d)[0];
  u0y=u0[1]-((double*)d)[1];
  u0z=u0[2]-((double*)d)[2];
  R=((double*)d)[3];

  if(verbose)
   {
    printf("In Project sphere\n");
    printf(" R %lf o=(%lf,%lf,%lf)\n",R,((double*)d)[0],((double*)d)[1],((double*)d)[2]);fflush(stdout);
    printf(" u0 (%lf,%lf,%lf) : ",u0x,u0y,u0z);MFPrintNVector(stdout,vu0);printf("\n");
    printf(" t1 (%lf,%lf,%lf) t2 (%lf,%lf,%lf) :\n",v0x,v0y,v0z,v1x,v1y,v1z);MFPrintNKMatrix(stdout,mPhi);printf("\n");
   }

  A=nx*nx+ny*ny+nz*nz;
  B=nx*u0x+ny*u0y+nz*u0z;
  C=u0x*u0x+u0y*u0y+u0z*u0z-R*R;
  if(verbose)printf(" quadratic  (%lf)t^2+(%lf)t+(%lf)=0\n",A,B,C);
  if(A>0.)
   {
    if(B>0)
      t=(-B+sqrt(B*B-A*C))/A;
     else
      t=(-B-sqrt(B*B-A*C))/A;
   }else if(A==0.)
     t=-.5*C/B;
    else return 0;

  u[0]=u0[0]+t*nx;
  u[1]=u0[1]+t*ny;
  u[2]=u0[2]+t*nz;

  if(verbose)
   {
    printf(" n  (%lf,%lf,%lf)\n",nz,ny,nz);
    printf(" u  (%lf,%lf,%lf) : ",u[0],u[1],u[2]);MFPrintNVector(stdout,vu);printf("\n");

/* Test */

    err=fabs((u[0]-u0[0])*Phi[3]+(u[1]-u0[1])*Phi[4]+(u[2]-u0[2])*Phi[5]);
    err+=fabs((u[0]-u0[0])*Phi[0]+(u[1]-u0[1])*Phi[1]+(u[2]-u0[2])*Phi[2]);
    err+=fabs(sqrt( pow(u[0]-((double*)d)[0],2)+pow(u[1]-((double*)d)[1],2)+pow(u[2]-((double*)d)[2],2))-((double*)d)[3]);
    printf("Error %le\n",err);
   }

  *index=0;
  return 1;
 }

int MFTangentSphere(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentSphere"};
  double s;
  double u0x,u0y,u0z;
  double v0x,v0y,v0z;
  double v1x,v1y,v1z;
  double err;
  double *u,*Phi;
  int verbose=0;

  if(verbose)printf("In %s\n",RoutineName);

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  u0x=(u[0]-((double*)d)[0])/((double*)d)[3];
  u0y=(u[1]-((double*)d)[1])/((double*)d)[3];
  u0z=(u[2]-((double*)d)[2])/((double*)d)[3];
  if(verbose)
   {
    printf(" R %lf o=(%lf,%lf,%lf)\n",((double*)d)[3],((double*)d)[0],((double*)d)[1],((double*)d)[2]);fflush(stdout);
    printf(" u0 (%lf,%lf,%lf) : ",u0x,u0y,u0z);MFPrintNVector(stdout,vu);printf("\n");
   }

  if(u0x!=1.)
   {
    v0x=1.-u0x*u0x;
    v0y=  -u0x*u0y;
    v0z=  -u0x*u0z;
   }else if(u0y!=1.)
   {
    v0x=  -u0y*u0x;
    v0y=1.-u0y*u0y;
    v0z=  -u0y*u0z;
   }else{
    v0x=  -u0z*u0x;
    v0y=  -u0z*u0y;
    v0z=1.-u0z*u0z;
   }
  s=1./sqrt(v0x*v0x+v0y*v0y+v0z*v0z);
  v0x=s*v0x;
  v0y=s*v0y;
  v0z=s*v0z;

  v1x=-v0y*u0z+v0z*u0y;
  v1y=-v0z*u0x+v0x*u0z;
  v1z=-v0x*u0y+v0y*u0x;

  Phi[0]=v0x;
  Phi[1]=v0y;
  Phi[2]=v0z;
  Phi[0+3]=v1x;
  Phi[1+3]=v1y;
  Phi[2+3]=v1z;

  if(verbose)
   {
    printf(" t1 (%lf,%lf,%lf) t2 (%lf,%lf,%lf) :\n",v0x,v0y,v0z,v1x,v1y,v1z);MFPrintNKMatrix(stdout,mPhi);printf("\n");

/* Test */
    err=fabs(Phi[0]*Phi[0]+Phi[1]*Phi[1]+Phi[2]*Phi[2]-1.);
    err+=fabs(Phi[3]*Phi[0]+Phi[4]*Phi[1]+Phi[5]*Phi[2]);
    err+=fabs(Phi[3]*Phi[3]+Phi[4]*Phi[4]+Phi[5]*Phi[5]-1.);
    err+=fabs(Phi[0]*u0x+Phi[1]*u0y+Phi[2]*u0z);
    err+=fabs(Phi[3]*u0x+Phi[4]*u0y+Phi[5]*u0z);

    printf("Error %le\n",err);
   }

  return 1;
 }

extern double MFEpsilon;

double MFScaleSphere(int n,int k,MFNVector u,MFNKMatrix Phi,void *d)
 {
  static char RoutineName[]={"MFScaleSphere"};
/*return sqrt(.5*((double*)d)[3]/.1);*/

  return .95*sqrt(2*MFEpsilon)*((double*)d)[3];
 }

void MFWriteSphereData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWriteSphereData"};
  double *data;

  data=(double*)d;

  fprintf(fid,"%lf %lf %lf %lf\n",data[0],data[1],data[2],data[3]);
  fflush(fid);
  return;
 }

MFImplicitMF MFReadSphere(FILE *fid)
 {
  static char RoutineName[]={"MFReadSphere"};
  MFImplicitMF sphere;
  double x=0.;
  double y=0.;
  double z=0.;
  double R=0.;

  fscanf(fid,"%lf %lf %lf %lf\n",&x,&y,&z,&R);
  sphere=MFIMFCreateSphere(x,y,z,R);

  return sphere;
 }

int MFSphereProjectToSave(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 3;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);
  x[2]=MFNV_C(u,2);

  return 0;
 }

int MFSphereProjectToDraw(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 3;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);
  x[2]=MFNV_C(u,2);

  return 0;
 }

int MFSphereProjectForBB(MFNVector u, double *x, void *d)
 {
  if(x==(double*)NULL)return 3;

  x[0]=MFNV_C(u,0);
  x[1]=MFNV_C(u,1);
  x[2]=MFNV_C(u,2);

  return 0;
 }
