/*
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 2, 1999                      */

#include <math.h>
#include <MFAtlas.h>
#include <MFKVector.h>
#include <MFNVector.h>
#include <MFChart.h>
#include <MFNRegion.h>
#include <MFError.h>
#include <MFPrint.h>
#include <stdlib.h>
#include <string.h>
#include <MFEnumPolytope.h>
#include <MFEnumDualPolytope.h>
#include <MFFortran.h>

int MFCheckHC(MFChart,MFNRegion,int,int,double*,double*);
int MFCheckHCAllPolytopes(MFAtlas,int,int,double*,double*);
void MFSubdivideHC(int,int,int*,int*,double***,double***);
MFChart MFAtlasChart(MFAtlas,int);
void Vol(MFEnumPolytope,int,int,int*,int,int*,double*);
double MFVolumeOfPolytope(MFPolytope);
double MFVolumeOfAllPolytopes(MFAtlas);

MFKVector MFVolumeTestPtK=(MFKVector)NULL;
MFNVector MFVolumeTestPtN=(MFNVector)NULL;
int      *MFVolumeIndex=(int*)NULL;

double MFVolumeOfChart(MFChart chart, MFNRegion Omega)
 {
  int i,j;
  int n,k;
  int rc;
  int nHC,mHC;
  double **HCMin=(double**)NULL;
  double **HCMax=(double**)NULL;
  double volume,volHC;
  int nIncluded;
  double refVolHC,refVolBall;
  double testVolume;
  double R;
  int verbose;

  verbose=0;

  if(verbose){printf("%s\n","MFVolumeOfChart");fflush(stdout);}

  mHC=100;
  nHC=0;
  HCMin=(double**)malloc(mHC*sizeof(double*));
  HCMax=(double**)malloc(mHC*sizeof(double*));
  for(i=0;i<mHC;i++)
   {
    HCMin[i]=(double*)NULL;
    HCMax[i]=(double*)NULL;
   }

  k=MFChartK(chart);
  n=MFChartN(chart);

  if(MFVolumeTestPtK==(MFKVector)NULL)MFVolumeTestPtK=MFCreateKVector(k);
  if(MFVolumeTestPtN==(MFNVector)NULL)MFVolumeTestPtN=MFCreateNVector(n);
  if(MFVolumeIndex==(int*)NULL)MFVolumeIndex=(int*)malloc(k*sizeof(int));
  
  volume=0.;
  testVolume=0.;
  nIncluded=0.;
  HCMin[0]=(double*)malloc(k*sizeof(double));
  HCMax[0]=(double*)malloc(k*sizeof(double));

  R=MFChartRadius(chart);
  refVolHC=1.;
  for(i=0;i<k;i++)
   {
    (HCMin[0])[i]=-R;
    (HCMax[0])[i]=R;
    refVolHC=refVolHC*2*R;
   }
  nHC=1;

  if(k==1)
    refVolBall=2*R;
   else if(k==2)
    refVolBall=3.1415926*R*R;
   else if(k==3)
    refVolBall=4./3.*3.1415926*R*R*R;
   else if(k%2==0)
    {
     refVolBall=3.1415926*R*R;
     for(i=4;i<=k;i+=2)refVolBall=2*3.1415926*R*R/i;
    }else{
     refVolBall=4./3.*3.1415926*R*R*R;
     for(i=5;i<=k;i+=2)refVolBall=2*3.1415926*R*R/i;
    }

  if(verbose)
   {
    printf("Initial List of Hypercubes %d\n",nHC);fflush(stdout);
    for(i=0;i<nHC;i++)
     {
      printf("   %d ",i);
      printf("(%lf",(HCMin[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
      printf(")<->");
      printf("(%lf",(HCMax[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
      printf(")\n");fflush(stdout);
     }
   }
  MFSubdivideHC(k,0,&nHC,&mHC,&HCMin,&HCMax);

  if(verbose)
   {
    printf("Initial Subdivided List of Hypercubes %d\n",nHC);fflush(stdout);
    for(i=0;i<nHC;i++)
     {
      printf("   %d ",i);
      printf("(%lf",(HCMin[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
      printf(")<->");
      printf("(%lf",(HCMax[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
      printf(")\n");fflush(stdout);
     }
   }

  while(nHC>0)
   {
    if(verbose){printf("Checking HyperCube\n");fflush(stdout);}
    rc=MFCheckHC(chart,Omega,k,n,HCMin[0],HCMax[0]);
    volHC=1.;
    for(i=0;i<k;i++)volHC*=(HCMax[0])[i]-(HCMin[0])[i];
    if(rc<0)
     {
      testVolume+=volHC;
      if(verbose){printf("All vertices are out, discarding\n");fflush(stdout);}
      free(HCMin[0]);
      free(HCMax[0]);
      HCMin[0]=(double*)NULL;
      HCMax[0]=(double*)NULL;
      if(nHC>1)
       {
        HCMin[0]=HCMin[nHC-1];
        HCMax[0]=HCMax[nHC-1];
       }
      nHC--;
     }else if(rc>0){
      testVolume+=volHC;
      if(verbose){printf("All vertices are in, volume is %lf\n",volHC);fflush(stdout);}
      volume+=volHC;
      nIncluded++;
      free(HCMin[0]);
      free(HCMax[0]);
      HCMin[0]=(double*)NULL;
      HCMax[0]=(double*)NULL;
      if(nHC>1)
       {
        HCMin[0]=HCMin[nHC-1];
        HCMax[0]=HCMax[nHC-1];
       }
      nHC--;
     }else{
      if(verbose){printf("Some vertices are in, some are out, volume is %lf\n",volHC);fflush(stdout);}
      if(volHC>1.e-10)
       {
        if(verbose){printf("  Subdividing it.\n");fflush(stdout);}
        MFSubdivideHC(k,0,&nHC,&mHC,&HCMin,&HCMax);
       }else{
        testVolume+=volHC;
        if(verbose){printf("  Too small, discarding it.\n");fflush(stdout);}
        free(HCMin[0]);
        free(HCMax[0]);
        HCMin[0]=(double*)NULL;
        HCMax[0]=(double*)NULL;
        if(nHC>1)
         {
          HCMin[0]=HCMin[nHC-1];
          HCMax[0]=HCMax[nHC-1];
         }
        nHC--;
       }
     }

    if(verbose)
     {
      printf("List of Hypercubes %d\n",nHC);fflush(stdout);
      for(i=0;i<nHC;i++)
       {
        printf("   %d ",i);
        printf("(%lf",(HCMin[i])[0]);
        for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
        printf(")<->");
        printf("(%lf",(HCMax[i])[0]);
        for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
        printf(")\n");fflush(stdout);
       }
     }
   }

  if(verbose)
   {
    if(verbose)printf("done %s\n","MFVolumeOfChart");
    printf("    volume of chart=%lf, nCubesInside=%d\n",volume,nIncluded);
    printf("    volume of ball =%lf,\n",refVolBall);
    printf("    volume of Cube =%lf\n",refVolHC);
    printf("    test volume    =%lf\n",testVolume);
    fflush(stdout);
   }
  return volume;
 }

int MFCheckHC(MFChart chart, MFNRegion Omega, int k, int n, double *HCMin, double *HCMax)
 {
  int i;
  int rc;
  int allIn,allOut;
  int verbose;

  verbose=0;

  if(verbose)
   {
    printf("%s k=%d, nHC=%d\n","MFCheckHC",k,n);fflush(stdout);
    printf("(%lf",HCMin[0]);
    for(i=1;i<k;i++)printf(",%lf",HCMin[i]);
    printf(")<->");
    printf("(%lf",HCMax[0]);
    for(i=1;i<k;i++)printf(",%lf",HCMax[i]);
    printf(")\n");fflush(stdout);
   }

  if(MFVolumeTestPtK==(MFKVector)NULL)MFVolumeTestPtK=MFCreateKVector(k);
  if(MFVolumeTestPtN==(MFNVector)NULL)MFVolumeTestPtN=MFCreateNVector(n);
  if(MFVolumeIndex==(int*)NULL)MFVolumeIndex=(int*)malloc(k*sizeof(int));

  allIn=1;
  allOut=1;

  for(i=0;i<k;i++)MFVolumeIndex[i]=0;

  while(1)
   {
    if(verbose)
     {
      printf(" index (%d",MFVolumeIndex[k-1]);
      for(i=1;i<k;i++)printf(",%d",MFVolumeIndex[k-1-i]);
      printf(") s=");fflush(stdout);
     }

    for(i=0;i<k;i++)
     MFKVSetC(MFVolumeTestPtK,i,HCMin[i]+MFVolumeIndex[i]*(HCMax[i]-HCMin[i]));
    if(verbose)MFPrintKVector(stdout,MFVolumeTestPtK);
    rc=MFChartInterior(chart,MFVolumeTestPtK);
    if(rc)
     {
      if(verbose)printf(" In chart, u=");
      MFChartPointInTangentSpace(chart,MFVolumeTestPtK,MFVolumeTestPtN);
      if(verbose)MFPrintNVector(stdout,MFVolumeTestPtN);
      rc=MFNRegionInterior(Omega,MFVolumeTestPtN);
      if(verbose)
       {
        if(rc)printf(" In Omega.\n");
         else printf(" Not in Omega.\n");
       }
      if(rc)allOut=0;
       else allIn=0;
     }else{
       allIn=0;
       if(verbose)printf(" Not in chart.\n");
     }

    i=0;
    MFVolumeIndex[i]++;
    while(MFVolumeIndex[i]==2)
     {
      MFVolumeIndex[i]=0;
      i++;
      if(i>=k)
       {
        if(verbose){printf("done %s allIn=%d, allOut=%d\n","MFCheckHC",allIn,allOut);fflush(stdout);}
        if(allIn)return 1;
         else if(allOut)return -1;
         else return 0;
       }
      MFVolumeIndex[i]++;
     }
   }
 }

void MFSubdivideHC(int k,int iHC,int *nHC,int *mHC,double ***HCMin,double ***HCMax)
 {
  int i;
  int verbose;

  verbose=0;
  if(verbose){printf("%s, k=%d, iHC=%d\n","MFSubdivideHC",k, iHC);fflush(stdout);}

  if(MFVolumeIndex==(int*)NULL)MFVolumeIndex=(int*)malloc(k*sizeof(int));
  for(i=0;i<k;i++)MFVolumeIndex[i]=0;

  while(1)
   {
    if(verbose)
     {
      printf(" index (%d",MFVolumeIndex[k-1]);
      for(i=1;i<k;i++)printf(",%d",MFVolumeIndex[k-1-i]);
      printf(")\n");fflush(stdout);
     }

    if(*nHC+1>*mHC)
     {
      if(verbose){printf("Resizing HCMin and HCMax\n");fflush(stdout);}
      (*mHC)+=100;
      *HCMin=(double**)realloc((void*)(*HCMin),(*mHC)*sizeof(double*));
      *HCMax=(double**)realloc((void*)(*HCMax),(*mHC)*sizeof(double*));
      if(*HCMin==(double**)NULL||*HCMax==(double**)NULL)
       {
        printf("Out of memory in SubdivideHC\n");fflush(stdout);
        exit(12);
       }
      for(i=0;i<100;i++)
       {
        (*HCMin)[*mHC-101+i]=(double*)NULL;
        (*HCMax)[*mHC-101+i]=(double*)NULL;
       }
     }

    (*nHC)++;
    (*HCMin)[*nHC-1]=(double*)malloc(k*sizeof(double));
    (*HCMax)[*nHC-1]=(double*)malloc(k*sizeof(double));

    for(i=0;i<k;i++)
     {
      if(MFVolumeIndex[i]==0)
       {
        ((*HCMin)[*nHC-1])[i]=((*HCMin)[iHC])[i];
        ((*HCMax)[*nHC-1])[i]=((*HCMin)[iHC])[i]+.5*(((*HCMax)[iHC])[i]-((*HCMin)[iHC])[i]);
       }else{
        ((*HCMin)[*nHC-1])[i]=((*HCMin)[iHC])[i]+.5*(((*HCMax)[iHC])[i]-((*HCMin)[iHC])[i]);
        ((*HCMax)[*nHC-1])[i]=((*HCMax)[iHC])[i];
       }
     }

    i=0;
    MFVolumeIndex[i]++;
    while(MFVolumeIndex[i]==2)
     {
      MFVolumeIndex[i]=0;
      i++;
      if(i>=k)
       {
        if(*nHC>1)
         {
          free((*HCMin)[iHC]);
          free((*HCMax)[iHC]);
          (*HCMin)[0]=(double*)NULL;
          (*HCMax)[0]=(double*)NULL;
          if(*nHC>1)
           {
            (*HCMin)[iHC]=(*HCMin)[(*nHC)-1];
            (*HCMax)[iHC]=(*HCMax)[(*nHC)-1];
            (*HCMin)[(*nHC)-1]=(double*)NULL;
            (*HCMax)[(*nHC)-1]=(double*)NULL;
           }
          (*nHC)--;
         }
        if(verbose){printf("done %s %d\n","MFSubdivideHC",*nHC);fflush(stdout);}
        return;
       }
      MFVolumeIndex[i]++;
     }
   }
 }

double MFVolumeOfAtlas(MFAtlas A, MFNRegion Omega)
 {
  int i,n;
  double volume;
  double volumeP;
  int verbose;

  verbose=0;

  if(verbose){printf("%s\n","MFVolumeOfAtlas");fflush(stdout);}

  volume=0.;
  volumeP=0.;
  n=MFAtlasNumberOfCharts(A);
  for(i=0;i<n;i++)
   {
/*  volume+=MFVolumeOfChart(MFAtlasChart(A,i),Omega);*/
    volumeP=MFVolumeOfPolytope(MFChartPolytope(MFAtlasChart(A,i)));
    if(verbose)printf("  volume of polytope[%d]=%lf, radius of chart=%f\n",i,volumeP,MFChartRadius(MFAtlasChart(A,i)));
    volume+=volumeP;
   }
/*volumeP=MFVolumeOfAllPolytopes(A);
  if(verbose)printf("  volume of all polytopes=%lf\n",volumeP);*/

  if(verbose){printf("done %s\n","MFVolumeOfAtlas");fflush(stdout);}
  return volume;
 }

static int Subset(int,int*,int,int*);
static int Element(int,int,int*);
int MFEnumPolytopeVerticesOnCell(MFEnumPolytope,int,int,int,int**);
static int AddElement(int,int,int*,int**);
double MFVolumeDet(MFEnumPolytope,int,int*);
static int Fact(int);
#include <MFFortran.h>


double MFVolumeOfPolytope(MFPolytope P)
 {
  double volume;
  MFEnumPolytope Q;
  int nlast,*last;
  int nS,*S;
  int i;
  int d;
  int j,iv,nv,*v;
  MFKVector Pt;
  int verbose;

  verbose=0;

  if(verbose){printf("MFVolumeOfPolytope\n");fflush(stdout);}

  if(verbose){printf(" Enumerate\n");fflush(stdout);}
  Q=MFEnumeratePolytope(P);
  if(verbose){printf(" done Enumerate\n");fflush(stdout);}

  volume=0;
  d=MFEnumPolytopeDimension(Q);
  nlast=MFEnumPolytopeNumberOfVertices(Q);
  last=(int*)malloc(nlast*sizeof(int));
  for(i=0;i<nlast;i++)last[i]=i;
  nS=0;
  S=(int*)NULL;

  if(verbose)
   {
    printf("Enumerated Polytope:\n");

    printf("  vertices:\n");
    for(i=0;i<MFEnumPolytopeNumberOfVertices(Q);i++)
     {
      Pt=MFEnumPolytopeVertex(Q,i);
      printf("  %d [%lf",i,MFKV_C(Pt,0));
      for(j=1;j<d;j++)printf(",%lf",MFKV_C(Pt,j));
      printf("]\n");fflush(stdout);
     }
  
    for(i=1;i<=d;i++)
     {
      printf("  %d-cells:\n",i);
      for(j=0;j<MFEnumPolytopeNumberOfCells(Q,i);j++)
       {
        nv=0;
        v=(int*)NULL;
        nv=MFEnumPolytopeVerticesOnCell(Q,i,j,nv,&v);
        printf("  %d",j);
        if(nv>0)
         {
          printf(" [%d",v[0]);
          for(iv=1;iv<nv;iv++)
           {
            printf(",%d",v[iv]);
           }
          printf("]\n");fflush(stdout);
         }else{
          printf(" [empty set]\n");fflush(stdout);
         }
        free(v);
       }
     }

    printf(" calc Vol\n");fflush(stdout);
   }
  Vol(Q,d,nlast,last,nS,S,&volume);
  if(verbose){printf(" done Vol, vol=%lf/%d!=%lf/%d\n",volume,d,volume,Fact(d));fflush(stdout);}
  volume=volume/Fact(d);

  MFFreeEnumPolytope(Q);
  free(last);

  return volume;
 }

void Vol(MFEnumPolytope Q, int d, int nlast, int *last, int nS, int *S, double *volume)
 {
  int i;
  int n;
  int nPsi,*psi;
  int nSnew,*Snew=(int*)NULL;
  int verbose;

  verbose=0;

  if(verbose)
   {
    printf(" Vol, dimension %d",d);fflush(stdout);
    if(nlast>0)
     {
      printf(", last=[%d",last[0]);
      for(i=1;i<nlast;i++)printf(",%d",last[i]);
      printf("]");fflush(stdout);
     }else{
      printf(", last=empty");
     }
    if(nS>0)
     {
      printf(", S=[%d",S[0]);
      for(i=1;i<nS;i++)printf(",%d",S[i]);
      printf("]");fflush(stdout);
     }else{
      printf(", S=empty");
     }
    printf("\n");fflush(stdout);
   }

  n=MFEnumPolytopeNumberOfCells(Q,d);

  if(d>-1)
   {
    for(i=0;i<n;i++)
     {
      psi=(int*)NULL;
      nPsi=0;
      nPsi=MFEnumPolytopeVerticesOnCell(Q,d,i,nPsi,&psi);
      if(Subset(nPsi,psi,nlast,last) && !Element(psi[0],nS,S))
       {
        nSnew=AddElement(psi[0],nS,S,&Snew);
        Vol(Q,d-1,nPsi,psi,nSnew,Snew,volume);
        free(Snew);
       }
      free(psi);
     }
   }else{
    if(verbose)
     {
      printf("Simplex: ");
      printf("[%d",S[0]);
      for(i=1;i<nS;i++)printf(",%d",S[i]);
      printf("]\n");fflush(stdout);
     }
    *volume+=fabs(MFVolumeDet(Q,nS,S));
   }
  return;
 }

int Subset(int n1,int *S1,int n2,int *S2)
 {
  static int i,j;
  static int result;
  static int verbose;

  verbose=0;

  if(verbose)
   {
    printf("    Subset, is");
    if(n1>0)
     {
      printf(" [%d",S1[0]);
      for(i=1;i<n1;i++)printf(",%d",S1[i]);
      printf("]");
     }else{
      printf(" the empty set");
     }
    printf(" a subset of ");
    if(n2>0)
     {
      printf(" [%d",S2[0]);
      for(i=1;i<n2;i++)printf(",%d",S2[i]);
      printf("]");
     }else{
      printf(" the empty set");
     }
    printf("?");fflush(stdout);
   }

  for(i=0;i<n1;i++)
   {
    result=0;
    for(j=0;j<n2;j++)if(S1[i]==S2[j])result=1;
    if(!result){if(verbose){printf(" No\n");fflush(stdout);}return 0;}
   }

  if(verbose){printf(" Yes\n");fflush(stdout);}
  return 1;
 }

int Element(int e,int n,int *S)
 {
  static int i;
  static int verbose;

  verbose=0;

  if(verbose)
   {
    printf("    Element, is %d an element of ");
    if(n>0)
     {
      printf(" [%d",S[0]);
      for(i=1;i<n;i++)printf(",%d",S[i]);
      printf("]");
     }else{
      printf(" the empty set");
     }
    printf("?");fflush(stdout);
   }

  for(i=0;i<n;i++)
   if(e==S[i]){if(verbose){printf(" Yes.\n");fflush(stdout);}return 1;}

  if(verbose){printf(" No.\n");fflush(stdout);}
  return 0;
 }

int AddElement(int e,int n,int *S,int **Snew)
 {
  static int i;
  static int verbose;

  verbose=0;

  if(verbose)
   {
    printf("       AddElement %d to",e); 
    if(n>0)
     {
      printf(" [%d",S[0]);
      for(i=1;i<n;i++)printf(",%d",S[i]);
      printf("]");
     }else{
      printf(" [empty list]");
     }
   }
  
  *Snew=(int*)malloc((n+1)*sizeof(int));
  i=0;
  while(i<n && S[i]<e)
   {
    (*Snew)[i]=S[i];
    i++;
   }
  (*Snew)[i]=e;
  i++;
  while(i<n+1)
   {
    (*Snew)[i]=S[i-1];
    i++;
   }

  if(verbose)
   {
    printf("=[%d",(*Snew)[0]);
    for(i=1;i<n+1;i++)printf(",%d",(*Snew)[i]);
    printf("]\n");fflush(stdout);
   }

  return n+1;
 }

double MFVolumeDet(MFEnumPolytope Q,int n,int *S)
 {
  static int i,j;
  static int d;
  static double *A=(double*)NULL;
  static MFKVector row;
  static MFKVector lastRow;
  static int *ipivot=(int*)NULL;
  static int info;
  static double det;
  static int verbose;

  verbose=0;

  d=MFEnumPolytopeDimension(Q);

  A=(double*)realloc((void*)A,d*d*sizeof(double));

  lastRow=MFEnumPolytopeVertex(Q,S[d]);
  if(n!=d+1)
   {
    printf(" problem in MFVolumeDet, n=%d, while dimension is %d\n",n,d);
    fflush(stdout);
    abort();
   }

  if(verbose)printf(" MFVolumeDet:\n");
  for(i=0;i<d;i++)
   {
    if(verbose)printf("    [");
    row=MFEnumPolytopeVertex(Q,S[i]);
    for(j=0;j<d;j++)
     {
      A[i+d*j]=MFKV_C(row,j)-MFKV_C(lastRow,j);
      if(verbose)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+d*j]);
       }
     }
    if(verbose){printf("]\n");fflush(stdout);}
   }

  if(d==1)
   {
    det=A[0];
   }else if(d==2){
    det=A[0]*A[3]-A[1]*A[2];
   }else if(d==3){
    det=A[0]*(A[4]*A[8]-A[5]*A[7])
       -A[1]*(A[3]*A[8]-A[5]*A[6])
       +A[2]*(A[3]*A[7]-A[4]*A[6]);
   }else if(d==4){
    det=A[0]*( 
               A[5]*(A[10]*A[15]-A[11]*A[14])
              -A[6]*(A[ 9]*A[15]-A[11]*A[13])
              +A[7]*(A[ 9]*A[14]-A[10]*A[13])
             )
       -A[1]*(
               A[4]*(A[10]*A[15]-A[11]*A[14])
              -A[6]*(A[ 8]*A[15]-A[11]*A[12])
              +A[7]*(A[ 8]*A[14]-A[10]*A[12])
             )
       +A[2]*(
               A[4]*(A[ 9]*A[15]-A[11]*A[13])
              -A[5]*(A[ 8]*A[15]-A[11]*A[12])
              +A[7]*(A[ 8]*A[13]-A[ 9]*A[12])
             )
       -A[3]*(
               A[4]*(A[ 9]*A[14]-A[10]*A[13])
              -A[5]*(A[ 8]*A[14]-A[10]*A[12])
              +A[6]*(A[ 8]*A[13]-A[ 9]*A[12])
             );
   }else{
/*  ipivot=(int*)realloc((void*)ipivot,d*sizeof(int));
    F77_FUNC(dgetrf,DGETRF)(&d,&d,A,&d,ipivot,&info);
    F77_FUNC(dgetrs,DGETRS)(&trans,&n,&one,A,&n,ipvt,b,&n,&info);*/
   }

  if(verbose){printf("   MFVolumeDet: %lf\n",det);fflush(stdout);}
  return det;
 }

int Fact(int n)
 {
  if(n>2)return n*Fact(n-1);
   else return 2;
 }

double MFVolumeOfAllPolytopes(MFAtlas A)
 {
  int i,j;
  int n,k;
  int rc;
  int nHC,mHC;
  double **HCMin=(double**)NULL;
  double **HCMax=(double**)NULL;
  double volume,volHC;
  int nIncluded;
  MFChart chart;
  int c;
  double R;
  MFNVector u;
  double t;
  int verbose;

  verbose=0;

  if(verbose){printf("%s\n","MFVolumeOfAllPolytopes");fflush(stdout);}

  mHC=100;
  nHC=0;
  HCMin=(double**)malloc(mHC*sizeof(double*));
  HCMax=(double**)malloc(mHC*sizeof(double*));
  for(i=0;i<mHC;i++)
   {
    HCMin[i]=(double*)NULL;
    HCMax[i]=(double*)NULL;
   }

  k=MFAtlasK(A);
  n=MFAtlasN(A);

  if(k!=n)return -1.;

  if(MFVolumeTestPtK==(MFKVector)NULL)MFVolumeTestPtK=MFCreateKVector(k);
  if(MFVolumeTestPtN==(MFNVector)NULL)MFVolumeTestPtN=MFCreateNVector(n);
  if(MFVolumeIndex==(int*)NULL)MFVolumeIndex=(int*)malloc(k*sizeof(int));
  
  volume=0.;
  nIncluded=0.;
  HCMin[0]=(double*)malloc(k*sizeof(double));
  HCMax[0]=(double*)malloc(k*sizeof(double));

  for(c=0;c<MFAtlasNumberOfCharts(A);c++)
   {
    chart=MFAtlasChart(A,c);
    R=MFChartRadius(chart);
    u=MFChartCenter(chart);
    for(i=0;i<k;i++)
     {
      if(c==0)
       {
        (HCMin[0])[i]=-1.05*R+MFNV_C(u,i);
        (HCMax[0])[i]=1.05*R+MFNV_C(u,i);
       }else{
        t=-1.05*R+MFNV_C(u,i);
        if((HCMin[0])[i]>t)(HCMin[0])[i]=t;
        t= 1.05*R+MFNV_C(u,i);
        if((HCMax[0])[i]<t)(HCMax[0])[i]=t;
       }
     }
   }

  nHC=1;
  if(verbose)
   {
    printf("Initial List of Hypercubes %d\n",nHC);fflush(stdout);
    for(i=0;i<nHC;i++)
     {
      printf("   %d ",i);
      printf("(%lf",(HCMin[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
      printf(")<->");
      printf("(%lf",(HCMax[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
      printf(")\n");fflush(stdout);
     }
   }
  MFSubdivideHC(k,0,&nHC,&mHC,&HCMin,&HCMax);

  if(verbose)
   {
    printf("Initial Subdivided List of Hypercubes %d\n",nHC);fflush(stdout);
    for(i=0;i<nHC;i++)
     {
      printf("   %d ",i);
      printf("(%lf",(HCMin[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
      printf(")<->");
      printf("(%lf",(HCMax[i])[0]);
      for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
      printf(")\n");fflush(stdout);
     }
   }

  while(nHC>0)
   {
    if(verbose){printf("Checking HyperCube\n");fflush(stdout);}
    rc=MFCheckHCAllPolytopes(A,k,n,HCMin[0],HCMax[0]);
    volHC=1.;
    for(i=0;i<k;i++)volHC*=(HCMax[0])[i]-(HCMin[0])[i];
    if(rc<0)
     {
      if(verbose){printf("All vertices are out, discarding\n");fflush(stdout);}
      free(HCMin[0]);
      free(HCMax[0]);
      HCMin[0]=(double*)NULL;
      HCMax[0]=(double*)NULL;
      if(nHC>1)
       {
        HCMin[0]=HCMin[nHC-1];
        HCMax[0]=HCMax[nHC-1];
       }
      nHC--;
     }else if(rc>0){
      if(verbose){printf("All vertices are in, volume is %lf\n",volHC);fflush(stdout);}
      volume+=volHC;
      nIncluded++;
      free(HCMin[0]);
      free(HCMax[0]);
      HCMin[0]=(double*)NULL;
      HCMax[0]=(double*)NULL;
      if(nHC>1)
       {
        HCMin[0]=HCMin[nHC-1];
        HCMax[0]=HCMax[nHC-1];
       }
      nHC--;
     }else{
      if(verbose){printf("Some vertices are in, some are out, volume is %lf\n",volHC);fflush(stdout);}
      if(volHC>1.e-10)
       {
        if(verbose){printf("  Subdividing it.\n");fflush(stdout);}
        MFSubdivideHC(k,0,&nHC,&mHC,&HCMin,&HCMax);
       }else{
        if(verbose){printf("  Too small, discarding it.\n");fflush(stdout);}
        free(HCMin[0]);
        free(HCMax[0]);
        HCMin[0]=(double*)NULL;
        HCMax[0]=(double*)NULL;
        if(nHC>1)
         {
          HCMin[0]=HCMin[nHC-1];
          HCMax[0]=HCMax[nHC-1];
         }
        nHC--;
       }
     }

    if(verbose)
     {
      printf("List of Hypercubes %d\n",nHC);fflush(stdout);
      for(i=0;i<nHC;i++)
       {
        printf("   %d ",i);
        printf("(%lf",(HCMin[i])[0]);
        for(j=1;j<k;j++)printf(",%lf",(HCMin[i])[j]);
        printf(")<->");
        printf("(%lf",(HCMax[i])[0]);
        for(j=1;j<k;j++)printf(",%lf",(HCMax[i])[j]);
        printf(")\n");fflush(stdout);
       }
     }
   }

  if(verbose)
   {
    if(verbose)printf("done %s\n","MFVolumeOfAllPolytopes");
    printf("    volume of atlas=%lf, nCubesInside=%d\n",volume,nIncluded);
    fflush(stdout);
   }
  return volume;
 }

int MFCheckHCAllPolytopes(MFAtlas A,int k,int n,double *HCMin,double *HCMax)
 {
  int i;
  int rc;
  int allIn,allOut;
  int verbose;
  int c;
  MFChart chart;
  double R;
  MFNVector u;
  int in;

  verbose=0;

  if(verbose)
   {
    printf("%s\n","MFCheckHCAllPolytopes");fflush(stdout);
    printf("(%lf",HCMin[0]);
    for(i=1;i<k;i++)printf(",%lf",HCMin[i]);
    printf(")<->");
    printf("(%lf",HCMax[0]);
    for(i=1;i<k;i++)printf(",%lf",HCMax[i]);
    printf(")\n");fflush(stdout);
   }

  if(MFVolumeTestPtK==(MFKVector)NULL)MFVolumeTestPtK=MFCreateKVector(k);
  if(MFVolumeTestPtN==(MFNVector)NULL)MFVolumeTestPtN=MFCreateNVector(n);
  if(MFVolumeIndex==(int*)NULL)MFVolumeIndex=(int*)malloc(k*sizeof(int));

  allIn=1;
  allOut=1;

  for(i=0;i<k;i++)MFVolumeIndex[i]=0;

  while(1)
   {
    if(verbose)
     {
      printf(" index (%d",MFVolumeIndex[k-1]);
      for(i=1;i<k;i++)printf(",%d",MFVolumeIndex[k-1-i]);
      printf(") s=");fflush(stdout);
      i=0;
      printf("[%lf",HCMin[i]+MFVolumeIndex[i]*(HCMax[i]-HCMin[i]));
      for(i=1;i<k;i++)printf(",%lf",HCMin[i]+MFVolumeIndex[i]*(HCMax[i]-HCMin[i]));
      printf("]");fflush(stdout);
     }

    rc=0;
    for(c=0;c<MFAtlasNumberOfCharts(A);c++)
     {
      chart=MFAtlasChart(A,c);
      R=MFChartRadius(chart);
      u=MFChartCenter(chart);
      in=1;
      for(i=0;i<k;i++)
       {
        if(HCMin[i]+MFVolumeIndex[i]*(HCMax[i]-HCMin[i])-MFNV_C(u,i)<-1.05*R)in=0;
        if(HCMin[i]+MFVolumeIndex[i]*(HCMax[i]-HCMin[i])-MFNV_C(u,i)> 1.05*R)in=0;
       }
      if(in)rc=1;
     }

    if(rc)
     {
      if(verbose)printf(" In a Chart.\n");
      if(rc)allOut=0;
       else allIn=0;
     }else{
       allIn=0;
       if(verbose)printf(" Not in a chart.\n");
     }

    i=0;
    MFVolumeIndex[i]++;
    while(MFVolumeIndex[i]==2)
     {
      MFVolumeIndex[i]=0;
      i++;
      if(i>=k)
       {
        if(verbose){printf("done %s allIn=%d, allOut=%d\n","MFCheckHCAllPolytopes",allIn,allOut);fflush(stdout);}
        if(allIn)return 1;
         else if(allOut)return -1;
         else return 0;
       }
      MFVolumeIndex[i]++;
     }
   }
 }
