/* 
    @(#)MFWeightedNSpace.c	1.7
    03/06/13 17:12:07
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */

#include <MFNSpace.h>
#include <MFNVector.h>
#include <MFError.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>

void MFSetError(int,char*,char*,int,char*);
static char MFWeightedNSpaceErrorMsg[256]="";

struct MFWeightedNSpaceData
 {
  int n;
  int l;
  double s;
  int *e;
 };

void MFFreeWeightedNSpaceData(void*);
double MFWeightedNSpaceDistance(MFNSpace,MFNVector,MFNVector,void*);
void MFWeightedNSpaceDirection(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void MFWeightedNSpaceAdd(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
void MFWeightedNSpaceScale(MFNSpace,double,MFNVector,MFNVector,void*);
double MFWeightedNSpaceInner(MFNSpace,MFNVector,MFNVector,void*);
void MFWriteWeightedNSpaceData(FILE*,MFNSpace,void*);
void MFReadWeightedNSpaceData(FILE*,MFNSpace);

MFNSpace MFCreateNSpace(int n)
 {
  static char RoutineName[]={"MFCreateNSpace"};
  MFNSpace this;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateNSpace++;
  starttime=clock();
#endif

  this=MFCreateNSpaceBaseClass("WeightedNSpace");
  MFNSpaceSetDistance(this,MFWeightedNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFWeightedNSpaceInner);
  MFNSpaceSetDirection(this,MFWeightedNSpaceDirection);
  MFNSpaceSetAdd(this,MFWeightedNSpaceAdd);
  MFNSpaceSetScale(this,MFWeightedNSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeWeightedNSpaceData);
  MFNSpaceSetWriteData(this,MFWriteWeightedNSpaceData);
  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));
  
  data->n=n;
  data->l=0;
  data->s=1.;
  data->e=(int*)NULL;

  MFNSpaceSetData(this,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFCreateNSpace+=clock()-starttime;
#endif
  return this;
 }

MFNSpace MFCreateWeightedNSpace(int n,int l,double s)
 {
  static char RoutineName[]={"MFCreateWeightedNSpace"};
  MFNSpace this;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateWeightedNSpace++;
  starttime=clock();
#endif

  this=MFCreateNSpaceBaseClass("WeightedNSpace");
  MFNSpaceSetDistance(this,MFWeightedNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFWeightedNSpaceInner);
  MFNSpaceSetDirection(this,MFWeightedNSpaceDirection);
  MFNSpaceSetAdd(this,MFWeightedNSpaceAdd);
  MFNSpaceSetScale(this,MFWeightedNSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeWeightedNSpaceData);
  MFNSpaceSetWriteData(this,MFWriteWeightedNSpaceData);
  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));
  data->n=n;
  data->l=l;
  data->s=s;
  data->e=(int*)NULL;

  MFNSpaceSetData(this,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFCreateWeightedNSpace+=clock()-starttime;
#endif
  return this;
 }

MFNSpace MFCreateNSpaceWithExponents(int n, int *e)
 {
  static char RoutineName[]={"MFCreateNSpaceWithExponents"};
  MFNSpace this;
  struct MFWeightedNSpaceData *data;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateNSpaceWithExponents++;
  starttime=clock();
#endif

  this=MFCreateNSpaceBaseClass("WeightedNSpace");
  MFNSpaceSetDistance(this,MFWeightedNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFWeightedNSpaceInner);
  MFNSpaceSetDirection(this,MFWeightedNSpaceDirection);
  MFNSpaceSetAdd(this,MFWeightedNSpaceAdd);
  MFNSpaceSetScale(this,MFWeightedNSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeWeightedNSpaceData);
  MFNSpaceSetWriteData(this,MFWriteWeightedNSpaceData);
  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));
  data->n=n;
  data->l=0;
  data->s=1.;
  data->e=(int*)malloc(n*sizeof(int));
  for(i=0;i<n;i++)data->e[i]=e[i];

  MFNSpaceSetData(this,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFCreateNSpaceWithExponents+=clock()-starttime;
#endif
  return this;
 }

MFNSpace MFCreateWeightedNSpaceWithExponents(int n,int l,double s, int *e)
 {
  static char RoutineName[]={"MFCreateWeightedNSpaceWithExponents"};
  MFNSpace this;
  struct MFWeightedNSpaceData *data;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateWeightedNSpaceWithExponents++;
  starttime=clock();
#endif

  this=MFCreateNSpaceBaseClass("WeightedNSpace");
  MFNSpaceSetDistance(this,MFWeightedNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFWeightedNSpaceInner);
  MFNSpaceSetDirection(this,MFWeightedNSpaceDirection);
  MFNSpaceSetAdd(this,MFWeightedNSpaceAdd);
  MFNSpaceSetScale(this,MFWeightedNSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeWeightedNSpaceData);
  MFNSpaceSetWriteData(this,MFWriteWeightedNSpaceData);

  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));
  data->n=n;
  data->l=l;
  data->s=s;
  data->e=(int*)malloc(n*sizeof(int));
  for(i=0;i<n;i++)data->e[i]=e[i];

  MFNSpaceSetData(this,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFCreateWeightedNSpaceWithExponents+=clock()-starttime;
#endif
  return this;
 }

void MFFreeWeightedNSpaceData(void *data)
 {
  static char RoutineName[]={"MFFreeWeightedNSpaceData"};
  struct MFWeightedNSpaceData *d;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeNSpace++;
  starttime=clock();
#endif

  d=(struct MFWeightedNSpaceData*)data;
  if(d!=(struct MFWeightedNSpaceData*)NULL)
   {
    if(d->e!=(int*)NULL)free(d->e);
    free(d);
   }

#ifdef MFTIMINGS
    MFTimeMFFreeNSpace+=clock()-starttime;
#endif
  return;
 }

double MFWeightedNSpaceDistance(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"MFWeightedNSpaceDistance"};
  double result;
  int i;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceDistance++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  if(data->e==(int*)NULL)
   {
    result=0.;
    for(i=0;i<data->l;i++)result+=pow(MFNV_C(v0,i)-MFNV_C(v1,i),2)*data->s;
    for(i=data->l;i<data->n;i++)result+=pow(MFNV_C(v0,i)-MFNV_C(v1,i),2);
    result=sqrt(result);
   }else{
    result=0.;
    for(i=0;i<data->l;i++)
      result+=pow(MFNV_C(v0,i)-MFNV_C(v1,i),2*data->e[i])*data->s;
    for(i=data->l;i<data->n;i++)
      result+=pow(MFNV_C(v0,i)-MFNV_C(v1,i),2*data->e[i]);
    result=sqrt(result);
   }

#ifdef MFTIMINGS
    MFTimeMFNSpaceDistance+=clock()-starttime;
#endif
  return result;
 }

void MFWeightedNSpaceDirection(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector diff,void *d)
 {
  static char RoutineName[]={"MFWeightedNSpaceDirection"};
  int i;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceDirection++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(diff,i,MFNV_C(v1,i)-MFNV_C(v0,i));

#ifdef MFTIMINGS
    MFTimeMFNSpaceDirection+=clock()-starttime;
#endif
  return;
 }

double MFWeightedNSpaceInner(MFNSpace this,MFNVector v0,MFNVector v1,void *d)
 {
  static char RoutineName[]={"MFWeightedNSpaceInner"};
  double result;
  int i;
  int verbose;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceInner++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  verbose=0;
  if(verbose){printf("%s 0x%8.8x\n",RoutineName,this);fflush(stdout);}
  if(data->e==(int*)NULL)
   {
    if(verbose)
     {
      printf("%s - e==NULL\n",RoutineName);fflush(stdout);
      printf("s==%lf\n",data->s);fflush(stdout);
      printf("l==%d\n",data->l);fflush(stdout);
     }
    result=0.;
    for(i=0;i<data->l;i++)
     {
      result+=MFNV_C(v0,i)*MFNV_C(v1,i)*data->s;
      if(verbose){printf("   %lf * %lf * %lf += %lf\n",MFNV_C(v0,i),MFNV_C(v1,i),data->s,result);fflush(stdout);}
     }
    for(i=data->l;i<data->n;i++)
     {
      result+=MFNV_C(v0,i)*MFNV_C(v1,i);
      if(verbose){printf("   %lf * %lf += %lf\n",MFNV_C(v0,i),MFNV_C(v1,i),result);fflush(stdout);}
     }
   }else{
    if(verbose)
     {
      printf("%s - e!=NULL\n",RoutineName);fflush(stdout);
      printf("s==%lf\n",data->s);fflush(stdout);
      printf("l==%d\n",data->l);fflush(stdout);
     }
    result=0.;
    for(i=0;i<data->l;i++)
      result+=pow(MFNV_C(v0,i),data->e[i])*pow(MFNV_C(v1,i),data->e[i])*data->s;
    for(i=data->l;i<data->n;i++)
      result+=pow(MFNV_C(v0,i)*MFNV_C(v1,i),data->e[i]);
   }

#ifdef MFTIMINGS
    MFTimeMFNSpaceInner+=clock()-starttime;
#endif
  return result;
 }

void MFWriteWeightedNSpaceData(FILE *fid,MFNSpace this,void *d)
 {
  static char RoutineName[]={"MFWriteWeightedNSpaceData"};
  int i;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteNSpace++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  fprintf(fid,"%d %d %lf\n",data->n,data->l,data->s);
  if(data->e==(int*)NULL)
   {
    fprintf(fid,"0\n");
   }else{
    fprintf(fid,"1\n");
    for(i=0;i<data->n;i++)
      fprintf(fid,"%d\n",data->e[i]);
   }

#ifdef MFTIMINGS
    MFTimeMFWriteNSpace+=clock()-starttime;
#endif
  return;
 }

void MFReadWeightedNSpaceData(FILE *fid,MFNSpace this)
 {
  static char RoutineName[]={"MFReadWeightedNSpaceData"};
  int i,n;
  char tag[100]="";
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadNSpace++;
  starttime=clock();
#endif

  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));
  if(data==(struct MFWeightedNSpaceData*)NULL)
   {
    sprintf(MFWeightedNSpaceErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFWeightedNSpaceData));
    printf("%s -- %s\n",RoutineName,MFWeightedNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFWeightedNSpaceErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
    return;
   }

  MFNSpaceSetDistance(this,MFWeightedNSpaceDistance);
  MFNSpaceSetInnerProduct(this,MFWeightedNSpaceInner);
  MFNSpaceSetDirection(this,MFWeightedNSpaceDirection);
  MFNSpaceSetAdd(this,MFWeightedNSpaceAdd);
  MFNSpaceSetScale(this,MFWeightedNSpaceScale);
  MFNSpaceSetFreeData(this,MFFreeWeightedNSpaceData);
  MFNSpaceSetWriteData(this,MFWriteWeightedNSpaceData);

  data=(struct MFWeightedNSpaceData*)malloc(sizeof(struct MFWeightedNSpaceData));

  fscanf(fid,"%d %d %lf\n",&(data->n),&(data->l),&(data->s));
  fscanf(fid,"%d\n",&n);
  if(n==0)
    data->e=(int*)NULL;
   else{
    data->e=(int*)malloc(data->n*sizeof(int));
    for(i=0;i<data->n;i++)
              fscanf(fid,"%d\n",&(data->e[i]));
   }
  MFNSpaceSetData(this,(void*)data);

#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
  return;
 }

void MFWeightedNSpaceAdd(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector sum,void *d)
 {
  static char RoutineName[]={"MFWeightedNSpaceAdd"};
  int i;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceAdd++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(sum,i,MFNV_C(v1,i)+MFNV_C(v0,i));

#ifdef MFTIMINGS
    MFTimeMFNSpaceAdd+=clock()-starttime;
#endif
  return;
 }

void MFWeightedNSpaceScale(MFNSpace this,double s, MFNVector v,MFNVector prod,void *d)
 {
  static char RoutineName[]={"MFWeightedNSpaceScale"};
  int i;
  struct MFWeightedNSpaceData *data;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceScale++;
  starttime=clock();
#endif
  data=(struct MFWeightedNSpaceData*)d;

  for(i=0;i<data->n;i++)MFNVSetC(prod,i,s*MFNV_C(v,i));

#ifdef MFTIMINGS
    MFTimeMFNSpaceScale+=clock()-starttime;
#endif
  return;
 }

