#include <stdlib.h>
#include <stdio.h>
#include <Polytope.h>
#include <math.h>
#define VERTEXLISTINCREMENT 100;
#define EDGELISTINCREMENT 100;

struct Polytope
 {
  int n;
  int nf;

  int firstVertex;        /* Doubly Linked List of Vertices */
  int freeVertex;         /* Doubly Linked List of Space for Vertices */

  int nv;                 /* Number of Vertices */
  int mv;                 /* Number of Vertices + Number of Unused Vertices */
  int *nextVertex;
  int *prevVertex;
  double *vc;             /* Coordinates of each Vertex */
  int **vi;               /* Index Set of each Vertex */

  int firstEdge;          /* Doubly Linked List of Edges */
  int freeEdge;           /* Doubly Linked List of Space for Edges */

  int ne;                 /* Number of Vertices */
  int me;                 /* Number of Vertices + Number of Unused Edges */
  int *nextEdge;
  int *prevEdge;
  int *ev;                /* Endpoints of each Edge */
  int *ei;                /* Index Set of each Edge */
 };

struct Polytope *createHyperCube(int n,double side, double *center)
 {
  static struct Polytope *P;
  static int *index;
  static double *c;
  static int i,j,v,w;

  P=createEmptyPolytope(n);
 
  index=(int*)malloc((n+1)*sizeof(int));
  c    =(double*)malloc(n*sizeof(double));

  for(i=0;i<n;i++)c[i]=0.;

  index[0]=n;
  j=0;
  while(j<n)
   {
    for(i=0;i<n;i++)
     {
      if(c[i]<.5)index[i+1]=2*i;
       else index[i+1]=2*i+1;
     }
    addVertex(P,c,index);
    j=0;
    i=0;
    c[0]=c[0]+1.;
    while(i<n&&c[i]>1.)
     {
      c[i]=0.;
      if(i<n-1)c[i+1]=c[i+1]+1.;
      j++;
      i++;
     }
   }

  v=P->firstVertex;
  while(v>-1)
   {
    w=P->nextVertex[v];
    while(w>-1)
     {
      index=intersectIndex(P->vi[v],P->vi[w]);
      if(index[0]==n-1)addEdge(P,v,w,index);
      w=P->nextVertex[w];
     }
    v=P->nextVertex[v];
   }

  P->nf=2*n;
  for(i=0;i<P->nv;i++)
   {
    for(j=0;j<P->n;j++)
     {
      P->vc[j+P->n*i]=center[j]+side*(P->vc[j+P->n*i]-.5);
     }
   }

  free(index);
  free(c);

  return P;
 }

void subtractHalfSpaceFromPolytope(struct Polytope *P, double on, double *n)
 {
  static int j,v,w,e,next;
  static double *c,*d;
  static int *iv;
  static int *index;
  static int k;
  static int i0;
  static int i1;
  static int ii;
  static double t;
  static int i;

  c=(double*)malloc(P->n*sizeof(double));
  d=(double*)malloc(4*(P->ne+P->nv)*sizeof(double));
  iv=(int*)malloc(P->ne*sizeof(int));
  index=(int*)malloc(P->n*sizeof(int));

  v=P->firstVertex;
  while(v>-1)
   {
    d[v]=-on;
    for(j=0;j<P->n;j++)d[v]=d[v]+P->vc[j+P->n*v]*n[j];
/*  printf("Vertex %d, d=%le\n",v,d[v]);fflush(stdout);*/
    v=P->nextVertex[v];
   }

  k=0;
  e=P->firstEdge;
  while(e>-1)
   {
    i0=P->ev[0+2*e];
    i1=P->ev[1+2*e];
    next=P->nextEdge[e];
    if( (d[i0]>0.&&d[i1]>0.) || (fabs(d[i0])<1.e-10&&d[i1]>0.) || (fabs(d[i1])<1.e-10&&d[i0]>0.) || ( fabs(d[i0])<1.e-10&&fabs(d[i1])<1.e-10) )
     {
      removeEdge(P,e);
     }else if(d[i0]*d[i1]<0. && fabs(d[i0])>1.e-10 && fabs(d[i1])>1.e-10)
     {
      if(d[i0]>0.)
       {
        ii=i0;
        i0=i1;
        i1=ii;
       }
      t=-d[i0]/(d[i1]-d[i0]);
      for(j=0;j<P->n;j++)c[j]=(1-t)*P->vc[j+P->n*i0]+t*P->vc[j+P->n*i1];
      
      for(j=0;j<P->n;j++)index[j]=P->ei[j+P->n*e];
      insertIndex(&index,P->nf);

      iv[k]=addVertex(P,c,index);
      d[iv[k]]=-1.; /* To keep from adding an index to it, or removing it */

      P->ev[0+2*e]=i0;
      P->ev[1+2*e]=iv[k];
      k=k+1;
     }

    e=next;
   }

  v=P->firstVertex;
  while(v>-1)
   {
    w=P->nextVertex[v];
/*  if(d[v]>1.e-10)printf("remove vertex %d, d=%lf\n",v,d[v]);fflush(stdout);*/
    if(d[v]>1.e-10)removeVertex(P,v);
    if(fabs(d[v])<1.e-10)
     {
      insertIndex(&P->vi[v],P->nf);
      iv[k]=v;k++;
     }
    v=w;
   }

  for(i=0;i<k;i++)
   {
    for(j=i+1;j<k;j++)
     {
      index=intersectIndex(P->vi[iv[i]],P->vi[iv[j]]);
      if(index[0]==P->n-1)addEdge(P,iv[i],iv[j],index);
     }
   }

  P->nf++;
  free(c);
  free(d);
  free(iv);
  free(index);

  return;
 }

int addVertex(struct Polytope *P, double *c, int *index)
 {
  static int i;

  if(P->freeVertex<0)
   {
    P->freeVertex=P->mv;
    P->mv+=VERTEXLISTINCREMENT;
    P->nextVertex=(int*)realloc((void*)P->nextVertex,P->mv*sizeof(int));
    P->prevVertex=(int*)realloc((void*)P->prevVertex,P->mv*sizeof(int));
    P->vc=(double*)realloc((void*)P->vc,P->mv*P->n*sizeof(double));
    P->vi=(int**)realloc((void*)P->vi,P->mv*sizeof(int*));

    for(i=P->freeVertex;i<P->mv;i++)
     {
      if(i+1<P->mv)P->nextVertex[i]=i+1;
       else P->nextVertex[i]=-1;
      if(i-1>P->freeVertex)P->prevVertex[i]=i-1;
       else P->prevVertex[i]=-1;
      P->vi[i]=(int*)NULL;
     }
   }

  i=P->firstVertex;
  P->firstVertex=P->freeVertex;
  P->freeVertex=P->nextVertex[P->freeVertex];

  if(i>-1)
    P->prevVertex[i]=P->firstVertex;
  if(P->freeVertex>-1)
    P->prevVertex[P->freeVertex]=-1;

  P->nextVertex[P->firstVertex]=i;
  P->prevVertex[P->firstVertex]=-1;

  for(i=0;i<P->n;i++)P->vc[i+P->n*P->firstVertex]=c[i];

  P->vi[P->firstVertex]=(int*)malloc((index[0]+1)*sizeof(int));
  for(i=0;i<index[0]+1;i++)(P->vi[P->firstVertex])[i]=index[i];

  P->nv++;
  i=P->firstVertex;
  
/*printf(" vertices -- \n");fflush(stdout);
  printf(" first=%d\n",i);fflush(stdout);
  while(i>-1){printf(" %d next=%d, prev=%d\n",i,P->nextVertex[i],P->prevVertex[i]);i=P->nextVertex[i];}
  printf(" last =%d\n",i);fflush(stdout); 
  i=P->freeVertex;
  printf(" free=%d\n",i);fflush(stdout);
  while(i>-1){printf(" %d next=%d, prev=%d\n",i,P->nextVertex[i],P->prevVertex[i]);fflush(stdout);i=P->nextVertex[i];}
  printf(" last =%d\n",i);fflush(stdout); */

  return P->firstVertex;
 }

int addEdge(struct Polytope *P, int e0, int e1, int *index)
 {
  static int i,j;

  if(P->freeEdge<0)
   {
    P->freeEdge=P->me;
    P->me+=EDGELISTINCREMENT;
    P->nextEdge=(int*)realloc((void*)P->nextEdge,P->me*sizeof(int));
    P->prevEdge=(int*)realloc((void*)P->prevEdge,P->me*sizeof(int));
    P->ev=(int*)realloc((void*)P->ev,2*P->me*P->n*sizeof(int));
    P->ei=(int*)realloc((void*)P->ei,P->me*P->n*sizeof(int));
    for(i=P->freeEdge;i<P->me;i++)
     {
      if(i+1>P->me)P->nextEdge[i]=i+1;
       else P->nextEdge[i]=-1;
      if(i-1<P->freeEdge)P->prevEdge[i]=i-1;
       else P->prevEdge[i]=-1;
      P->ei[0+P->n*i]=0;
      for(j=1;j<P->n;j++)P->ei[j+P->n*i]=-1;
     }
   }

  i=P->firstEdge;
  P->firstEdge=P->freeEdge;
  P->freeEdge=P->nextEdge[P->freeEdge];

  if(i>-1)P->prevEdge[i]=P->firstEdge;
  if(P->freeEdge>-1)P->prevEdge[P->freeEdge]=-1;

  P->nextEdge[P->firstEdge]=i;
  P->prevEdge[P->firstEdge]=-1;

  P->ev[0+2*P->firstEdge]=e0;
  P->ev[1+2*P->firstEdge]=e1;
  for(i=0;i<P->n;i++)P->ei[i+P->n*P->firstEdge]=index[i];

  P->ne++;

  return P->firstEdge;
 }

void removeVertex(struct Polytope *P,int vertex)
 {
  static int i;

  if(vertex==P->firstVertex)
   {
    P->firstVertex=P->nextVertex[vertex];
    P->prevVertex[P->firstVertex]=-1;
    P->nextVertex[vertex]=P->freeVertex;
    P->prevVertex[vertex]=-1;
    if(P->freeVertex>-1)
      P->prevVertex[P->freeVertex]=vertex;
    P->freeVertex=vertex;
   }else{
    if(P->prevVertex[vertex]>-1)
      P->nextVertex[P->prevVertex[vertex]]=P->nextVertex[vertex];
    if(P->nextVertex[vertex]>-1)
      P->prevVertex[P->nextVertex[vertex]]=P->prevVertex[vertex];

    P->nextVertex[vertex]=P->freeVertex;
    P->prevVertex[vertex]=-1;

    if(P->freeVertex>-1)
      P->prevVertex[P->freeVertex]=vertex;
    P->freeVertex=vertex;
   }

  if(P->vi[vertex]!=(int*)NULL)free(P->vi[vertex]);
  P->vi[vertex]=(int*)NULL;

  P->nv--;

  i=P->firstVertex;
/*printf("after remove vertex %d\n",vertex);fflush(stdout);
  printf(" first=%d\n",i);fflush(stdout);
  while(i>-1){printf(" %d next=%d, prev=%d\n",i,P->nextVertex[i],P->prevVertex[i]);i=P->nextVertex[i];}
  printf(" last =%d\n",i);fflush(stdout);
  i=P->freeVertex;
  printf(" free=%d\n",i);fflush(stdout);
  while(i>-1){printf(" %d next=%d, prev=%d\n",i,P->nextVertex[i],P->prevVertex[i]);i=P->nextVertex[i];}
  printf(" last =%d\n",i);fflush(stdout);*/

  return;
 }

void removeEdge(struct Polytope *P,int edge)
 {
  static int i,j;

  if(edge==P->firstEdge)
   {
    P->firstEdge=P->nextEdge[edge];
    P->prevEdge[P->firstEdge]=-1;
    P->nextEdge[edge]=P->freeEdge;
    P->prevEdge[edge]=-1;
    if(P->freeEdge>-1)P->prevEdge[P->freeEdge]=-1;
    P->freeEdge=edge;
   }else{
    if(P->prevEdge[edge]>-1)
      P->nextEdge[P->prevEdge[edge]]=P->nextEdge[edge];
    if(P->nextEdge[edge]>-1)
      P->prevEdge[P->nextEdge[edge]]=P->prevEdge[edge];

    P->nextEdge[edge]=P->freeEdge;
    P->prevEdge[edge]=-1;

    if(P->freeEdge>-1)P->prevEdge[P->freeEdge]=edge;
    P->freeEdge=edge;
   }

  P->ev[0+2*edge]=-1;
  P->ev[1+2*edge]=-1;
  P->ei[0+P->n*edge]=0;
  for(j=1;j<P->n;j++)P->ei[j+P->n*edge]=-1;

  P->ne--;

  return;
 }

struct Polytope *createEmptyPolytope(int n)
 {
  static struct Polytope *P;
  static int i,j;

  P=(struct Polytope*)malloc(sizeof(struct Polytope));

  P->n=n;
  P->nf=0;

  P->firstVertex=-1;
  P->freeVertex=0;

  P->nv=0;
  P->mv=VERTEXLISTINCREMENT;
  P->nextVertex=(int*)malloc(P->mv*sizeof(int));
  P->prevVertex=(int*)malloc(P->mv*sizeof(int));
  P->vc=(double*)malloc(P->mv*P->n*sizeof(double));
  P->vi=(int**)malloc(P->mv*sizeof(int*));
  for(i=0;i<P->mv;i++)
   {
    if(i+1<P->mv)P->nextVertex[i]=i+1;
     else P->nextVertex[i]=-1;
    if(i-1>0)P->prevVertex[i]=i-1;
     else P->prevVertex[i]=-1;
    P->vi[i]=(int*)NULL;
   }

  P->firstEdge=-1;
  P->freeEdge=0;

  P->ne=0;
  P->me=EDGELISTINCREMENT;
  P->nextEdge=(int*)malloc(P->me*sizeof(int));
  P->prevEdge=(int*)malloc(P->me*sizeof(int));
  P->ev=(int*)malloc(2*P->me*sizeof(int));
  P->ei=(int*)malloc(P->n*P->me*sizeof(int));

  for(i=0;i<P->me;i++)
   {
    if(i+1<P->me)P->nextEdge[i]=i+1;
     else P->nextEdge[i]=-1;
    if(i-1>0)P->prevEdge[i]=i-1;
     else P->prevEdge[i]=-1;
    P->ei[0+P->n*i]=0;
    for(j=1;j<P->n;j++)P->ei[j+P->n*i]=-1;
   }

  return P;
 }

int *intersectIndex(int *index0,int *index1)
 {
  static int n0,i0,n1,i1,i,n;
  static int *index;

  n0=index0[0];
  n1=index1[0];

  n=n0;if(n1>n)n=n1;
  index=(int*)malloc((n+1)*sizeof(int));

  i0=1;
  i1=1;
  i=1;
  while(i0<n0+1&&i1<n1+1)
   {
    if(index0[i0]==index1[i1])
     {
      index[i]=index0[i0];
      i0++;
      i1++;
      i++;
     }else if(index0[i0]<index1[i1])i0++;
      else if(index1[i1]<index0[i0])i1++;
   }
  index[0]=i-1;

  return index;
 }

void insertIndex(int **index,int f)
 {
  static int i,j,n;

/*
  printf("insertIndex %d into (",f);
  for(i=1;i<(*index)[0]+1;i++)
   {
    if(i>1)printf(",");
    printf("%d",(*index)[i]);
   }
  printf(")\n");fflush(stdout);
*/

  n=(*index)[0];
  i=1;
  while(i<n && (*index)[i]<f)i++;

  if(i<n+1&&(*index)[i]==f)return;

  *index=(int*)realloc((void*)(*index),(n+2)*sizeof(int));

  if(i==n)
   {
    (*index)[n+1]=f;
   }else{
    for(j=n-1;j>=i;j--)(*index)[j+1]=(*index)[j];
    (*index)[i]=f;
   }

  (*index)[0]++;

/*
  printf("done insertIndex (");
  for(i=1;i<(*index)[0]+1;i++)
   {
    if(i>1)printf(",");
    printf("%d",(*index)[i]);
   }
  printf(")\n");fflush(stdout);
*/
  return;
 }

void printPolytope(struct Polytope *P)
 {
  int i,e,v;

  printf("Polytope, dimension %d:\n",P->n);
  printf("  %d vertices\n",P->nv);
  v=P->firstVertex;
  while(v>-1)
   {
    printf(" %d   (",v);
    for(i=0;i<P->n;i++)
     {
      if(i>0)printf(",");
      printf("%lf",P->vc[i+P->n*v]);
     }
    printf(")   [");
    for(i=1;i<(P->vi[v])[0]+1;i++)
     {
      if(i>1)printf(",");
      printf("%ld",(P->vi[v])[i]);
     }
    printf("]\n");

    v=P->nextVertex[v];
   }

  printf("  %d edges\n",P->ne);
  e=P->firstEdge;
  while(e>-1)
   {
    printf(" %d   (%d,%d)",e,P->ev[0+2*e],P->ev[1+2*e]);
    printf("   [");
    for(i=1;i<P->ei[0+P->n*e]+1;i++)
     {
      if(i>1)printf(",");
      printf("%ld",P->ei[i+P->n*e]);
     }
    printf("]\n");

    e=P->nextEdge[e];
   }

  return;
 }

void PolytopeToDXFile(char *name,struct Polytope *P)
 {
  static char RoutineName[]={"PolytopeToDXFile"};
  int nV,nF,n,d;
  FILE *fid;
  int *index,*iv,*nedges;
  int i,j,k,l,m,prevpt;
  int *edges;

  fid=fopen(name,"w");
  iv=(int*)malloc(P->mv*sizeof(int));for(i=0;i<P->mv;i++)iv[i]=-1;

  nV=P->nv;
  d=P->n;
  fprintf(fid,"object \"Points\" class array type float rank 1 shape %d items %d data follows\n",d,nV);fflush(stdout);
  i=P->firstVertex;k=0;
  while(i>-1)
   {
    fprintf(fid,"     ");
    for(j=0;j<d;j++)
     {
      if(j>0)fprintf(fid," ");
      fprintf(fid,"%lf",P->vc[j+d*i]);fflush(stdout);
     }
    fprintf(fid,"\n");
    iv[i]=k;k++;
    i=P->nextVertex[i];
   }
  fprintf(fid,"\n");fflush(stdout);

  index=(int*)malloc((P->nf+1)*sizeof(int));
  index[0]=0;
  i=P->firstEdge;
  while(i>-1)
   {
    for(j=1;j<P->n;j++)insertIndex(&index,P->ei[j+P->n*i]);
    i=P->nextEdge[i];
   }

  nF=index[0];

  nedges=(int*)malloc(nF*sizeof(int));

  fprintf(fid,"object \"Faces\" class array type int rank 1 shape 1 items %d data follows\n",nF);fflush(stdout);
  for(j=0;j<nF;j++)
   {
    fprintf(fid,"%d\n",j);
   }
  fprintf(fid,"attribute \"ref\" string \"loops\"\n");

  l=0;
  m=0;
  for(j=0;j<nF;j++)
   {
    nedges[j]=0;
    i=P->firstEdge;
    while(i>-1)
     {
      for(k=1;k<P->ei[0+d*i]+1;k++)
       {
        if(P->ei[k+d*i]==index[j+1])nedges[j]++;
       }
      i=P->nextEdge[i];
     }
    l+=nedges[j];
    if(nedges[j]>m)m=nedges[j];
   }
  fprintf(fid,"\n");fflush(stdout);
  edges=(int*)malloc(m*sizeof(int));

  fprintf(fid,"object \"Edges\" class array type int rank 1 shape 1 items %d data follows\n",l);fflush(stdout);
 
  for(j=0;j<nF;j++)
   {
    m=0;
    i=P->firstEdge;
    while(i>-1)
     {
      for(k=1;k<P->ei[0+d*i]+1;k++)
       {
        if(P->ei[k+d*i]==index[j+1])
         {
          edges[m]=i;
          m++;
         }
       }
      i=P->nextEdge[i];
     }
    prevpt=P->ev[0+2*edges[0]];
    fprintf(fid,"%d\n",iv[P->ev[0+2*edges[0]]]);
    edges[0]=-1;
    for(k=0;k<m;k++)
     {
      if(edges[k]>-1 && P->ev[0+2*edges[k]]==prevpt)
       {
        prevpt=P->ev[1+2*edges[k]];
        fprintf(fid,"%d\n",iv[P->ev[1+2*edges[k]]]);
        edges[k]=-1;
        k=0;
       }else if(edges[k]>-1 && P->ev[1+2*edges[k]]==prevpt)
       {
        prevpt=P->ev[0+2*edges[k]];
        fprintf(fid,"%d\n",iv[P->ev[0+2*edges[k]]]);
        edges[k]=-1;
        k=0;
       }
     }
   }
  fprintf(fid,"attribute \"ref\" string \"positions\"\n\n");

  fprintf(fid,"object \"Loops\" class array type int rank 1 shape 1 items %d data follows\n",nF);fflush(stdout);
  k=0;
  for(j=0;j<nF;j++)
   {
    fprintf(fid,"%d\n",k);k+=nedges[j];
   }
  fprintf(fid,"attribute \"ref\" string \"loops\"\n\n");
  fprintf(fid,"attribute \"ref\" string \"edges\"\n\n");

  fprintf(fid,"object \"Polytope\" class field\n");
  fprintf(fid,"component \"positions\" value \"Points\"\n");
  fprintf(fid,"component \"faces\" value \"Faces\"\n");
  fprintf(fid,"component \"loops\" value \"Loops\"\n");
  fprintf(fid,"component \"edges\" value \"Edges\"\n");

  fprintf(fid,"\n");
  
  fprintf(fid,"object \"All\" class group\n");
  i=0;
  fprintf(fid,"member \"polytope\" value \"Polytope\"\n",i);i++;
  fprintf(fid,"\nDefault \"All\"\n");
  fprintf(fid,"\nend\n");

  fclose(fid);
  free(iv);
  free(index);
  free(nedges);

  return;
 }
