/* 
    %W%
    %D% %T%
   

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFNRegion.h>
#include <MFNVector.h>
#include <math.h>

#define PI 3.14159265358979323846264338327950288

/*
u'=f(u,t,l);                     n equations
a(u(0),l)+b(u(1),l)+c(l)=0;    n equations
int l(u(t),t,l) dt + m(l) = 0;   nc equations
*/

double q(double l){return l*l*exp(-l/2);}
double ql(double l){return l*(2-l/2)*exp(-l/2);}
double p(double z){return z*(1-z*(1-z*(1-z*(1-z*(1-z*(1-z*(1-z*(1-2*z))))))));}
double pz(double z){return 1-z*(2-z*(3-z*(4-z*(5-z*(6-z*(7-z*(8-18*z)))))));}
double u0(double x,double l,double t){return t*q(l)*x*(1-x);}
double u0l(double x,double l,double t){return t*ql(l)*x*(1-x);}
double u0t(double x,double l,double t){return q(l)*x*(1-x);}

void f(double *u,double *f)
 {
  f[0]=u[1];
  f[1]=-2*q(u[3])-PI*PI*u[3]*p(u[0]-u0(u[2],u[3],u[4]));

  return;
 }

void fu(double *u,double *fu)
 {
  fu[0+2*0]=0.;
  fu[0+2*1]=1.;
  fu[1+2*0]=-PI*PI*u[3]*pz(u[0]-u0(u[2],u[3],u[4]));
  fu[1+2*1]=0.;

  return;
 }

void fl(double *u,double *fl)
 {
  fl[0+2*0]=0.;
  fl[0+2*1]=0.;
  fl[1+2*0]=-2*ql(u[3])-PI*PI*p(u[0]-u0(u[2],u[3],u[4]))
                       -PI*PI*u[3]*pz(u[0]-u0(u[2],u[3],u[4]))*u0l(u[2],u[3],u[4]);
  fl[1+2*1]=           -PI*PI*u[3]*pz(u[0]-u0(u[2],u[3],u[4]))*u0t(u[2],u[3],u[4]);

  return;
 }

void a(double *u,double *a)
 {
  a[0]=u[0];
  a[1]=0;
  return;
 }

void au(double *u,double *au)
 {
  au[0+2*0]=1.;
  au[0+2*1]=0.;
  au[1+2*0]=0.;
  au[1+2*1]=0.;
  return;
 }

void al(double *u,double *al)
 {
  al[0]=0.;
  al[1]=0.;
  return;
 }

void b(double *u,double *b)
 {
  b[0]=0.;
  b[1]=u[0];
  return;
 }

void bu(double *u,double *bu)
 {
  bu[0+2*0]=0.;
  bu[0+2*1]=0.;
  bu[1+2*0]=1.;
  bu[1+2*1]=0.;
  return;
 }

void bl(double *u,double *bl)
 {
  bl[0]=0.;
  bl[1]=0.;
  return;
 }

void c(double *p,double *c)
 {
  c[0]=0.;
  c[1]=0.;
  return;
 }

void cl(double *p,double *cl)
 {
  cl[0]=0.;
  cl[1]=0.;
  return;
 }

void l(double *u,double *l)
 {
  return;
 }

void lu(double *u,double *lu)
 {
  return;
 }

void ll(double *u,double *ll)
 {
  return;
 }

void m(double *u,double *m)
 {
  return;
 }

void ml(double *u,double *ml)
 {
  return;
 }

int main(int argc, char *argv[])
 {
  MFImplicitMF M;
  int i,j,n;
  MFNRegion Omega;
  MFAtlas S;
  MFNVector u0;
  FILE *fid;
  MFNVector ug;
  MFNKMatrix Tan;
  MFNVector dug[2];
  MFNKMatrix Tan0;
  int nw=1;

  double t,L;

  int nx=5;
  int np=2;
  int nu=2;
  int nc=0;

  M=MFIMFCreateTPBVP(nx,nu,np,nc,f,fu,fl,a,au,al,b,bu,bl,c,cl,l,lu,ll,m,ml);
  n=MFIMF_N(M);
  Omega=MFNRegionCreateHyperCube(n,100.);

  ug=MFCreateNVector(n);
  dug[0]=MFCreateNVector(n);
  dug[1]=MFCreateNVector(n);

  L=1.;

  for(i=0;i<nx;i++)
   {
    t=(i-.5)/(nx-2);

    MFNVSetC( ug,nu*nx+np+i,t);

    MFNVSetC( ug,nu*i+0,0.);  /* x[i] */
    MFNVSetC( ug,nu*i+1,0.);  /* y[i] */

    MFNVSetC(dug[0],nu*nx+np+i,t);
    MFNVSetC(dug[0],nu*i+0,0.); /* dx[i] */
    MFNVSetC(dug[0],nu*i+1,0.); /* dy[i] */

    MFNVSetC(dug[1],nu*nx+np+i,t);
    MFNVSetC(dug[1],nu*i+0,0.); /* dx[i] */
    MFNVSetC(dug[1],nu*i+1,0.); /* dy[i] */
   }

  t=(nx-.5)/(nx-2);
  MFNVSetC( ug,nu*nx+np+nx,t);
  MFNVSetC( ug,nu*nx  ,L);

  MFNVSetC(dug[0],nu*nx+np+nx,t);
  MFNVSetC(dug[0],nu*nx  ,1.);
  MFNVSetC(dug[0],nu*nx+1,0.);

  MFNVSetC(dug[1],nu*nx+np+nx,t);
  MFNVSetC(dug[1],nu*nx  ,0.);
  MFNVSetC(dug[1],nu*nx+1,1.);

  Tan0=MFCreateNKMatrix(MFIMF_K(M),dug);
  Tan=MFIMFTangentSpaceWithGuess(M,ug,Tan0);
  u0=MFCreateNVector(n);
  MFIMFProject(M,ug,Tan,u0);
  MFFreeNVector(ug);

  S=MFComputeAtlas(M,Omega,u0,.0001,100,1);

  fid=fopen("TPBVP.atlas","w");
  MFWriteAtlas(fid,S);

  MFFreeImplicitMF(M);
  MFFreeNRegion(Omega);
  MFFreeNVector(u0);

  return(0);
 }
